//
// Project:	Lhogho
// File:	atoms.c
//
// Copyright (C) 2007 P.Boytchev
//
// Revision history:
//   2006-09-28	- file created
//   2006-09-30	- adjusted to work with pools
//   2006-10-01 - new_word()
//		- new_subword()
//   2006-10-07 - introduced *_t types
//   2006-10-08 - mbstowcs(), wcstombs(), load_file)
//   2006-10-10 - mbstowcs(), wcstombs(), load_file() moved
//		  to new unit util.c
//		- changed debug and statistic macros
//		- made statistics more detailed
//   2006-10-11 - removed ADVANCED macro
//   2006-10-11 - UNICODE macro
//		- parse_data_list()
//   2006-10-13 - append()
//		- list support in dump_atom()
//		- fixed parse_data_list()
//		- parse_data_list() moved to parser.c
//   2006-10-14 - fixed bug in dump of empty list
//   2006-10-26 - new_error()
//		- error aroms supported by dump()
//   2006-10-27 - clearing flags in new_list()
//   2007-02-15 - removing EXPR and LISP types of lists
//   2007-02-25 - changed dump_atom() for error atoms
//		- dump_stdout() accepts len==0
//   2007-02-26 - dump_atom() for parsed lists
//   2007-02-27 - new_var()
//		- dump_atom() for vars
//		- module renamed to ATOMS
//   2007-03-01 - fixed dump_atom() for vars
//		- DEVELOPMENT macro is used
//   2007-03-18 - new_atom()
//		- same_words()
//   2007-05-17 - license info
//   2007-05-22 - doxygen-friendly documentation
//   2007-05-27 - dumps error text of error atoms
//		- behead()
//		- added ERROR_VACUUM_TO_END, ERROR_WHERES_NAME
//		  to error_texts[]
//		- new_error_atom accepts list input
//   2007-05-29 - dump function bodies
//		- init_output(), use_stdout()
//		- cfg_case_sensitive renamed to option_case_insensitive
//   2007-06-01 - added TR_ERROR_BEG_ME_PARDON to error_texts[]
//		- error texts moved to errors.h
//		- STATISTICS and DEVELOPMENT merged into ADVANCED
//		- new_var() accepts anonymous vars
//   2007-06-05 - definitions spread to where they belong
//   2007-06-08 - Subwords can be used as host words
//   2007-06-09 - dump_integer(), dump_float(), dump_list(), dump_word()
//		- word-related stuff moved to words module
//		- number-related stuff moved to words module
//   2007-06-13	- fixed bug #1736021 "Alloc/dealloc statistics"
//   2007-06-19 - added mem atoms
//   2007-12-21 - fixed bug #1851865 Locale information not considered
//   2009-06-02 - added use_stdin()
//   2010-06-25 - added innereof()
//   2011-10-07 - Fixed problem with GCC 4+
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//



#include "globals.h"
#ifdef UNICODE_CHARS
#include <wchar.h>	// UNICODE support: wcslen()
#endif //UNICODE_CHARS

#include <assert.h>	// assert()
#include <errno.h>	// errno, ENOMEM
#include <stdio.h>	// fopen(), fread(), fclose()
#include <stdlib.h>     // malloc()
#include <string.h>     // memcpy()
#include <locale.h>	// LC_ALL

#include "atoms.h"
#include "numbers.h"
#include "words.h"
#include "lists.h"
#include "errors.h"
#include "mems.h"
#include "pools.h"
#include "unicode.h"
#include "translate.h"
#include "vars.h"
#include "parser.h"
#include "compiler.h"
#include "options.h"




inner_t		  std_inner;		      ///< hook variable for inputting text
inner_eof_t	std_inner_eof;	    ///< hook variable for inputting text
outter_t	  std_outter;		      ///< hook variable for outputting text
int	    	  outter_size=0;	///< counts outted characters
FILE*       input_stream  = NULL;  ///< current input stream
FILE*       output_stream = NULL;  ///< current input stream
FILE*       dribble_handle = NULL; ///< Handle of the dribble file
struct lconv *locale_info;	///< locale information


//===================================================
/// \fn void init_atoms( )
/// \brief	initializes the Atoms module
///
/// Creates a data pool for all atoms and creates one
/// special atom representing the empty list.
//===================================================
void init_atoms( )
{
  #ifdef DEBUG_ATOM
  printf("<ATOM> Atoms initialized\n");
  #endif //DEBUG_ATOM

//#ifdef __LP64__
//#warning Definitely 64 bit
//#endif
//#ifndef __LP64__
//#warning Maybe 32 bit
//#endif
  
  #ifdef ADVANCED
  int i;
  stats_free = 0;
  stats_allocs = 0;
  for( i=MIN_ID; i<MAX_ID; i++ )
  {
    stats[i].max      = 0;
    stats[i].allocs   = 0;
    stats[i].deallocs = 0;
  }
  #endif //ADVANCED

  #ifdef SAFEMODE
  assert( sizeof(atomrec_t)==16 );
  #endif

  //setlocale( LC_ALL, "" );
  locale_info = localeconv();

  init_pool( &data_pool, ATOM_SIZE );
  init_pool( &data_pool_ex, ATOM_SIZE_EX );

  empty_list = (atom_t)take_from_pool( &data_pool );
  REF(empty_list) = 1;
  ID(empty_list) = LIST_ID;
  CAR(empty_list) = empty_list;
  CDR(empty_list) = empty_list;
  FLAGS(empty_list) = 0;
}




//===================================================
/// \fn void finit_atoms( )
/// \brief	finalizes the Atoms module
///
/// Returns the empty list back to the data pool.
//===================================================
void finit_atoms( )
{
  return_to_pool( &data_pool, empty_list );

  #ifdef DEBUG_MEMORY_LEAKS
  dump_pool();
  #endif

  #ifdef DEBUG_ATOM
  printf("<ATOM> Atoms finalized\n");
  #endif //DEBUG_ATOM
}




//===================================================
/// \fn		atom_t use( atom_t a ) 
/// \brief	adds link to atom
/// \param a	atom to use
///
/// Adds a link to an atom by increasing atom's
/// reference count.
//===================================================
atom_t use( atom_t a )
{
  if( IS_UNBOUND(a) || IS_EMPTY(a) || IS_STOPPED(a) )
  return a;

  //if( a==0x49f3b8)
  //{
  //printf("USE.BUG[%08x] ref %d->%d\n", (int)a, REF(a), REF(a)+1 );
  //}


  #ifdef DEBUG_RUNTIME_ATOMS
  if( running_compiled_code )
  {
    outter( TEXT("<RUNTIME> use  "), -1 );
    dump_atom_address( a );
    dump_atom( a, 1 );
    outter( TEXT("\n"), -1 );
  }
  #endif
  #ifdef DEBUG_COMPILETIME_ATOMS
  if( compiling_code )
  {
    outter( TEXT("<COMPILETIME> use  "), -1 );
    dump_atom_address( a );
    dump_atom( a, 1 );
    outter( TEXT("\n"), -1 );
  }
  #endif

  //if( IS_ERROR(a) )
  //  {
  //    printf("TO BE USEIT hex(a)=%x id=%d REF=%d a=",(int)a,ID(a),REF(a)); dumpln(a);
  //  }

  #ifdef SAFEMODE
  assert( a );
  assert( ID(a)<MAX_ID );
  assert( REF(a)>=0 ); // 2009 was >0
  #endif //SAFEMODE

  //if( IS_INTEGER(a) || IS_FLOAT(a))
  //{
  //printf("deuse[addr=%x ref=%d]\n",a,REF(a));
  //printf("use atom="); dumpln(a);
  //}

  REF(a)++;

  #ifdef DEBUG_ATOM
  printf("<ATOM>  [%08x] ref+1\n",(int)a);
  #endif //DEBUG_ATOM
  return a;
}




//===================================================
/// \fn		void deuse( atom_t a ) 
/// \brief	removes link to atom
/// \param a	atom to unlink
///
/// Decrements the reference count for the atom and
/// if it reaches 0 then frees the atom using one of
/// the functions ::delete_numeric, ::delete_list,
/// ::delete_word, ::delete_subword, ::delete_error,
/// ::delete_var.
//===================================================
#define void void __attribute__ ((used,noinline,regparm(0),stdcall))
void deuse ( atom_t a )
#undef void
{
  __asm__ volatile ( ASM_STORE_RESULT:::ASM_CLOBBER_REGISTERS );

  typedef void(*deleter_t)(atom_t);

  static deleter_t deleters[MAX_ID] = {
    delete_numeric,	// INTEGER_ID
    delete_numeric,	// FLOAT_ID
    delete_list,	// LIST_ID
    delete_word,	// WORD_ID
    delete_subword,	// SUBWORD_ID
    delete_error,	// ERROR_ID
    delete_var,		// VAR_ID
    delete_mem,		// MEM_ID
  }; // array of deleter functions for each atom type

  //if( a==0x49f3b8)
  //{
  //printf("DEUSE.BUG[%08x] ref %d->%d\n", (int)a, REF(a), REF(a)-1 );
  //}

  //if( IS_INTEGER(a) || IS_FLOAT(a))
  //  {
  //    //printf("deuse[addr=%x ref=%d]\n",(int)a,REF(a));
  //    printf("deuse atom="); dumpln(a);
  //  }


  if( !IS_EMPTY(a) && !IS_UNBOUND(a) && !IS_STOPPED(a))
  {
    //printf("deuse "); dumpln(a);
    #ifdef DEBUG_RUNTIME_ATOMS
    if( running_compiled_code )
    {
      outter( TEXT("<RUNTIME> deuse"), -1 );
      dump_atom_address( a );
      dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
    #endif
    #ifdef DEBUG_COMPILETIME_ATOMS
    if( compiling_code )
    {
      outter( TEXT("<COMPILETIME> deuse"), -1 );
      dump_atom_address( a );
      dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
    #endif

    //if( IS_ERROR(a) )
    //{
    //printf("TO BE DEUSE hex(a)=%x id=%d REF=%d a=",(int)a,ID(a),REF(a)); dumpln(a);
    //}

    #ifdef SAFEMODE
    assert( a );
    //if(ID(a)>=MAX_ID) {printf("ASSERT[%x]\n",a);}
    assert( (ID(a)<MAX_ID) );
    assert( REF(a)>0 );
    #endif // SAFEMODE
    //printf("GOODY hex(a)=%x id=%d REF=%d\n\n",(int)a,ID(a),REF(a));

    if( !--REF(a) )
    {
      #ifdef DEBUG_ATOM
      printf("<ATOM>  [%08x] ref-1\n",(int)a);
      #endif //DEBUG_ATOM

      #ifdef ADVANCED
      stats[ID(a)].deallocs++;
      stats_free++;
      #endif //ADVANCED

      deleter_t deleter = deleters[ID(a)];
      deleter(a);
    }
  }
  __asm__ volatile ( ASM_RESTORE_RESULT:::ASM_CLOBBER_REGISTERS );
}




//===================================================
/// \fn		void dump_atom( atom_t a, int level )
/// \brief		dumps atom contents
/// \param	a	atom to dump
/// \param	level	level of nesting
///
/// Dumps atom's contents using a call-back \c outter
/// function. Texts are automatically indented
/// according to the \c level. Typically this function
/// is used to print an atom when the call-back
/// function forwards contents to standard output.
/// The call-back function is set by ::init_output().
//===================================================
void dump_atom( atom_t a, int level )
{
  typedef void(*dumper_t)(atom_t,int);

  static dumper_t dumpers[MAX_ID] = {
    dump_integer,	// INTEGER_ID
    dump_float,		// FLOAT_ID
    dump_list,		// LIST_ID
    dump_word,		// WORD_ID
    dump_word,		// SUBWORD_ID
    dump_error,		// ERROR_ID
    dump_var,		// VAR_ID
    dump_mem,		// MEM_ID
  }; ///< array of deleter functions for each atom type


  #ifdef SAFEMODE
  assert( a );
  //assert( outter );
  if( IS_NOT_EMPTY(a) ) assert( ID(a)<MAX_ID );
  #endif //SAFEMODE

  #ifdef DEBUG_REF_COUNT
  if( IS_EMPTY(a) )
  {
    outter( TEXT("##"), 2 );
  }
  else
  {
    int n;
    #define DUMP_BUF_SIZE 64
    char_t buf[DUMP_BUF_SIZE];
    n = SPRINTF( buf, DUMP_BUF_SIZE, TEXT(" %d#"), REF(a) );
    outter( buf, n );
  }
  #endif

  dumper_t dumper = dumpers[ID(a)];
  dumper(a,level);
}




//===================================================
/// \fn		void dump( atom_t a )
/// \brief		dumps atom contents
/// \param	a	atom to dump
///
/// Dumps atom's contents throught the current
/// text output function (e.g. ::use_stdout).
//===================================================
void dump( atom_t a )
{
  dump_atom( a, 0 );
}




//===================================================
/// \fn		void dumpln( atom_t a )
/// \brief		dumps atom contents
/// \param	a	atom to dump
///
/// Dumps atom's contents throught the current
/// text output function (e.g. ::use_stdout)
/// and moves cursor to the next line.
//===================================================
void dumpln( atom_t a )
{
  dump_atom( a, 0 );
  outter( TEXT("\n"), 1 );
}




//===================================================
/// \fn		void init_output( outter_t new_outter )
/// \brief		initializes text output
/// \param  new_outter	new outter function to use by ::dump and ::dumpln
///
/// Initializes the text output system by setting the
/// ::std_outter function to output text throught it.
//===================================================
void init_output( outter_t new_outter )
{
  std_outter = new_outter;
}




//===================================================
/// \fn		void init_input( inner_t new_inner, inner_eof_t new_inner_eof )
/// \brief		initializes text input
/// \param  new_inner	new inner function
/// \param  new_inner_eof	new inner_eof function
///
/// Initializes the text input system by setting the
/// ::std_inner and :std_inner_eof functions to input text
/// throught it.
//===================================================
void init_input( inner_t new_inner, inner_eof_t new_inner_eof )
{
  std_inner = new_inner;
  std_inner_eof = new_inner_eof;
}



//===================================================
/// \fn		void outter( chars_t string, int len )
/// \brief		prints text
/// \param	string	text to print
/// \param	len	length of text
///
/// This function prints text to either the hooked
/// output or to a file.
//===================================================
void outter( chars_t string, int len )
{
  if( output_stream==NULL )
  {
    std_outter( string, len );
  }
  else
  {
    if( len==-1 ) len = STRLEN( string );
    for( ; len>0; len--, string++ )
    {
      int crlf = (DEBAR(*string)==0x0D) && (DEBAR(*(string+1))==0x0A);
      if( !crlf ) 
      {
        char_t wc[2];
        wc[0] = DEBAR(*string);
        wc[1] = 0;

        char* buf =(char*) UTF16_to_UTF8(wc);
        DEALLOC( buf );
        fprintf( output_stream, "%S", wc );
      }
    }
  }
  
  if( dribble_handle )
  {
    if( len==-1 ) len = STRLEN( string );
    for( ; len>0; len--, string++ )
    {
      int crlf = (DEBAR(*string)==0x0D) && (DEBAR(*(string+1))==0x0A);
      if( !crlf ) 
      {
        char_t wc[2];
        wc[0] = DEBAR(*string);
        wc[1] = 0;

        char* buf =(char*) UTF16_to_UTF8(wc);
        DEALLOC( buf );
        fprintf( dribble_handle, "%S", wc );
      }
    }
  }

}


//===================================================
/// \fn		void inner_eof(  )
/// \brief		returns eof status
///
/// Return EOF status of a hooked or file stream.
//===================================================
int inner_eof( )
{
  if( input_stream==NULL )
  {
    return std_inner_eof();
  }
  else
  {
    return feof(input_stream)?1:0;
  }
}


//===================================================
/// \fn		  char_t use_stdin( )
/// \brief	reads a character
///
/// Read a character from the hooked of file stream,
/// which must be ASCII.
//===================================================
char_t inner( )
{
  if( input_stream==NULL )
  {
    char_t ch = std_inner();
    
    if( dribble_handle )
    {
      fprintf( dribble_handle, "%C", (wint_t)ch );
    }
    
    return ch;
  }
  else
  {
    char_t ch;
    //ch = GETCHAR( input_stream );
    //while( ch=='\r' ) ch = GETCHAR( input_stream ); // remove ^M from input, keep ^J

    ch = (char_t)getc( input_stream );
    while( ch=='\r' ) ch = (char_t)getc( input_stream ); // remove ^M from input, keep ^J
    return ch;
  }
}
