//
// Project:	Lhogho
// File:	atoms.h
//
// Copyright (C) 2007 P.Boytchev
//
// Revision history:
//   2006-09-28	- file created
//   2006-09-30	- removed things supported by pools
//   2006-10-01 - added WORD_SIZE
//   2006-10-07 - introduced *_t types
//   2006-10-10 - read_word() and write_word()
//   2006-10-11 - UNICODE macro
//		- export of parse_data_list()
//   2006-10-13 - append()
//   2006-10-17 - errpos/errchar to parse_data_list()
//		  SAFEMODE for parse()
//   2006-10-26 - IS_ERROR(), ERRPOS(), ERRSRC(), ERRCODE()
//		- parse() return error atom
//		- export of parse()
//   2006-10-27 - FLAG_*
//   2006-10-28 - bytes_t
//   2007-02-15 - removing EXPR and LISP types of lists
//   2007-02-27 - DESCR1(), DESCR2()
//		- export of new_var()
//		- module renamed to ATOMS
//   2007-03-01 - NULL
//   2007-03-18 - NEXT(),PREV(),DATA()
//		- equal()
//   2007-03-20 - FLAG_PRIORITY_MUL/ADD/LOG/CMP
//   2007-05-17 - added license info
//   2007-05-22	- Added doxygen-friendly documentation
//   2007-05-27 - behead()
//   2007-05-31 - use_stdout()
//		- hook
//   2007-06-05 - definitions spread to where they belong
//   2007-06-09 - word-related stuff moved to words module
//		- list-related stuff moved to words module
//		- number-related stuff moved to words module
//   2007-06-19 - added mem atoms
//   2009-06-02 - added use_stdin()
//   2010-06-25 - added inner_eof
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#ifndef LHOGHO_ATOMS_H
#define LHOGHO_ATOMS_H

#include "stdint.h"
#include "stddef.h"
#include "stdio.h"
#include "globals.h"




//===================================================
/// \struct		node atoms.h
/// \brief		core atom structure
///
/// This structure represents the core structure of
/// atoms. It makes various projection in order to
/// treat atom's contents as arrays of various types.
//===================================================
typedef struct node {
  union {
    byte_t    b[16];		///< atom structure as 16 bytes
    uint_t    u[4];		///< atom structure as 4 unsigned ints
    int_t     n[4];		///< atom structure as 4 ints
    ushort_t  s[8];		///< atom structure as 8 unsigned shorts
    chars_t   c[4];		///< atom structure as 16 or 8 chars
    struct node* a[4];		///< atom structure as 4 atoms
    int64_t   i[2];		///< atom structure as 2 int64
    float64_t f[2];		///< atom structure as 2 float64
    void*     v[4];		///< atom structure as 4 pointers
  } unode;			///< union of various views of atom's structure
} atomrec_t;			///< core structure of an atom

typedef atomrec_t* atom_t;	///< global atom type

typedef struct node_ex {
  union {
    struct node* a[5];		///< atom structure as 5 atoms
  } unode_ex;			///< union of various views of atom's structure
} atomrec_ex_t;			///< core structure of an atom

typedef atomrec_ex_t* atom_ex_t;	///< global atom type



//==============================================
/// \name	Atom IDs
///
/// These constants represent atom ID. \c MIN_ID
/// and \c MAX_ID are not real atom IDs.
//==============================================

#define MIN_ID		INTEGER_ID
#define INTEGER_ID	0
#define FLOAT_ID	1
#define LIST_ID		2
#define WORD_ID		3
#define SUBWORD_ID	4
#define ERROR_ID	5
#define VAR_ID		6
#define MEM_ID		7
#define MAX_ID		(MEM_ID+1)




//==============================================
/// \name	Common atoms
//==============================================
#define OFFSET_ID		4
#define OFFSET_INT		8
#define OFFSET_REPCOUNT		8
#define OFFSET_REPLIMIT		12

#define REF(x)			((x)->unode.u[0])
#define ID(x)			((x)->unode.b[OFFSET_ID])
#define FLAGS(x)		((x)->unode.s[3])

#define GET_FLAGS(a,bitmask)	(FLAGS(a) &   (bitmask))
#define SET_FLAGS(a,bitmask)	 FLAGS(a) |=  (bitmask)
#define CLR_FLAGS(a,bitmask)	 FLAGS(a) &= ~(bitmask)

//#define DEUSE(x)	{printf("deuse  %x  %s:%d %s()\n",(int)x,__FILE__,__LINE__,__FUNCTION__); deuse(x); }

#define DEUSE(x)	deuse(x)
#define USE(x)          use(x)


//==============================================
/// \name	Sizes
//==============================================
// @{
#define ATOM_SIZE	sizeof(atomrec_t)	///< atom size (in bytes)
#define ATOM_SIZE_EX	(ATOM_SIZE+1*sizeof(int_t)) ///< size of extended atom
#define CHAR_SIZE	sizeof(char_t)		///< char size (in bytes)
#define UNKNOWN		-1
// @}




#ifdef ADVANCED
  extern int  stats_free;
#endif //ADVANCED




extern void   init_atoms();
extern void   finit_atoms();
extern atom_t use( atom_t a );
extern void   __attribute__((used,noinline,regparm(0),stdcall)) deuse( atom_t a );

extern FILE*  input_stream;
extern FILE*  output_stream;
extern int		outter_size;
extern void   outter( chars_t string, int len );
extern char_t inner( );
extern int    inner_eof( );
extern FILE*  dribble_handle;

extern struct lconv *locale_info;
extern void   dump( atom_t a );
extern void   dumpln( atom_t a );
extern void   dump_atom( atom_t a, int level );
extern void   init_output( outter_t new_outter );
extern void   init_input( inner_t new_inner, inner_eof_t new_inner_eof );

#endif //LHOGHO_ATOMS_H
