//
// Project:	Lhogho
// File:	errors.c
//
// Copyright (C) 2007 P.Boytchev
//
// Revision history:
//   2007-06-01	- file created
//   2007-06-03 - errors renamed
//		- TR_ERROR_CROWDED_SOURCES, TR_ERROR_OS_ERROR
//		- ERROR_INCOMPATIBLE_REDEFINITION
//   2007-06-05 - definitions spread to where they belong
//   2007-06-09 - added dump_error()
//   2007-06-26 - ERROR_NOT_A_NUMBER
//   2007-09-09 - ERROR_DO_NOT_KNOW
//		- ERROR_UNKNOWN_VAR
//   2007-09-17 - ERROR_BOOLEAN_EXPECTED
//   2007-09-21 - ERROR_NOT_AN_INTEGER, ERROR_NOT_A_LIST
//   2007-10-06 - ERROR_NOT_A_WORD
//   2007-10-13 - ERROR_INCOMPATIBLE_DATA
//   2007-10-16 - Fixed bug #1814277 Memory leak in LAST [ ]
//		- Fixed bug #1814278 Memory leak in BL [ ]
//		- Fixed bug #1814439 No error position in some cases
//   2007-10-17 - ERROR_TOO_BIG_NUMBER, ERROR_TOO_SMALL_NUMBER
//   2007-11-12 - EXIT_BY_BYE, EXIT_BY_THROW_TOPLEVEL,
//		  EXIT_BY_THROW_SYSTEM, EXIT_BY_THROW_ERROR
//   2007-11-15 - EXIT_BY_THROW_TAG, EXIT_BY_THROW_TAG_VALUE
//   2007-12-04 - ERROR_NOT_A_USER_FUNCTION
//   2008-01-05 - ERROR_VAR_HAS_NO_VALUE
//   2008-09-11 - ERROR_NOT_A_TYPE_NAME
//   2008-09-13 - ERROR_BAD_PROTOTYPE
//   2008-11-22 - ERROR_NOT_A_MEM
//   2009-08-29 - Fixed bug #2845099 Memory leak in THROW
//		- Fixed bug #2845579 DEBUG_MEMORY_LEAKS finds two bad tests
//		- Fixed bug #2846281 Errors not (de)used correctly
//   2009-08-31 - Fixed bug #2847389 Memory leak for unknown file
//   2011-10-07 - Removed TR_WARNING_UNSUPPORTED_COMPILER, because GCC 4+ is now supported
//   2011-12-26 - Error messages taken from string, stored to array of atoms
//   2012-01-20 - Added new_os_error_atom()
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>	// getcwd()

#include "atoms.h"
#include "words.h"
#include "lists.h"
#include "translate.h"
#include "unicode.h"
#include "pools.h"
#include "errors.h"
#include "vars.h"
#include "compiler.h"
#include "numbers.h"
#include "parser.h"

atom_t error_texts[LAST_ERROR_CODE+1]; ///< Texts for error messages




int last_os_error; ///< records last OS error code
atom_t all_errors = NULL; ///< list of all occurred errors



//===================================================
/// \fn    void init_errors( )
/// \brief  initializes errors
///
/// Initializes the table with error messages taken
/// from ::TR_ERRORS.
//===================================================
void init_errors( )
{
  error_texts[0] = unbound;
  
  atom_t names = new_word( TR_ERRORS, UNKNOWN );
  atom_t tokens = tokenize( names, TOKENIZE_DATA );
  atom_t t = tokens; 

  // create error texts
  int i;
  for( i = 1; i<=LAST_ERROR_CODE; i++, t=CDR(t) )
    {
      #ifdef SAFEMODE
        assert( IS_NOT_EMPTY(t) ); // too few words in TR_PRIMITIVES
      #endif

      error_texts[i] = USE(CAR(t));
  }

  #ifdef SAFEMODE
    assert( IS_EMPTY(t) ); // too many words in TR_PRIMITIVES
  #endif

  DEUSE( names );
  DEUSE( tokens );
}




//===================================================
/// \fn    void finit_errors( )
/// \brief  finalizes errors
///
/// Frees atoms holding error texts.
//===================================================
void finit_errors( )
{
  int i;
  for( i = 1; i<=LAST_ERROR_CODE; i++ )
    {
      DEUSE(error_texts[i]);
  }
}




//===================================================
/// \fn 	atom_t new_error( uint_t code, atom_t data )
/// \brief		creates a new error atom
/// \param	code	error code
/// \param	data    error data
/// \return		error atom
///
/// Creates an error atom with reference count 1.
/// The reference count of \c data is increased.
/// Initializes the error position to be a list
/// containing one element, which is the \c data
/// parameter itself.
//===================================================
atom_t new_error( uint_t code, atom_t data )
{
  //printf("-----------\n");
  //printf("new error code=%d\n",code);
  //printf("new error data="); dumpln(data);
  //printf("-----------\n");

  atom_t a = take_from_pool( &data_pool );

  #ifdef SAFEMODE
    assert( data );
  #endif //SAFEMODE

  REF(a) = 1;
  ID(a)  = ERROR_ID;
  ERRCODE(a)  = code;
  ERRPOS(a)   = new_list( USE(data), empty_list );
  ERRDATA(a)  = USE(data);

  #ifdef DEBUG_ATOM
    printf("<ATOM>  [%08x] error=[code=%d]\n",(int)a,code);
  #endif //DEBUG_ATOM

  #ifdef ADVANCED
    stats[ID(a)].allocs++;
    if( stats[ID(a)].max<(stats[ID(a)].allocs-stats[ID(a)].deallocs) )
      stats[ID(a)].max=(stats[ID(a)].allocs-stats[ID(a)].deallocs);
    stats_free--;
  #endif //ADVANCED

  #ifdef DEBUG_RUNTIME_ATOMS
  if( running_compiled_code )
    {
      outter( TEXT("<RUNTIME> new  "), -1 );
      dump_atom_address( a );
      dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
  #endif
  #ifdef DEBUG_COMPILETIME_ATOMS
  if( compiling_code )
    {
      outter( TEXT("<COMPILETIME> new  "), -1 );
      dump_atom_address( a );
      dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
  #endif

  if( all_errors==NULL ) all_errors = empty_list;
  all_errors = new_list( a, all_errors );

  return a;
}



//===================================================
/// \fn		atom_t new_os_error( chars_t filename )
/// \brief		creates error atom for OS error
/// \param filename	file name associated with the error
/// \return		error atom
///
/// Creates error atom describing OS error.
//===================================================
atom_t new_os_error( chars_t filename )
{
  atom_t wrd = new_word( filename, UNKNOWN );
  atom_t res =  new_os_error_atom( wrd );
  DEUSE( wrd );
  return res;
  //return new_error( ERROR_OS_ERROR, wrd );
}
atom_t new_os_error_atom( atom_t filename )
{
  last_os_error = errno;
  return new_error( ERROR_OS_ERROR, filename );
}




//===================================================
/// \fn		atom_t new_parse_error( uint_t code, int position, atom_t source )
/// \brief		creates error atom for parser's error
/// \param code		error code
/// \param position	character position within the source
/// \param source	(sub)word containing the source
/// \return		error atom
///
/// Creates error atom describing an error generated
/// by the parser.
//===================================================
atom_t new_parse_error( uint_t code, int position, atom_t source )
{
  atom_t errsrc = new_subword( source, STRING(source)+position, 1 );
  atom_t error = new_error( code, unbound );
  CAR(ERRPOS(error)) = errsrc;
  return error;
}




//===================================================
/// \fn		void add_error_source( atom_t error, atom_t source )
/// \brief		adds new error position to an error
/// \param error	error to which new position will be added
/// \param source	error source
///
/// Adds a new error position to an already existing
/// error atom. The error position is simply an atom
/// which may have been extracted from the actual
/// source. The new error position is inserted in the
/// first place of the list of error positions.
//===================================================
void add_error_source( atom_t error, atom_t source )
{
  //printf("register error pos>>>---------------\n");
  //printf("register error src>>>"); dumpln(source);
  //printf("register error err>>>"); dumpln(error);
  //printf("register error pos>>>---------------\n");
  //printf("errpos old ref=%d\n",REF(ERRPOS(error)));
  ERRPOS(error) = new_list( USE(source), ERRPOS(error) );
  //printf("errpos new ref=%d\n",REF(ERRPOS(error)));
}




//===================================================
/// \fn		void delete_error( atom_t a ) 
/// \brief	deletes error atom
/// \param a	atom to delete
///
/// Deletes error atom by returning it back to the
/// data pool. The source atom is dereferenced.
//===================================================
void delete_error( atom_t a )
{
  //printf("****** DELETING ERROR ****** [%x]\n",(int)a);
  DEUSE( ERRPOS(a) );
  DEUSE( ERRDATA(a) );
  return_to_pool( &data_pool, a );
}




//===================================================
/// \fn		int get_error_position( atom_t data, chars_t* source );
/// \brief		gets error position
/// \param data		data to search for
/// \param source	source where the data is found
/// \return		position of data within found source
///
/// Tries to identify the position of the first
/// (sub)word in the data in the original source.
/// This function returns the character position
/// and the source. If the source cannot be identified,
/// then return -1 and set \c source to NULL.
//===================================================
int get_error_position( atom_t data, chars_t* source )
{
  // dig in a list to find the first (sub)word
  while( IS_LIST(data) & IS_NOT_EMPTY(data) ) data = CAR(data);

  // if it is a word then return we have found something
  if( IS_WORD(data) )
    {
      *source = STRING(data);
      return 0;
    }

  // if it is a subword then return we have found something
  if( IS_SUBWORD(data) )
    {
      *source = STRING(WORD(data));
      return STRING(data)-STRING(WORD(data));
    }

  // no, could not find the error position
  *source = NULL;
  return -1;
}



//===================================================
/// \fn		void dump_neighbourhood( chars_t source, int position )
/// \brief		dumps source around some position
/// \param source	source containing the neighbourhood
/// \param position	position marking the neighbourhood
///
/// Dumps a piece of source code containing given
/// position. The dumped neighbourhood starts before
/// the position and end after it in a way that:
///	(1) at most LEFT_DIST chars before the position
///	    are included
///	(2) at most RIGHT_DIST chars after the position
///	    are included
///	(3) the neightbourhood contains characters from
///	    a single line only
/// Dumped neighbourhood is output as a single line
/// with an arrow anove it pointing the exact position.
//===================================================
#define LEFT_DIST 10
#define RIGHT_DIST 40
void dump_neighbourhood( chars_t source, int position )
{
  if( !source || position<0 ) return;

  int left;
  int right;
  int ellipses = 1;
  for( left=0; left<LEFT_DIST; left++ )
    {
      // reached the beginning of the source?
      if( position-left == 0 ) {ellipses=0; break;}

      // reached the beginning of the line?
      char_t ch = *(source+(position-left-1));
      if( ch==TEXT('\n') || ch==TEXT('\r') ) {ellipses=0; break;}
    }
  while( left && ((*(source+position-left))<=TEXT(' ')) ) left--;

  for( right=0; right<RIGHT_DIST; right++ )
    {
      // reached the end of the line/file?
      char_t ch = *(source+(position+right+1));
      if( ch=='\n' || ch=='\r' || ch=='\0') break;
    }

  int i;
  chars_t s;
  if( ellipses ) outter( TEXT("..."), 3 );
  for( i=0,s=source+position-left; i<left+right+1; i++,s++ ) outter( s, 1 );
  if( *(source+position+right+1)>=TEXT(' ') )
    outter( TEXT("..."), 3 );
  outter( TEXT("\n"), -1 );
  for( i=0; i<left+3*ellipses; i++ ) outter( TEXT(" "), 1 );
  outter( TEXT("^"), -1 );
}








//===================================================
/// \fn		void dump_error( atom_t a, int level ) 
/// \brief	dumps error atom
/// \param a	atom to dump
/// \param level dump level
///
/// Dumps error atom through the current outter function.
//===================================================
void dump_error( atom_t a, int level )
{
  #define DUMP_BUF_SIZE 128
  char_t buf[DUMP_BUF_SIZE];
  int n;

  //printf("<<err ref=%d>> ",REF(a));
  //printf("temporary dump of an error\n");
  //printf(">>>>>>error ref  = %d\n",REF(a));
  //printf(">>>>>>error code = %d\n",ERRCODE(a));
  //printf(">>>>>>error posn = "); dumpln(ERRPOS(a));
  //printf(">>>>>>error data = "); dumpln(ERRDATA(a));
  //return;

  int errpos = -1;
  int errpos2 = -1;
  int errpos3 = -1;
  chars_t errsrc = NULL;
  chars_t errsrc2 = NULL;
  chars_t errsrc3 = NULL;
  atom_t p;
  //printf("errpos="); dump_atom(ERRPOS(a),1); printf("\n------------\n");
  for( p=ERRPOS(a); IS_NOT_EMPTY(p); p=CDR(p) )
    {
      //printf("ERRPOS=");dumpln(CAR(p));
      chars_t src;
      int pos = get_error_position( CAR(p), &src );
      if( pos>-1 )
	{
	  errpos3 = errpos2;
	  errsrc3 = errsrc2;
	  errpos2 = errpos;
	  errsrc2 = errsrc;
	  errpos = pos;
	  errsrc = src;
	}
      //printf("pos=%d source=",pos); dumpln(CAR(p));
    }

  if( ERRCODE(a)==EXIT_BY_THROW_USER_ERROR )
    {
      errpos = errpos3;
      errsrc = errsrc3;
    }

  // print error code
  n = SPRINTF( buf, DUMP_BUF_SIZE, FORMAT_ERR_CODE, ERRCODE(a) );
  outter( buf, n );

  // print error position
  n = SPRINTF( buf, DUMP_BUF_SIZE, FORMAT_ERR_POS, errpos );
  outter( buf, n );

  // print error message
  if( ERRCODE(a)==EXIT_BY_THROW_USER_ERROR )
    {
      outter( TEXT(" - "), 3 );
      dump( ERRDATA(a) );
    }
  else
    {
      #ifdef SAFEMODE
        assert( ERRCODE(a)>0 );
        assert( ERRCODE(a)<=LAST_ERROR_CODE );
      #endif
      outter( TEXT(" - "), 3 );
      dump( error_texts[ERRCODE(a)] );
    }

  if( ERRCODE(a)==ERROR_OS_ERROR )
    {
      chars_t msg = UNFILENAME(strerror(last_os_error));
      outter( TEXT(": "), 2 );
      outter( msg, -1 );
      DEALLOC( msg );
    }
  outter( TEXT("\n"), -1 );

  // print error source
  dump_neighbourhood( errsrc, errpos );
}




//===================================================
/// \fn		void clear_all_errors( ) 
/// \brief	clears all errors
///
/// Clears all errors recorder in ::all_errors except
/// for the one stored in last_error
//===================================================
void clear_all_errors(  )
{
  atom_t err = all_errors;
  if( err==NULL ) return;

  //printf("===INSIDE CLEAR_ALL_ERRORS===\n");
  //printf("===EXCEPTION(%d):  ",REF(last_error)); dumpln(last_error);
  //printf("===ALL ERRORS (BEFORE): "); dumpln(all_errors);

  //printf("===START ERROR SCANNING===\n");
  while( IS_NOT_EMPTY(err) )
    {
      //printf("===ERROR: [%x]\n",CAR(err));
      if( CAR(err)!=last_error ) REF(CAR(err)) = 1;
      err = CDR(err);
    }
  //printf("===END ERROR SCANNING===\n");

  USE( last_error );
  //printf("AEREF=%d !!!!!!!!!!!!!!!!\n",REF(all_errors));
  DEUSE( all_errors );
  //printf("AEREF=%d !!!!!!!!!!!!!!!!\n",REF(all_errors));
  DEUSE( last_error );

  all_errors = empty_list;
  //printf("===END OF CLEAR_ALL_ERRORS===\n\n");
}
