//
// Project:	Lhogho
// File:	globals.h
//
// Copyright (C) 2007 P.Boytchev
//
// Revision history:
//   2006-09-28	- file created
//   2006-10-10 - added DEBUG_* macros
//   2006-10-11 - removed ADVANCED macro
//		- added UNICODE macro
//   2006-10-13 - added DEBUG_TOKENS macro
//   2007-03-01 - added DEVELOPMENT macro
//   2007-03-19 - DEBUG_LIST_FLAGS macro
//   2007-03-25 - DEBUG_PARENTHESES macro
//   2007-05-17 - added license info
//   2007-05-22	- added doxygen-friendly documentation
//   2007-06-02 - STATISTICS & DEVELOPMENT merged into ADVANCED
//   2007-06-05 - definitions spread to where they belong
//   2007-07-03 - DEBUG_COMPILE macro
//   2007-07-31 - DEBUG_FIND_VAR macro
//   2007-12-02 - DEBUG_LIST_TOKEN_FLAGS macro
//   2010-06-25 - support for inner_eof
//   2011-02-27 - check for GCC version
//   2012-09-23 - DEBUG_ATOM_LIST macro
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef LHOGHO_GLOBALS_H
#define LHOGHO_GLOBALS_H


#include "stdint.h"
#include "stddef.h"




//===================================================
/// \name Build options
//===================================================
#define SAFEMODE	///< full error checking
#define UNICODE_CHARS	///< use UTF-16LE for internal data
#define ADVANCED	///< include features needed for development




//===================================================
/// \name Debug options
/// Defined debug symbols. When some of them is
/// defined the compilation contains code to dump
/// related functions. To define a symbol remove
/// the \e no prefix from its name */
//===================================================
#define noDEBUG_POOL	///< remove \e no prefix to debug pool functions
#define noDEBUG_HEAP	///< remove \e no prefix to debug heap functions
#define noDEBUG_ATOM	///< remove \e no prefix to debug atom functions
#define noDEBUG_ATOM_LIST	///< remove \e no prefix to debug atom-list functions
#define noDEBUG_VAR		///< debug var functions
#define noDEBUG_TOKENS		///< debug token generation
#define noDEBUG_LIST_FLAGS		///< debug flags of list nodes
#define noDEBUG_LIST_TOKEN_FLAGS	///< debug token flags of list nodes
#define noDEBUG_REF_COUNT	///< dump ref counts for each atom
#define noDEBUG_TOKENIZATION	///< debug grouping tokens in lists
#define noDEBUG_PARENTHESES	///< debug parsing parenthesis
#define noDEBUG_PARSE   	///< debug parsing
#define noDEBUG_TO_END		///< debug TO...END processing
#define noDEBUG_COMPILE		///< debug compilation process
#define noDEBUG_FIND_VAR	///< debug syntax-scope find var
#define noDEBUG_FIND_RUNTIME_VAR	///< debug runtime searching of vars
#define noDEBUG_RUNTIME_ATOMS	///< debug atoms during run time
#define noDEBUG_COMPILETIME_ATOMS ///< debug atoms during compilation time
#define DEBUG_CLEAR_FREED_MEM	///< clear (with 0xFF) all freed atoms
#define noDEBUG_MEMORY_LEAKS	///< dumps all unfreed atoms

#ifdef DEBUG_MEMORY_LEAKS
  #define DEBUG_CLEAR_FREED_MEM
#endif


//===================================================
/// \name Platform defines
/// Platform-specific macros. Currently supported are:
/// 32-bit Intel Pentium for Windows, Linux and Mac OS X
//===================================================
#ifdef PROCESSOR_x86
  #define PROCESSOR_NAME	"i386"	///< Platform name
  //#define ASM_STORE_RESULT	"push %%ebx;  push %%eax; push %%esi;" ///< store result
  //#define ASM_RESTORE_RESULT	"pop %%esi; pop %%eax; pop %%ebx"   ///< restore result
  #define ASM_STORE_RESULT	"pusha" ///< store result
  #define ASM_RESTORE_RESULT	"popa"   ///< restore result
  #define ASM_CLOBBER_REGISTERS "%eax","%ebx","%ecx","%edx","%esi","%edi","memory","cc"
//#define ASM_STATUS		"movl %0,%%ebx"::"m"(status):"%ebx" ///< set OUTPUT flag
//#define ASM_SET_STATUS	"stc" ///< set OUTPUT flag
//#define ASM_CLEAR_STATUS	"clc" ///< clear OUTPUT flag
#endif

#ifdef WINDOWS
  #define OS_NAME	"windows"
#endif

#ifdef __GNUC__
#define GCC_VERSION (__GNUC__ * 10000 \
                   + __GNUC_MINOR__ * 100 \
                   + __GNUC_PATCHLEVEL__)
          
   // GCC versions 4.4 and 4.5 generate crashing Lhogho
   // GCC versions 3.4 and 4.1 generate working Lhogho
   // Thus, set UNSUPPORTED_COMPILER flag for GCC > 4.1.99
   #if GCC_VERSION > 40199 // GCC
      #define UNSUPPORTED_COMPILER
   #endif
#else
   #define UNSUPPORTED_COMPILER
#endif
          
          
#ifdef LINUX
  #define OS_NAME	"linux"
#endif

#ifdef APPLE
  #define OS_NAME	"darwin"
#endif

#ifndef OS_NAME
  #error "Undefined operating system"
#endif

#ifndef PROCESSOR_NAME
  #error "Undefined processor"
#endif


//===================================================
/// \name Lhogho datatypes
//===================================================
// @{
typedef double		float64_t;	// 64-bit fp
typedef float		float32_t;	// 32-bit fp
//typedef long long	int64_t;	// 64-bit int
typedef unsigned char	byte_t;		// 8-bit byte
typedef unsigned short	ushort_t;	// unsigned short
typedef int		int_t;		// integer
typedef unsigned int	uint_t;		// unsigned integer
typedef void*		ptr_t;		// pointer
typedef byte_t*		bytes_t;	// array of bytes
typedef void(*fn)();                    // general function

#ifdef UNICODE_CHARS
  typedef wchar_t	char_t;		// character
  typedef wchar_t*	chars_t;	// string
#else
  typedef char		char_t;		// character
  typedef char*		chars_t;	// string
#endif //UNICODE_CHARS
/// @}

#define Z   printf("--- %s() in %s:%d---\n",__FUNCTION__,__FILE__,__LINE__)

//===================================================
/// \fn		void(*outter_t)(chars_t,int)
/// \brief		Functional-typedef for \c dump/dumpln functions
///
/// A functional type definition used as a pattern
/// for all private atom outters like ::use_stdout.
//===================================================
typedef void(*outter_t)(chars_t,int);



//===================================================
/// \fn		char_t(*inner_t)(void)
/// \brief		Functional-typedef for input functions
///
/// A functional type definition used as a pattern
/// for all inners like ::use_stdin.
//===================================================
typedef char_t(*inner_t)(void);



//===================================================
/// \fn		int(*inner_eof_t)(void)
/// \brief		Functional-typedef for input functions
///
/// A functional type definition used as a pattern
/// for all inner_eofs.
//===================================================
typedef int(*inner_eof_t)(void);




#ifndef NULL
#define NULL 0
#endif


#ifndef __USE_ISOC99
#define __USE_ISOC99
#endif

#endif //LHOGHO_GLOBALS_H
