//
// Project:	Lhogho
// File:	i386.c
//
// Copyright (C) 2007 P.Boytchev
//
// Revision history:
//   2007-06-19	- file created
//   2007-08-05 - asm_result_make()
//   2007-08-08 - asm_reference()
//   2007-08-09 - Fixed bug #1771163 Warning building the dev docs
//   2007-08-12 - Fixed bug #1767825 Bad return value for empty programs
//   2007-08-13 - Support for primitive variables
//              - Fixed bug #1773113 MAKE does not work with primitive variables
//   2007-09-03 - Fixed bug #1787023 ARCTAN and RADARCTAN
//   2007-09-18 - Support for OUTPUT and STOP
//   2007-09-23 - Support for IF
//   2007-10-21 - Support for REPEAT, FOREVER
//   2007-11-04 - Support for WHILE, UNTIL
//   2007-11-09 - Fixed bug #1828215 MAKE does not work with unknown variables
//              - Fixed bug #1828217 ':' does not work with unknown variables
//   2007-12-23 - Support for RUN
//   2008-01-07 - Fixed bug #1856864 Defined but unset variables
//              - Support for RUNRESULT
//   2009-08-03 - Fixed bug #2191139 Errors/exits in internal functions
//   2009-08-16 - Fixed bug #2838617 RUN suppresses STOP
//   2009-08-18 - EBX=1 value from output/stop; EBX=0 value from expression
//   2009-08-19 - Support for MAYBEOUTPUT
//   2011-02-05 - Fixed bug #3173377 Warnings building developer's documentation
//   2011-03-03 - Fixed bug #3197053 DO.UNTIL behaves exactly like UNTIL
//              - Fixed bug #3197052 DO.WHILE behaves exactly like WHILE
//	 2011-10-14 - Support for RUNMACRO
//   2011-12-31 - Fixed bug #3466110 Error in SE's input cases a crash
//   2012-01-02 - Command-line options are translatable
//   2012-02-26 - Fixed bug #3493871 Externals do not capture errors
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include "stdio.h"
#include "atoms.h"
#include "words.h"
#include "vars.h"
#include "lists.h"
#include "numbers.h"
#include "unicode.h"
#include "options.h"
#include "runtime.h"
#include "asm.h"


#define EXIT		(void*)ctx->exit_addr
#define START_IP	(ctx->generate?(int)MEMORY(ctx->generate):0)///< start IP
#define LOCAL_IP	(ctx->size)			///< local IP (0-based)
#define IP		START_IP+LOCAL_IP		///< current IP
#define REL(addr)	relative( ctx, (uint_t)addr )	///< relative address

#define EMIT1(a)		emit_1(ctx,a)
#define EMIT2(a)		emit_2(ctx,a)
#define EMIT3(a,b)		emit_2(ctx,a); emit_1(ctx,b)
#define EMIT4(a)		emit_4(ctx,a)
#define EMIT5(a,b)		emit_1(ctx,a); emit_4(ctx,(uint_t)b)
#define EMIT6(a,b)		emit_2(ctx,a); emit_4(ctx,(uint_t)b)

//===================================================
/// \fn		uint_t relative( context_t* ctx, int addr )
/// \brief	calculates relative offset
/// \param ctx	compilation context
/// \param addr	target address
/// \return	calculated offset
///
/// Calculates a relative offset, which is commonly
/// used in JPM and CALL instructions. The parameter
/// ADDR is absolute target address. The result is
/// the same target, but relative to the current
/// IP position.
//===================================================
uint_t relative( context_t* ctx, int addr )
{
  if( ctx->generate )
    {
      int next_instruction = IP + 4;
      return (uint_t)(addr-next_instruction);
    }
 else
   return 0;
}

//===================================================
/// \fn		void asm_label( context_t* ctx, chars_t label )
/// \brief		disassemble label
/// \param	ctx	context
/// \param	label	label
///
/// Dumps disassembler's label followed by semicolon
//===================================================
void asm_label( context_t* ctx, chars_t label )
{
  #ifdef ADVANCED
  disasm( ctx, 4, label );
#endif //ADVANCED
}

#define PUSH_EAX		{ EMIT1(0x50); INSTR("push  eax"); }
#define PUSH_EDX		{ EMIT1(0x52); INSTR("push  edx"); }
#define PUSH_EBP		{ EMIT1(0x55); INSTR("push  ebp"); }
#define PUSH_ESI		{ EMIT1(0x56); INSTR("push  esi"); }
#define PUSH_EBX		{ EMIT1(0x53); INSTR("push  ebx"); }

#define POP_EAX			{ EMIT1(0x58); INSTR("pop   eax"); }
#define POP_EDX			{ EMIT1(0x5A); INSTR("pop   edx"); }
#define POP_EBP			{ EMIT1(0x5D); INSTR("pop   ebp"); }
#define POP_ESI			{ EMIT1(0x5E); INSTR("pop   esi"); }
#define POP_EBX			{ EMIT1(0x5B); INSTR("pop   ebx"); }

#define PUSH_MEM_EAX(ofs)	{ EMIT6(0xB0FF,ofs); INSTR("push  [eax%p]",ofs); }
#define PUSH_MEM_EBP(ofs)	{ EMIT6(0xB5FF,ofs); INSTR("push  [ebp%p]",ofs); }
#define PUSH_MEM_ESI(ofs)	{ EMIT6(0xB6FF,ofs); INSTR("push  [esi%p]",ofs); }

#define PUSH_MEM(ofs)		{ EMIT6(0x35FF,ofs); INSTR("push  [%d]",ofs); }

#define PUSH_SOURCE(s)		{ EMIT5(0x68,s); INSTR("push  (%l)",s); }
#define PUSH_CONST(n)		{ EMIT5(0x68,n); INSTR("push  %d",n); }
#define PUSH_ATOM(a)		{ EMIT5(0x68,a); INSTR("push  %a",a); }
#define PUSH_ATOMSTR(a,s)	{ EMIT5(0x68,a); INSTR("push  %s",s); }

#define MOV_EAX_ATOM(atom)	{ EMIT5(0xB8,atom); INSTR("mov   eax,%a",atom); }
#define MOV_ESI_CONST(n)	{ EMIT5(0xBE,n);    INSTR("mov   esi,%d",n); }
#define MOV_ESI_CONST(n)	{ EMIT5(0xBE,n);    INSTR("mov   esi,%d",n); }
#define MOV_EBX_CONST(n)	{ EMIT5(0xBB,n);    INSTR("mov   ebx,%d",n); }
     
#define MOV_EAX_EBP		{ EMIT2(0xE889); INSTR("mov   eax,ebp"); }
#define MOV_EBP_ESP		{ EMIT2(0xE589); INSTR("mov   ebp,esp"); }
#define MOV_ESP_EBP		{ EMIT2(0xEC89); INSTR("mov   esp,ebp"); }
#define MOV_ESI_EBP		{ EMIT2(0xEE89); INSTR("mov   esi,ebp"); }

#define MOV_EAX_MEM_ESP		{ EMIT3(0x048B,0x24); INSTR("mov   eax,[esp]"); }
#define MOV_EAX_MEM_EBP(ofs)	{ EMIT3(0x458B,ofs);  INSTR("mov   eax,[ebp%p]",ofs); }
#define MOV_EAX_MEM_EAX(ofs)	{ EMIT3(0x408B,ofs);  INSTR("mov   eax,[eax%p]",ofs); }
#define MOV_EAX_MEM_OFS(ofs)    { EMIT6(0x058B,ofs);  INSTR("mov   eax,[%d]",ofs); }
#define MOV_EBP_MEM_OFS(ofs)    { EMIT6(0x2D8B,ofs);  INSTR("mov   ebp,[%d]",ofs); }
#define MOV_EBX_MEM_OFS(ofs)    { EMIT6(0x1D8B,ofs);  INSTR("mov   ebp,[%d]",ofs); }

#define MOV_MEM_EBP_EAX(ofs)	{ EMIT6(0x8589,ofs); INSTR("mov   [ebp%p],eax",ofs); }
#define MOV_MEM_ESI_EAX(ofs)	{ EMIT6(0x8689,ofs); INSTR("mov   [esi%p],eax",ofs); }
#define MOV_MEM_OFS_EAX(ofs)    { EMIT6(0x0589,ofs); INSTR("mov   [%d],eax",ofs); }
#define MOV_MEM_OFS_EBP(ofs)    { EMIT6(0x2D89,ofs); INSTR("mov   [%d],ebp",ofs); }
#define MOV_MEM_OFS_CONST(ofs,n) {EMIT6(0x05C7,ofs); EMIT4(n); \
				  INSTR("mov   [%d],%a", ofs, n); }

#define MOV_ESI_SSMEM_ESI_4	{ EMIT4(0xFC768B36); INSTR("mov   esi,ss:[esi-4]"); }
#define MOV_ESI_SSMEM_EBP_4	{ EMIT4(0xFC758B36); INSTR("mov   esi,ss:[ebp-4]"); }

#define INC_EAX_MEM(n)		{ EMIT3(0x40FF,n); INSTR("inc   [eax%p]",n); }
#define DEC_EAX_MEM(n)		{ EMIT3(0x48FF,n); INSTR("dec   [eax%p]",n); }

#define RET			{ EMIT1(0xC3); INSTR("ret"); }

#define ADD_EAX_CONST_4(n)	{ EMIT5(0x05,n);   INSTR("add   eax,%d",n); }
#define ADD_ESI_CONST_4(n)      { EMIT6(0xC681,n); INSTR("add   esi,%d",n); }

#define CMP_EBX_CONST(n)	{ EMIT3(0xFB83,n); INSTR("cmp   ebx,%d",n); }
#define CMP_EAX_CONST(n)	{ EMIT5(0x3D,n); INSTR("cmp   eax,%d",n); }
#define CMP_EAX_ATOM(a)		{ EMIT5(0x3D,a); INSTR("cmp   eax,%l",a); }
#define CMP_ID_ERROR		{ EMIT2(0x7880); EMIT1(OFFSET_ID); EMIT1(ERROR_ID);\
				  INSTR("cmp   [eax+ID],ERROR_ID"); }
#define CMP_MEM_EBP(ofs,val)	{ EMIT2(0x7D80); EMIT1(ofs); EMIT1(val);\
				  INSTR("cmp   [eax%p],%d",ofs,val); }
#define CMP_MEM_OFS_CONST_4(ofs,n) { EMIT6(0x3D81,ofs); EMIT4(n); INSTR("cmp   [%d],%a", ofs, n); }
#define CMP_MEM_OFS_CONST_1(ofs,n) { EMIT6(0x3D80,ofs); EMIT1(n); \
				     INSTR("cmp   [%d],%d", ofs, n); }

//#define JC_EXIT			{ EMIT6(0x820F,REL(EXIT)); INSTR("jc    exit"); }
#define JE_EXIT			{ EMIT6(0x840F,REL(EXIT)); INSTR("je    exit"); }
#define JE(addr,name)		{ EMIT6(0x840F,REL(addr)); INSTR("je    %s",name); }
#define JNE(addr,name)		{ EMIT6(0x850F,REL(addr)); INSTR("jne   %s",name); }
#define JNZ(addr,name)		{ EMIT6(0x850F,REL(addr)); INSTR("jnz   %s",name); }

#define JMP_EXIT		{ EMIT5(0xE9,REL(EXIT)); INSTR("jmp   exit"); }
#define JMP(addr,name)		{ EMIT5(0xE9,REL(addr)); INSTR("jmp   %s",name); }
#define JMP_ATOM(addr,name)	{ EMIT5(0xE9,REL(addr)); INSTR("jmp   %a",name); }
#define JMP_EAX(name)		{ EMIT2(0xE0FF); INSTR("jmp   %a",name); }

#define CALL(addr,name)		{ EMIT5(0xE8,REL(addr)); INSTR("call  %s",name); }
#define CALL_ATOM(addr,name)	{ EMIT5(0xE8,REL(addr)); INSTR("call  %a",name); }

#define CALL_MEM(addr,name)	{ EMIT6(0x15FF,addr); INSTR("call  [%a]",name); }
#define CALL_MEM_EAX(addr,name) { EMIT6(0x90FF,addr); INSTR("call  [eax+%s]",name); }

#define FSTPF_MEM_ESP		{ EMIT2(0x1CD9); EMIT1(0x24); INSTR("fstpf [esp]"); } // FSTP for float
#define FSTPD_MEM_ESP		{ EMIT2(0x1CDD); EMIT1(0x24); INSTR("fstpd [esp]"); } // FSTP for double

#define INT_3                   { EMIT1(0xCC); INSTR("int   3"); }
#define NOP                     { EMIT1(0x90); INSTR("nop"); }
//#define STC			{ EMIT1(0xF9); INSTR("stc"); }
//#define CLC			{ EMIT1(0xF8); INSTR("clc"); }

//===================================================
/// \fn		void asm_prologue( context_t* ctx, atom_t func, int check )
/// \brief	generates a function prologue
/// \param ctx	compilation context
/// \param func	function which prologue to generate
/// \param check add code for checking inputs
///
/// Generates the prologue of a function. This is an
/// initializating sequences of machine instructions
/// needed to setup stack, system data and local vars.
///
/// Generated code:
///
///	;; initialize stack
///	push	ebp		; caller's frame [EBP+0]
///	mov	ebp,esp		; EBP=current frame
///	push	esi		; parent's frame [EBP-4]
///	push	func		; parent atom [EBP-8]
///
///	;; initialize system data
///	push	repeat_chain	; for root; unbound for non-root
///	call	use
///	push	2		; result of IFTRUE [EBP-16]
///	mov	eax,unbound	; default result of the function
///
///	;; initialize local variables
///	push	empty_list	; runtime locals [EBP-20]
///	push	eax		; once for each local variable
///
/// ;; check all input params   ; if check!=0
/// push ebp
/// call rt_check_inputs
/// CMP_ID_ERROR
/// JE_EXIT;

//===================================================
void asm_prologue( context_t* ctx, atom_t func, int check )
{
  atom_t a;

  INFO( "" );
  INFO( "=========================" );
  INFO( " FUNCTION %a", NAME(func) );
  INFO( "=========================" );
  INFO( "" );

  LABEL( "%a:", NAME(func) );

  PUSH_EBP;	REM( "function prologue" );
  MOV_EBP_ESP;
  PUSH_ESI;
  PUSH_ATOM( func ); REM( "parent atom" );

  INFO( "" );
  INFO( "initialize system data" );
  if( func==root )
    {
      MOV_MEM_OFS_EBP( &root_frame ); REM( "store root frame" );
      PUSH_ATOMSTR( repeat_chain, TEXT("repeat_chain") );
      CALL( use, TEXT("use") );
    }
  else
    {
      PUSH_ATOM( unbound );
    }

  PUSH_CONST(2); REM( "IFTRUE status" );
  MOV_EAX_ATOM(unbound); REM( "default result of %a", NAME(func) );

  INFO( "" );
  INFO( "initialize local variables" );
  PUSH_ATOM( empty_list ); REM( "runtime locals [EBP-16]" );

  int ofs = BASE_OFFSET_LOCALS-4;
  for( a=LOCALS(func); IS_NOT_EMPTY(a); a=CDR(a) )
    if( IS_VARIABLE(CAR(a))&& IS_NORMAL(CAR(a)) )
      {
        // print the name and the address of the local var
        PUSH_EAX; REM( "local %a [EBP%p]", NAME(CAR(a)), ofs );
        ofs-=4;
      }

  // if we are not making the topmost prologue
  // and if we are not making the prologue of RUN
  // then add code to check all inputs
  if( check && func!=root && NAME(func)!=word_to )
  {
//printf("func anme="); dumpln(NAME(func)); //@@
    INFO( "check all inputs" );
    PUSH_EBP;	REM( "give local frame to checker" );
    CALL( rt_check_inputs, TEXT("rt_check_inputs") );
    CMP_ID_ERROR
    JE_EXIT;
  }
}

//===================================================
/// \fn		void asm_epilogue( context_t* ctx, atom_t func, int is_macro )
/// \brief	generates a function epilogue
/// \param ctx	compilation context
/// \param func	function which epilogue to generate
/// \param is_macro if set then local variables must be copies
///
/// Generates the epilogue of a function. This is a
/// finalizating sequences of machine instructions
/// needed to recover stack and local variables.
///
/// Generated code:
///
///     mov     ebx,0
/// exit:
///
/// push	ebp				; push current frame only if is_macro is set
/// call    copy_local_vars	; only if is_macro is set
///
///	;; free local variables
///	call	deuse	; once for each local
///	call	deuse	; for runtime locals
///
///	;; free system data
///	pop	edx	; free result of IFTRUE
///	call	deuse	; free repeat chain
///
///	;; recover frame pointers
///	pop	edx	; free parent atom
///	pop	esi   
///	mov	esp,ebp
///	pop	ebp
///	ret
//===================================================
void asm_preepilogue( context_t* ctx )
{
  MOV_EBX_CONST(0);
}
void asm_epilogue( context_t* ctx, atom_t func, int is_macro )
{
 atom_t a;
 
  INFO( "" );
LABEL( "exit:" );

  if( is_macro==COMPILE_AS_MACRO )
  {
    //INT_3; NOP; NOP; NOP;
	PUSH_EAX;
	PUSH_EBX;
	PUSH_EBP; REM("push current frame");
	CALL( copy_local_vars, TEXT("copy_local_vars") );
	POP_EDX;
	POP_EBX;
	POP_EAX;
	//NOP; NOP; NOP;
  }

  INFO( "free local variables" );
  for( a=LOCALS(func); IS_NOT_EMPTY(a); a=CDR(a) )
    if( IS_VARIABLE(CAR(a)) && IS_NORMAL(CAR(a)) )
      CALL( deuse, TEXT("deuse") );
  CALL( deuse, TEXT("deuse") ); REM("also free dynamic variables");

  INFO( "" );
  INFO( "free system data" );
  POP_EDX;			REM( "free result of IFTRUE" );
  CALL( deuse, TEXT("deuse") ); REM( "free repeat chain" );

  INFO( "" );
  INFO( "recover frame pointers" );
  POP_EDX;			REM( "delete link to parent" );
  POP_ESI;
  MOV_ESP_EBP;
  POP_EBP;
  RET;				REM( "end of %a", NAME(func) );
}


//===================================================
/// \fn		void asm_push_atom( context_t* ctx, atom_t atom )
/// \brief	genarates a code to push atom
/// \param ctx	compilation context
/// \param atom	atom to push
///
/// Generates machine code which pushes atom
/// in the stack and increases its reference
/// count. The atom reference count is NOT
/// increased if it is UNBOUND or EMPTY LIST.
///
/// \note
///	Assumes the reference count is at the
///	beginning of the atom - i.e. at offset 0.
///
/// Generated code:
///
///	push	{atom}
///	call	{use}	; if not UNBOUND or EMPTY LIST
///
//===================================================
void asm_push_atom( context_t* ctx, atom_t atom )
{
  PUSH_ATOM( atom );
  if( !IS_UNBOUND(atom) && !IS_EMPTY(atom) )
    {
      CALL( use, TEXT("use") );
    }
}

//===================================================
/// \fn		void asm_pop_atom( context_t* ctx )
/// \brief	genarates a code to pop atom
/// \param ctx	compilation context
///
/// Generates machine code which pops atom from
/// the stack and deuses it.
///
/// Generated code:
///	call  {deuse}
//===================================================
void asm_pop_atom( context_t* ctx )
{
  CALL( deuse, TEXT("deuse") );
}

//===================================================
/// \fn		void asm_pop_dummy( context_t* ctx )
/// \brief	genarates a code to pop (and ignore) 
/// \param ctx	compilation context
///
/// Generates machine code which pops data from the
/// the stack and then ignores it.
///
/// Generated code:
///	pop  edx
//===================================================
void asm_pop_dummy( context_t* ctx )
{
  POP_EDX;
}

//===================================================
/// \fn		void asm_push_result( context_t* ctx )
/// \brief	genarates a code to push function's result
/// \param ctx	compilation context
///
/// Generates machine code which push the value
/// returned by a function into the stack.
///
/// Generated code:
///	push	eax
//===================================================
void asm_push_result( context_t* ctx )
{
  PUSH_EAX; REM( "move result in stack " );
}

//===================================================
/// \fn 	void asm_result_proc( context_t* ctx, atom_t source )
/// \brief		check result of a command
/// \param	ctx	compilation context
/// \param	source	source of the function
///
/// Generates code which check whether the result
/// of a command is acceptible.  The \c source parameter
/// points to the source where the procedure is called.
/// It is used to locate the error position in case
/// of errors.
//===================================================
void asm_result_proc( context_t* ctx, atom_t source )
{
  INFO( "" );
  INFO( "check for errors in proc %a", CAR(source) );
  PUSH_SOURCE( source );
  CALL( rt_cmdchk, TEXT("rt_cmdchk") ); REM( "check result, exit if error" );
  CMP_ID_ERROR;
  JE_EXIT;
  MOV_EBX_CONST( 0 );
}

//===================================================
/// \fn 	void asm_result_unknown( context_t* ctx, atom_t source )
/// \brief		check result of expression
/// \param	ctx	compilation context
/// \param	source	source of the function
///
/// Generates code which check whether the result
/// of an expression is acceptible. Actually, any
/// result is acceptible. 
//===================================================
void asm_result_unknown( context_t* ctx, atom_t source )
{
  INFO( "" );
  INFO( "check for errors in %a", CAR(source) );
  PUSH_SOURCE( source );
  CALL( rt_exprchk, TEXT("rt_exprchk") ); REM( "check result" );
}

//===================================================
/// \fn 	void asm_result_func( context_t* ctx, atom_t source )
/// \brief		check result of a function
/// \param	ctx	compilation context
/// \param	source	source of the function
///
/// Generates code which check whether the result
/// of a function is acceptible. The \c source parameter
/// points to the source where the function is called.
/// It is used to locate the error position in case
/// of errors.
//===================================================
void asm_result_func( context_t* ctx, atom_t source )
{
  INFO( "" );
  INFO( "check for errors in func %a", CAR(source) );
  PUSH_SOURCE( source ); REM( "source" );
  CALL( rt_funchk, TEXT("rt_funchk") );  REM( "check result" );
  asm_push_result( ctx );
}

//===================================================
/// \fn 	int asm_static_link( context_t* ctx, atom_t var )
/// \brief		generated code for static link
/// \param	ctx	compilation context
/// \param	var	var to which link is sought
/// \return		offset of the variable
///
/// Generates code with store the stack frame address
/// of a variable in ESI regiser. The result of the
/// function is the offset of the variable (if it is
/// normal variable) or its address (if it is function
/// or command).
///
/// Generated code:
///
///   mov esi,ebp
///   mov esi,ss:[esi-4]  ; repeated depth times
//===================================================
int asm_static_link( context_t* ctx, atom_t var )
{
  int depth = LEVEL(ctx->parent)-LEVEL(var)+1;

  #ifdef SAFEMODE
    assert(depth>=0);
  #endif

  MOV_ESI_EBP; REM( "ESI=stack frame of %a", NAME(var) );
  for( ; depth; depth-- ) MOV_ESI_SSMEM_ESI_4;

  return OFFSET( var );
}

//===================================================
/// \fn 	void	compile_system_reference( context_t* ctx, atom_t name ) 
/// \brief		compiles reference to system variable
/// \param	ctx	compilation context
/// \param	name	variable's name
///
/// Compiles a reference to a system variable. It is
/// assumed that the variable exists.
//===================================================
void compile_system_reference( context_t* ctx, atom_t name )
{
  atom_t var = find_var( name, ctx->parent );

  #ifdef SAFEMODE
    assert( var );
    assert( IS_VARIABLE( var ) );
  #endif //SAFEMODE

  asm_reference( ctx, var, unbound, 0 );
  CALL( use, TEXT("use") );
}

//===================================================
/// \fn 	void asm_reference( context_t* ctx, atom_t var, atom_t source, int check )
/// \brief		generated code for reference
/// \param	ctx	compilation context
/// \param	var	variable which value is referenced
/// \param	source	source of var reference
/// \param	check	if 0 then do not check result
///
/// Generates code for referencing variable's value.
/// The generated code depends on \c depth - this is
/// the difference between depths of syntax levels
/// of the current context and the context where the
/// var's value exist.
///
/// If the variable is primitive then it is stored in
/// globals and is accessed directly -- i.e. it's value
/// is not stored in the stack.
///
/// If check==0 push the value to the stack and exit.
/// Otherwise generate code to check whether the value
/// is not UNBOUND (i.e. to check whether the variable
/// has been assigned a value)
///
/// (a) var is primitive
///     push    [{address}]
///
/// (b) user-defined var
///     {static link}
///	push	[esi+offset]		; push value
///
/// if check!=0 then also generate:
///
///	push	{source}
///	call	asm_use_var			; use new value
///	cmp	[eax+OFFSET_ID],ERROR_ID
///	je	exit
///	push	eax
//===================================================
void asm_reference( context_t* ctx, atom_t var, atom_t source, int check )
{
  if( IS_PRIMITIVE(var) )
    { // case (a) - primitive variable
      int addr = (int)&VALUE(var);
      PUSH_MEM( addr );
    }
  else
    { // case (b) - user-defined variable
      int offset = asm_static_link( ctx, var );
      PUSH_MEM_ESI( offset );
    }
  REM( "value of %a", NAME(var) );

  // generate code to check whether
  // the variable has a value
  if( check )
    {
      PUSH_SOURCE( source ); REM( "source" );
      CALL( rt_use_var, TEXT("rt_use_var") );
      //CMP_ID_ERROR;
      //JE_EXIT;
      PUSH_EAX; REM( "still value of %a", NAME(var) );
    }
}

//===================================================
/// \fn 	void asm_make_direct( context_t* ctx, atom_t var, atom_t source )
/// \brief		generated code for direct make statement
/// \param	ctx	compilation context
/// \param	var	variable which value is changed
/// \param	source	source of the make command
///
/// Generates code for direct \c make  statement. I.e.
/// the var which value is changed is already known.
///
/// Before the execution of asm_make_direct() the
/// stack contains the new value.
///
/// Primitive variables are accessed directly.

/// For all user-defined variables the value of
/// \c depth determines what code is generated.
/// This variable is the difference between the
/// context where the var is defined and the
/// context where the var is used. The value of
/// \c delpth is used to generate code to reach
///  the proper context.
///
/// Code generation patterns (assumes the new value
/// is already pushed onto the stack):
///
/// (a) primitive (system) variables
///	push	[{address}]		; push old value
///	call	deuse			; deuse and pop old value
///	mov	eax,[esp]		; get new value
///	mov	[{address}],eax		; store new value
///
/// (b) user-defined var
///    {static}                         ; calculate base pointer in ESI
///	push	[esi+offset]		; push old value
///	call	deuse			; deuse and pop old value
///	mov	eax,[esp]		; get new value
///	mov	[esi+offset],eax	; store new value
//===================================================
void asm_make_direct( context_t* ctx, atom_t var, atom_t source )
{
  if( IS_PRIMITIVE(var) || IS_GLOBAL(var) )
    { // case (a)
      int addr = (int)&VALUE(var);

      // NB! No need to override default segment,
      // because it is already SS
      PUSH_MEM( addr ); REM( "free old value of %a", NAME(var) );
      CALL( deuse, TEXT("deuse") );
      MOV_EAX_MEM_ESP;  REM( "set new value of %a", NAME(var) );
      MOV_MEM_OFS_EAX( addr );
    }
  else
    {
      INFO( "" );
      INFO( "assign new value to %a", NAME(var) );
      int offset = asm_static_link( ctx, var );
      PUSH_MEM_ESI( offset ); REM( "free old value of %a", NAME(var) );
      CALL( deuse, TEXT("deuse") );
      MOV_EAX_MEM_ESP; REM( "set new value of %a", NAME(var) );
      MOV_MEM_ESI_EAX( offset );
    }

  // Generates code which check whether the result
  // of a make is acceptible.  The \c source parameter
  // points to the source where the procedure is called.
  // It is used to locate the error position in case
  // of errors.

  INFO( "" );
  INFO( "check for errors of %a", CAR(source) );
  PUSH_SOURCE( source );
  CALL( rt_makechk, TEXT("rt_makechk") ); REM( "check result atom" );
  CMP_ID_ERROR; REM( "is error atom?" );
  JE_EXIT;
}

//===================================================
/// \fn 	void asm_push_value_addr( context_t* ctx, atom_t var )
/// \brief		generated code for address of var's value
/// \param	ctx	compilation context
/// \param	var	variable which value is looked for
///
/// Generates code that calculates and pushes in the
/// the stack the address of the value of a variable
/// known at compile time.
///
///
/// (a) primitive (system) variables
///	push	{address}		; push value's address
///
/// (b) user-defined var
///    {static}                         ; calculate base pointer in ESI
///	add	esi,offset		; adjust for offset
///     push    esi                     ; push value's address
//===================================================
void asm_push_value_addr( context_t* ctx, atom_t var )
{
  if( IS_PRIMITIVE(var) )
    { // case (a)
      int addr = (int)&VALUE(var);

      PUSH_CONST( addr ); REM( "address of %a's value", NAME(var) );
    }
  else
    {
      int offset = asm_static_link( ctx, var );
      ADD_ESI_CONST_4( offset ); REM( "address of %a's value", NAME(var) );
      PUSH_ESI;
    }
}

//===================================================
/// \fn		void asm_call_atom( context_t* ctx, atom_t var, int params )
/// \brief	genarates a code to call atom
/// \param ctx		compilation context
/// \param var		atom to call
/// \param params	number of actual inputs
///
/// Generates machine code which calls a logo function.
/// The input \c atom contains the name of the function.
///
/// Generated code (for functions with fixed number of inputs):
/// - call {function}
///
/// Generated code (for functions with variable number of inputs or not-primitive):
/// - push {param count}
/// - call {function}
/// - pop  edx
//===================================================
void asm_call_atom( context_t* ctx, atom_t var, int params )
{
  int variadic = GET_FLAGS(var,FLAG_INFINITE_ARGS|FLAG_CAN_BE_UNARY|FLAG_MAY_HAVE_EXTRA_ARG);

  // all variadic functions need one more parameter
  if( variadic || !GET_FLAGS(var,FLAG_PRIMITIVE) )
    {
      PUSH_CONST( params ); REM( "param count of %a", NAME(var) );
    }


  // test for special system-defined variables-inputs
  // first test for print-related system inputs
  if( GET_FLAGS(var,FLAG_PRINT_VARS) )
    {
      compile_system_reference( ctx, NAME(fullprintp) ); 
      compile_system_reference( ctx, NAME(printwidthlimit) );
      compile_system_reference( ctx, NAME(printdepthlimit) );
    }
  if( GET_FLAGS(var,FLAG_EQUAL_VARS) )
    {
      compile_system_reference( ctx, NAME(caseignoredp) );
    }

  // test for run-time access to parent
  if( GET_FLAGS(var,FLAG_PUSH_PARENT) )
    {
      PUSH_CONST( (int)ctx->parent ); REM( "push parent of %a", NAME(var) );
      PUSH_EBP; REM( "push current frame" );
    }

  if( IS_PRIMITIVE(var) )
    {
      if( !ADDRESS(var) )
	{
	  printf("This primitive is not defined yet -> ");
	  dumpln(NAME(var));
	  return;
	}
      CALL_ATOM( (void*)ADDRESS(var), NAME(var) );
    }
  else
    {
      // prepare static link
      asm_static_link( ctx, var );
      CALL_MEM( &ADDRESS(var), NAME(var) );
    }

  // postprocessing of run-time access to parent
  if( GET_FLAGS(var,FLAG_PUSH_PARENT) )
    {
      POP_EDX; REM("remove static link");
      POP_EDX; REM("remove parent");
    }

  // postprocessing of system inputs
  if( GET_FLAGS(var,FLAG_EQUAL_VARS) )
    {
      asm_pop_atom( ctx ); REM( "remove %a", NAME(caseignoredp) );
    }
  if( GET_FLAGS(var,FLAG_PRINT_VARS) )
    {
      asm_pop_atom( ctx ); REM( "remove %a", NAME(printdepthlimit) );
      asm_pop_atom( ctx ); REM( "remove %a", NAME(printwidthlimit) );
      asm_pop_atom( ctx ); REM( "remove %a", NAME(fullprintp) );
    }

  // postprocessing of all variadic functions
  if( variadic || !GET_FLAGS(var,FLAG_PRIMITIVE) )
    {
      POP_EDX; REM("remove param count");
    }
}


//===================================================
/// \fn		void asm_runtime_reference( context_t* ctx )
/// \brief	genarates a code to get variable value
/// \param ctx		compilation context
///
/// Generates machine code which finds the value of a
/// variable and pushes it in the stack. The name is
/// in the stack, i.e. it is not processed at compile
/// time.
///
///	push	{parent}
///	push	ebp
///	call	:
///	pop	edx
///	pop	edx
//===================================================
void asm_runtime_reference( context_t* ctx )
{
  // always push the parent
  PUSH_CONST( (int)ctx->parent ); REM( "push parent" );
  PUSH_EBP; REM( "push static link" );
  CALL( rt_reference, TEXT(":") );
  POP_EDX; REM( "remove static link" );
  POP_EDX; REM( "remove parent" );
}

//===================================================
/// \fn 	void asm_dump_source( context_t* ctx, atom_t source )
/// \brief		dump current source
/// \param	ctx	compilation context
/// \param	source	source to dump
///
/// Generates code which dumps the current source.
/// This code is generally placed before calling
/// user-defined function of command.
//===================================================
void asm_dump_source( context_t* ctx, atom_t source )
{
  INFO( "" );
  PUSH_SOURCE( source ); REM( "dump source" );
  CALL( rt_predump, TEXT("rt_predump") );
  INFO( "" );
}

//===================================================
/// \fn		void asm_empty_body( context_t* ctx )
/// \brief	generates the body of an empty function
/// \param ctx	compilation context
///
/// Generates the body of an empty function. This body
/// just sets the result to be unbound atom.
///
/// Generated code:
/// - push (unbound)
/// - call use
/// - pop  eax
//===================================================
void asm_empty_body( context_t* ctx )
{
  INFO( "" );
  MOV_EAX_ATOM( unbound ); REM( "there is no result" );
}

//===================================================
/// \fn 	void asm_output( context_t* ctx, atom_t source, int true_output )
/// \brief		generated code for OUTPUT statement
/// \param	ctx	compilation context
/// \param	source	source of the OUTPUT command
/// \param	true_output flag whether OUPUT is real or simulated
///
/// Generates code for \c output statement. Before
/// the execution of asm_output() the stack contains
/// the new value to output.
///
/// Generated code:
///
///     mov     ebx,true_output
///	pop	eax			; load value in EAX
///	jmp	exit
//===================================================
void asm_output( context_t* ctx, atom_t source, int true_output )
{
  INFO( "" );
  INFO( "output current result" );

  #ifdef ADVANCED
  if( OPTION_RUNTIME )
    {
      PUSH_SOURCE( source ); REM( "dump source" );
      CALL( rt_dump, TEXT("dump") );
    }
  #endif

  MOV_EBX_CONST( true_output );
  POP_EAX; REM( "get current result in EAX" );
  JMP_EXIT;
}

//===================================================
/// \fn 	void asm_stop( context_t* ctx, atom_t source )
/// \brief		generated code for STOP statement
/// \param	ctx	compilation context
/// \param	source	source of the STOP command
///
/// Generates code for \c stop statement.
///
/// Generated code:
///
///	mov     ebx,1
///	mov	eax,{stopped}	; load value in EAX
///	jmp	exit
//===================================================
void asm_stop( context_t* ctx, atom_t source )
{
  INFO( "exit with no result" );

  #ifdef ADVANCED
  if( OPTION_RUNTIME )
    {
      PUSH_SOURCE( source ); REM( "dump source" );
      CALL( rt_dump, TEXT("rt_dump") );
    }
  #endif

  MOV_EBX_CONST( 1 );
  MOV_EAX_ATOM( stopped ); REM( "stopped" );
  JMP_EXIT;
}

//===================================================
/// \fn 	void asm_boolean( context_t* ctx, atom_t source )
/// \brief		generated code for checking boolean values
/// \param	ctx	compilation context
/// \param	source	source of the expression
///
/// Generates code that checks the top value in
/// the stack is \c TRUE or \c FALSE. In case of
/// \c TRUE sets \c EAX = \c false_true[1]. In
/// case of \c FALSE sets \c EAX = \c false_true[0].
/// If the value is neither \c TRUE nor \c FALSE
/// sets EAX to an error atom
///
/// Generated code:
///
///     call    rt_boolchk
///	cmp	[eax+OFFSET_ID],ERROR_ID
///	je	exit
///
//===================================================
void asm_boolean( context_t* ctx, atom_t source )
{
  INFO( "" );
  INFO( "check boolean result" );
  PUSH_SOURCE( source ); REM( "source" );
  CALL( rt_boolchk, TEXT("rt_boolchk") ); REM( "exit if not boolean" );
  CMP_ID_ERROR;
  JE_EXIT;
}

//===================================================
/// \fn 	int asm_if_prologue( context_t* ctx )
/// \brief		generates prologue for IF command
/// \param	ctx	compilation context
/// \return		branch address
///
/// Generates prologue code for the \c if  command.
/// Prologue is the code which stands right before
/// the code for \c then. It is used to direct
/// execution towards the \c then or the \c else
/// parts.
///
/// It is assumed that the code for the condition is
/// already ready and \c EAX contains the address of
/// either \c false_true[0] or \c false_true[1].
///
/// The return value is set to the address of the
/// \c else component of the \c je instruction.
/// It is used to set the correct offset once the
/// size of then-commands is known.
///
/// Generated code:
///	cmp	eax,FALSE
///	je	else
//===================================================
int asm_if_prologue( context_t* ctx )
{
  CMP_EAX_ATOM( false_true[0] ); REM( "is it false?" );
  JE( 0, TEXT("$else") ); REM( "goto if-else block" );
  return LOCAL_IP-4;
}

//===================================================
/// \fn 	int asm_if_epilogue( context_t* ctx )
/// \brief		generates epilogue for IF command
/// \param	ctx	compilation context
/// \return		branch address
///
/// Generates epilogue code for the \c if  command.
/// Epilogue is the code which stands right after the
/// code for the then-instructions. It contains only
/// a jump over the else-instructions directly to the
/// end of the \c if command.
///
/// Generated code:
///	jmp	ifend
//===================================================
int asm_if_epilogue( context_t* ctx )
{
  JMP( 0, TEXT("ifend") ); REM( "skip if-else block" );
  return LOCAL_IP-4;
}

//===================================================
/// \fn 	void asm_fix( context_t* ctx, int addr )
/// \brief		fixes unresolved offset
/// \param	ctx	compilation context
/// \param	addr	address of data to fix
/// \return		branch address
///
/// Fixes a relative offset from a given position to
/// the current position as defined by ctx. This is
/// used when there is a forward conditional or
/// inconditional jump, which target is unknown at
/// compile time.
///
/// The parameter addr is a 0-based address (i.e. it
/// is an offset relative to START_IP).
//===================================================
void asm_fix( context_t* ctx, int addr )
{
  reemit_4( ctx, addr, ctx->size-addr-4 );
}

//===================================================
/// \fn 	int asm_repeat_prologue_const( context_t* ctx, int cnt )
/// \brief		generates const-prologue for REPEAT command
/// \param	ctx	compilation context
/// \param	cnt	number of repetitions
/// \return		branch address
///
/// Generates const-prologue code for the \c repeat
/// command. Const-prologue is used when the number
/// of repetitions is known at compile time
/// (currently this means it is a constant).
///
/// The prologue initializes the \c repeat by calling
/// ::rt_repeat_enter to set up a new element of the
/// repeat chain.
///
/// The return value of the function is the address
/// of the instruction right after the prologue.
/// This address is used by ::asm_repeat_epilogue()
/// and is absolute address.
///
/// Generated code:
///
///	push	{count}
///	push	ebp
///	call    rt_repeat_enter
//===================================================
int asm_repeat_prologue_const( context_t* ctx, int cnt )
{
  PUSH_CONST( cnt ); REM( "repeat count" );
  PUSH_EBP; REM( "static link" );
  CALL( rt_repeat_enter, TEXT("rt_repeat_enter") ); REM( "initialize repeat" );

  INFO( "" );
  asm_label( ctx, TEXT("$rep:") );
  INFO( "start bofy of repeat" );

  return IP;
}

//===================================================
/// \fn 	int asm_repeat_prologue_expr( context_t* ctx, atom_t source, int* branch2 )
/// \brief		generates expr-prologue for REPEAT command
/// \param	ctx	compilation context
/// \param	source	source of the expression
/// \param	branch2 address of branch to skip over the whole repeat
/// \return		branch address
///
/// Generates expr-prologue code for the \c repeat
/// command. Expr-prologue is used when the number
/// of repetitions is an expression, i.e. it is not
/// known at compile time.
///
/// The prologue generates code to calculate the
/// actual number of repetitions. It checks whether
/// there was an error during calculation. If not,
/// it checks whether the number of repetitions is
/// valid. Finally the prologue initializes the
/// \c repeat by calling ::rt_repeat_enter to set
/// up a new element of the repeat chain.
///
/// The return value of the function is the address
/// of the instruction right after the prologue. This
/// address is used by ::asm_repeat_epilogue() and is
/// absolute address.
///
/// Generated code:
///
///    {expr}
///	push    {source}
///	call	rt_repchk
///	cmp	[eax+OFFSET_ID],ERROR_ID
///	je	exit
///
///	push	eax
///	mov	[eax+REPLIMIT],eax
///	call	deuse
///
///	cmp	eax,0
///	je	skip_repeat	][	; <-branch2
///
///	push	eax
///	push	ebp
///	call    rt_repeat_enter
/// $rep:
//===================================================
int asm_repeat_prologue_expr( context_t* ctx, atom_t source, int* branch2 )
{
  INFO( "" );
  INFO( "check validity of repeat count" );
  PUSH_SOURCE( source ); REM( "source" );
  CALL( rt_repchk, TEXT("rt_repchk") ); REM( "exit if invalid" );
  CMP_ID_ERROR;
  JE_EXIT;

  PUSH_EAX; REM( "free repeat count expression" );
  MOV_EAX_MEM_EAX( OFFSET_INT ); REM( "get repeat count" );
  CALL( deuse, TEXT("deuse") );

  CMP_EAX_CONST( 0 ); REM( "exit if it is 0" );
  JE( 0, TEXT("skip_repeat") );
  *branch2 = LOCAL_IP-4;

  PUSH_EAX; REM( "repeat count" );
  PUSH_EBP; REM( "static link" );
  CALL( rt_repeat_enter, TEXT("rt_repeat_enter") ); REM( "initialize repeat" );

  asm_label( ctx, TEXT("$rep:") );

  return IP;
}

//===================================================
/// \fn 	void asm_repeat_epilogue( context_t* ctx, int branch, int branch2 )
/// \brief		generates epilogue for REPEAT command
/// \param	ctx	compilation context
/// \param	branch	branch for repeat's loop
/// \param	branch2 branch for skipping repeat
///
/// Generates epilogue code for the \c repeat command.
/// There are two prologues for \c repeat, but the
/// epilogue is the same for both of them.
///
/// The epilogue of \c repeat decrements the
/// repeat count and if it is not zero, then
/// loops back to the branch address, which is
/// provided as parameter.
///
/// If the value of \c branch2 is -1, then the
/// repeat with constant number of repetitions and
/// \c branch2 is ignored. Otherwise it contains
/// the address where a branch is made over the whole
/// body of the repeat when the repetition number is
/// 0.
///
/// Generated code:
/// (note: {body} is not a part of the epilogue)
///
/// $rep:
///    {body}
///	mov	eax,[ebp+REPEATCHAIN]
///	mov	eax,[eax+CAR_OFFSET]
///	inc	[eax+REPCOUNT]
///	dec	[eax+REPLIMIT]
///	jnz	$rep
///
/// $skip:
///	push	ebp
///	call	rt_repeat_exit
//===================================================
//void rt_repeat_debug ( int replefts, int repcount, atom_t data )
//{
//  printf("atom=");dump_atom(data,1);
//  printf("repcount=%d/%d replefts=%d/%d\n",repcount,REPCOUNT(CAR(data)),replefts,REPLIMIT(CAR(data)));
//}

void asm_repeat_epilogue( context_t* ctx, int branch, int branch2 )
{

  INFO( "" );
  INFO( "calculate next loop of repeat" );
  MOV_EAX_MEM_EBP( BASE_OFFSET_REPEATCHAIN ); REM( "get repeat chain" );
  MOV_EAX_MEM_EAX( CAR_OFFSET ); REM( "get current repeat node" );
  INC_EAX_MEM( OFFSET_REPCOUNT ); REM( "increase done repeats" );
  DEC_EAX_MEM( OFFSET_REPLIMIT ); REM( "decrease left repeats" );

  //PUSH_MEM_EBP(BASE_OFFSET_REPEATCHAIN);
  //PUSH_MEM_EAX(OFFSET_REPCOUNT);
  //PUSH_MEM_EAX(OFFSET_REPLIMIT);
  //CALL( rt_repeat_debug, TEXT("") );

  JNZ( branch, TEXT("$rep") );

  if( branch2!=-1 )
    {
      INFO( "" );
      asm_label( ctx, TEXT("skip_repeat:") );
      INFO( "no more loops" );
      asm_fix( ctx, branch2 );
    }

  PUSH_EBP; REM( "static link" );
  CALL( rt_repeat_exit, TEXT("rt_repeat_exit") );
}

//===================================================
/// \fn 	int asm_forever_prologue( context_t* ctx )
/// \brief		generates prologue for FOREVER command
/// \param	ctx	compilation context
/// \return		branch address
///
/// Generates prologue code for the \c forever command.
///
/// The prologue initializes the \c forever by calling
/// ::rt_forever_enter to set up a new element of the
/// repeat chain.
///
/// The return value of the function is the address
/// of the instruction right after the prologue.
/// This address is used by ::asm_forever_epilogue().
///
/// Generated code:
///
///	push	ebp
///	call	rt_forever_enter
/// forever:
//===================================================
int asm_forever_prologue( context_t* ctx )
{
  PUSH_EBP; REM("static link");
  CALL( rt_forever_enter, TEXT("rt_forever_enter") ); REM( "initialize forever" );

  INFO( "" );
  asm_label( ctx, TEXT("forever:") );
  INFO( "start bofy of forever" );

  return IP;
}

//===================================================
/// \fn 	void asm_forever_epilogue( context_t* ctx, int branch )
/// \brief		generates epilogue for FOREVER command
/// \param	ctx	compilation context
/// \param	branch	branch for forever's loop
///
/// Generates epilogue code for the \c forever command.
///
/// Generated code:
/// (note: {body} is not a part of the epilogue)
///
/// forever::
///    {body}
///	mov	eax,[ebp+REPEATCHAIN]
///	mov	eax,[eax+CAR_OFFSET]
///	inc	[eax+REPCOUNT]
///	jmp	forever
//===================================================
void asm_forever_epilogue( context_t* ctx, int branch )
{
  INFO( "" );
  INFO( "calculate next loop of forever" );
  MOV_EAX_MEM_EBP( BASE_OFFSET_REPEATCHAIN ); REM( "get current repeat node" );
  MOV_EAX_MEM_EAX( CAR_OFFSET ); REM( "get current repeat node" );
  INC_EAX_MEM( OFFSET_REPCOUNT ); REM( "increase done repeats" );
  JMP( branch, TEXT("forever") );
}

//===================================================
/// \fn 	void asm_push_frame( context_t* ctx )
/// \brief		generates code to push frame pointer
/// \param	ctx	compilation context
///
/// Generates code for pushing the frame pointer
/// which is stored in EBP.
///
/// Generated code:
///
///	push	ebp
//===================================================
void asm_push_frame( context_t* ctx )
{
  PUSH_EBP; REM("push current frame");
}

//===================================================
/// \fn 	void asm_pop_frame( context_t* ctx )
/// \brief		generates code to pop frame pointer
/// \param	ctx	compilation context
///
/// Generates code for popping the frame pointer
/// previously pushed by ::asm_push_frame(). The
/// actual value that is popped out is not used
/// for anything.
///
/// Generated code:
///
///	pop	edx
//===================================================
void asm_pop_frame( context_t* ctx )
{
  POP_EDX; REM( "remove pushed frame" );
}

//===================================================
/// \fn 	void asm_push_mode( context_t* ctx, int mode )
/// \brief		generates code to push mode
/// \param	ctx	compilation context
/// \param	mode	mode to push
///
/// Generates code for pushing the mode of a function
/// call. The mode indicates whether the function is
/// called as a function or as a command. The values
/// of mode should be either COMPILE_AS_FUNC or
/// COMPILE_AS_PROC.
///
/// Generated code:
///
///	push   {mode}
//===================================================
void asm_push_mode( context_t* ctx, int mode )
{
  PUSH_CONST( mode );
  if( mode==COMPILE_AS_FUNC )
    REM( "call as function" );
  else
    REM( "call as procedure" );
}

//===================================================
/// \fn 	void asm_adjust_result( context_t* ctx )
/// \brief		generated code for result adjustment
/// \param	ctx	compilation context
///
/// Generates code which adjust the result of a command.
/// This is needed when EAX register is left unchanged
/// from the last expression.
///
/// Generated code:
///
///	mov	eax,{unbound}	; load value in EAX
//===================================================
void asm_adjust_result( context_t* ctx )
{
  MOV_EAX_ATOM( unbound ); REM( "set void result" );
}

//===================================================
/// \fn 	void asm_pop_result( context_t* ctx )
/// \brief		generated code for result adjustment
/// \param	ctx	compilation context
///
/// Generates code which adjust the result of a function.
/// This is needed when the result is in the stack, but
/// it is needed to be in EAX.
///
/// Generated code:
///
///	pop	eax
//===================================================
void asm_pop_result( context_t* ctx )
{
  POP_EAX; REM( "get current result" );
}


//===================================================
/// \fn 	int asm_while_prologue( context_t* ctx, int is_while, int is_do )
/// \brief		generates prologue for WHILE command
/// \param   ctx	compilation context
/// \param   is_while	while=1, until=0
/// \param   is_do do.while&do.until=1, while&until=0
/// \return		branch address
///
/// Generates prologue code for the \c while or the
/// \c until command (depending on the value of
/// \c is_while parameter). Actually no code is
/// generated, as long as the the prologue is just
/// a label.
///
/// Generated code (is_while==1 & is_do==0):
///
/// while:
///
/// Generated code (is_while==0 & is_do==0):
///
/// until:
///
/// Generated code (is_while==1 & is_do==1):
///
/// do.while:
///
/// Generated code (is_while==0 & is_do==1):
///
/// do.until:
///
//===================================================
int asm_while_prologue( context_t* ctx, int is_while, int is_do )
{
  if( is_do )
      {
          if( is_while )
              asm_label( ctx, TEXT("do.while:") );
          else
              asm_label( ctx, TEXT("do.until:") );
      }
      else
      {
          if( is_while )
              asm_label( ctx, TEXT("while:") );
          else
              asm_label( ctx, TEXT("until:") );
      }
  return IP;
}

//===================================================
/// \fn 	int asm_while_inlogue( context_t* ctx, atom_t source, int is_while )
/// \brief		generates inlogue for WHILE command
/// \param   ctx	compilation context
/// \param   source	source of the expression
/// \param   is_while	while=1, until=0
/// \return		address of branch to skip over while
///
/// Generates inlogue code for the \c while or the
/// \c until command (depending on the value of
/// \c is_while parameter). Inlogue is used to
/// calculate whether to continue the execution of
/// \c while or \c until.
///
/// The inlogue generates code to test the condition.
///
/// The return value of the function is the address
/// of the instruction where a fix-up is needed.
///
/// Generated code:
///
///	push    {source}
///	call	rt_whlchk
///	cmp	[eax+OFFSET_ID],ERROR_ID
///	je	exit

///	push	eax
///	mov	eax,[eax+INT]
///	call	deuse

///	cmp	eax,0 (for while) or 1 (for until)
///	je	skip_while
//===================================================
int asm_while_inlogue( context_t* ctx, atom_t source, int is_while )
{
  INFO( "" );
  INFO( "test condition of while/until" );
  PUSH_SOURCE( source ); REM( "source" );
  CALL( rt_whlchk, TEXT("rt_whlchk") ); REM( "check condition" );
  CMP_ID_ERROR; REM( "exit if error" );
  JE_EXIT;

  PUSH_EAX; REM( "free condition atom" );
  MOV_EAX_MEM_EAX( OFFSET_INT ); REM( "get condition result" );
  CALL( deuse, TEXT("deuse") );

  CMP_EAX_CONST( 1-is_while ); REM( "exit if condition fails" );
  JE( 0, is_while?TEXT("skip_while"):TEXT("skip_until") );

  return LOCAL_IP-4;
}


//===================================================
/// \fn 	void asm_while_epilogue( context_t* ctx, int loop_branch, int skip_branch, int is_while )
/// \brief		generates epilogue for WHILE/UNTIL command
/// \param ctx		compilation context
/// \param loop_branch	branch for while/until's loop
/// \param skip_branch	branch to skip while/until
/// \param is_while	while=1, until=0
///
/// Generates epilogue code for the \c while or
/// \c until command.
///
/// Generated code:
/// (note: {body} is not a part of the epilogue)
///
/// $whl:
///     ...
///    {body}
///	jmp	$whl
/// skip_while:
///     mov     eax,(unbound)
//===================================================
void asm_while_epilogue( context_t* ctx, int loop_branch, int skip_branch, int is_while )
{
  INFO( "" );
  INFO( "finalization of while/until" );
  JMP( loop_branch, is_while?TEXT("while"):TEXT("until") );

  asm_label( ctx, is_while?TEXT("skip_while"):TEXT("$skip_until") );
  asm_fix( ctx, skip_branch );

  MOV_EAX_ATOM( unbound ); REM( "set no result" );
}

//===================================================
/// \fn 	void asm_set_output_status( context_t* ctx, int n )
/// \brief		generates code for output status
/// \param     ctx	compilation context
/// \param     n	output status
///
/// Generates code to set output/stop status in EBX
///
/// Generated code:
///     mov	ebx,n
//===================================================
void asm_set_output_status( context_t* ctx, int n )
{
  MOV_EBX_CONST( n );
}


//===================================================
// / \fn 	void asm_enter_if_output( context_t* ctx )
// / \brief		generates enter code
// / \param     ctx	compilation context
// /
// / Generates code to store EBX used by asm_exit_if_output().
// /
// / Generated code:
// /     push	ebx
//===================================================
//void asm_enter_if_output( context_t* ctx )
//{
//  POP_EAX;
//  PUSH_EBX;
//  PUSH_EAX;
//}

//===================================================
// / \fn 	void asm_exit_if_output( context_t* ctx )
// / \brief		generates exit code
// / \param     ctx	compilation context
// /
// / Generates code which exit if EBX shows that
// / there was an active STOP or OUTPUT.
// /
// / Generated code:
// /     pop	ebx
// /	cmp	ebx,1
// /	je	exit
//===================================================
//void asm_exit_if_output( context_t* ctx )
//{
//  POP_EBX;
//  CMP_EBX_CONST( 1 );
//  JE_EXIT;
//}

//===================================================
/// \fn 	int asm_catch_prologue( context_t* ctx )
/// \brief		generates prologue for CATCH command
/// \param     ctx	compilation context
/// \return		catch exit address
///
/// Generates prologue code for the \c catch command.
/// Prologue is the code which stands right before
/// the code for the body of the \c catch. The prologue
/// is a trampoline which temporarily overwrited the
/// default function exit address. Thus, all exits
/// done in the commands of the \c catch are directed
/// to this trampoline.
///
/// Note! This function changes the exit_addr to point
/// to jump_addr!
///
/// Generated code:
///		jmp	$1
/// jump_addr:	jmp     catch_epilogue
///        $1: {body}
//===================================================
int asm_catch_prologue( context_t* ctx )
{
  JMP( IP+9, TEXT("catch") ); // skip over next jump
  ctx->exit_addr = IP;
  JMP( 0, TEXT("exit_catch") );  // to be fixed later

  INFO( "" );
  asm_label( ctx, TEXT("catch:") );
  INFO( "start of catch block" );

  return LOCAL_IP-4;
}

//===================================================
/// \fn 	void asm_catch_epilogue( context_t* ctx )
/// \brief		generates epilogue for CATCH command
/// \param	ctx	compilation context
///
/// Generates epilogue code for the \c catch command.
/// It fixes unset branch in the prologue and generates
/// code which captures and verifies the result of the
/// monitored block of commands.
///
/// Generated code:
///     push	{tag} ; already done
///	push	{eax} ; already done
///     call    rt_catchchk
///	mov	ebx,[catch_output_flag]
///	cmp	ebx,1
///	je	exit
///	push	eax
///
//===================================================
void asm_catch_epilogue( context_t* ctx )
{
  INFO( "" );
  INFO( "check result of catch block" );

  PUSH_EBX;
  CALL( rt_catchchk, TEXT("rt_catchchk") );
  MOV_EBX_MEM_OFS( &catch_output_flag );
  CMP_EBX_CONST(1);
  JE_EXIT;

//__asm__ volatile ("int $3; nop");

  PUSH_EAX;
}

//===================================================
/// \fn 	void asm_goto( context_t* ctx, atom_t tag )
/// \brief		generates jump to specific address
/// \param	ctx	compilation context
/// \param	tag	tag for the target address
///
/// Generates unconditional jump to a given address
/// specified by a tag variable.
///
/// Generated code:
///     jmp     addr
///
//===================================================
void asm_goto( context_t* ctx, atom_t tag )
{
  if( ctx->generate )
    {
      #ifdef SAFEMODE
        assert( IS_INTEGER(VALUE(tag)) );
      #endif

      JMP_ATOM( START_IP+INTEGER(VALUE(tag)), NAME(tag) );
    }
  else
    {
      JMP( 0, TEXT("") );
    }
}

//===================================================
/// \fn 	void asm_goto_prologue( context_t* ctx, atom_t source )
/// \brief		generates prologue for GOTO command
/// \param	ctx	compilation context
/// \param	source	source of the function
///
/// Generates code to push the source. It will be
/// used by ::rt_goto().
///
/// Generated code:
///
///	push	{source}
//===================================================
void asm_goto_prologue( context_t* ctx, atom_t source )
{
  PUSH_SOURCE( source ); REM( "source" );
}

//===================================================
/// \fn 	void asm_goto_epilogue( context_t* ctx, atom_t source )
/// \brief		generates epilogue for GOTO command
/// \param	ctx	compilation context
/// \param	source	source of goto command
///
/// Generates epilogue for \c goto command which
/// input cannot be resolve at compile time.
///
/// The epilogue generates code to find the target
/// address of the \c goto command. The code assumes
/// that EAX contains an integer atom with the
/// target address.
///
/// Generated code:
///
///    {expr}			; already generated
///	cmp	[eax+OFFSET_ID],ERROR_ID
///	je	exit
///
///	push	eax			; tag
///	mov	eax,[eax+INTEGER]	; EAX=target
///	call	deuse			; tag
///
///	add	eax,{base}
///	jmp	eax
//===================================================
void asm_goto_epilogue( context_t* ctx, atom_t source )
{
  CMP_ID_ERROR; REM( "exit if error" );
  JE_EXIT;

  PUSH_EAX; REM( "get target relative address" );
  MOV_EAX_MEM_EAX( OFFSET_INT );
  CALL( deuse, TEXT("deuse") );

  ADD_EAX_CONST_4( START_IP ); REM( "make it absolute" );
  JMP_EAX( source );
}

//===================================================
/// \fn 	int asm_test_prologue( context_t* ctx, int criteria )
/// \brief		generates prologue for IFTRUE/IFFALSE command
/// \param   ctx	compilation context
/// \param   criteria	iftrue=1, iffalse=0
/// \return		address of branch to skip over iftrue/iffalse
///
/// Generates prologue code for the \c iftrue or the
/// \c iffalse command (depending on the value of
/// \c criteria parameter). Prologue is used to
/// skip the body if the criyeria is not fulfilled.
///
/// The return value of the function is the address
/// of the instruction where a fix-up is needed.
///
/// Generated code:
///
///	cmp	[ebp+OFFSET_TEST],criteria
///	jne	skip
//===================================================
int asm_test_prologue( context_t* ctx, int criteria )
{
  CMP_MEM_EBP( BASE_OFFSET_TEST, criteria ); REM( "check test result" );
  JNE( 0, TEXT("skip") );

  return LOCAL_IP-4;
}

//===================================================
/// \fn 	void asm_run_epilogue( context_t* ctx )
/// \brief		generates epilogue for RUN command
/// \param	ctx	compilation context
///
/// Generates epilogue for \c runmacro command which
/// does the actual execution of commands.
///
/// The epilogue generates code to get the address
/// of compiled code and execute it.
///
/// Generated code:
///
///    {run}			; already generated
///	cmp	[eax+OFFSET_ID],ERROR_ID
///	je	exit
///
///	push	eax			; EAX=var atom
///	mov	eax,[eax+DESCR1]	; get DESCR1 which contains the address
///	call	[eax+ADDRESS]		; code to run, EAX=result of run code
///	call	deuse			; deuse var atom (including run code)
///
///	cmp	ebx,1
///	je	exit
///
///	cmp	eax,{stopped}
///	je	exit
//===================================================
void asm_run_epilogue( context_t* ctx )
{
  INFO( "" );
  INFO( "actual execution of run" );

  CMP_ID_ERROR; REM( "exit if error" );
  JE_EXIT;

  MOV_ESI_EBP; REM( "static link" );
  PUSH_EAX; REM( "EAX=var atom" );
  MOV_EAX_MEM_EAX( OFFSET_DESCR1 ); REM( "EAX=var descriptor" );

  CALL_MEM_EAX( OFFSET_ADDRESS, TEXT("ADDRESS") ); REM( "get and call var address" );

  CALL( deuse, TEXT("deuse") );

  CMP_EBX_CONST( 1 );
  JE_EXIT;

  CMP_EAX_ATOM(stopped); REM( "exit if stopped" );
  JE_EXIT;
}


//===================================================
/// \fn 	void asm_runresult_epilogue( context_t* ctx )
/// \brief		generates epilogue for RUNRESULT command
/// \param	ctx	compilation context
///
/// Generates epilogue for \c runresult command which
/// does the actual execution of commands.
///
/// The epilogue generates code to get the address
/// of compiled code and execute it. It reuses code
/// generated by ::asm_run_epilogue().
///
/// Generated code:
///
///    {run}			; already generated
///	  cmp	[eax+OFFSET_ID],ERROR_ID
///	  je	skip
///
///	  push	eax			; EAX=var atom
///	  mov	eax,[eax+DESCR1]	; get DESCR1 which contains the address
///	  call	[eax+ADDRESS]		; code to run, EAX=result of run code
///	  call	deuse			; deuse var atom (including run code)
/// skip:
///	  push	eax
///	  call	rt_runresult_fix
//===================================================
void asm_runresult_epilogue( context_t* ctx )
{
  INFO( "" );
  INFO( "actual execution of runresult" );

  CMP_ID_ERROR; REM( "skip if error" );
  JE( 0, TEXT("$skip") );
  int branch=LOCAL_IP-4;

  MOV_ESI_EBP; REM( "static link" );
  PUSH_EAX; REM( "EAX=var atom" );
  MOV_EAX_MEM_EAX( OFFSET_DESCR1 ); REM( "EAX=var descriptor" );

  CALL_MEM_EAX( OFFSET_ADDRESS, TEXT("ADDRESS") ); REM( "get and call var address" );

  CALL( deuse, TEXT("deuse") );
  
LABEL( "$skip:" );
  asm_fix( ctx, branch );
  
  PUSH_EAX;
  CALL( rt_runresult_fix, TEXT("rt_runresult_fix") );
}




//===================================================
/// \fn 	int asm_for_prologue( context_t* ctx, atom_t step_var, atom_t source, int* branch2 )
/// \brief		generates prologue for FOR command
/// \param	ctx	compilation context
/// \param      step_var    variable which will hold the value of FOR's step
/// \param	source	source of the expression
/// \param	branch2 address of branch to skip over the whole repeat
/// \return		branch address
///
/// Generates prologue code for the \c for
/// command. 
///
/// The prologue generates code to calculate the
/// actual number of repetitions. It checks whether
/// there was an error during calculation. If not,
/// it checks whether the number of repetitions is
/// valid. Finally the prologue initializes the
/// \c for by calling ::rt_repeat_enter to set
/// up a new element of the repeat chain.
///
/// The return value of the function is the address
/// of the instruction right after the prologue. This
/// address is used by ::asm_for_epilogue() and is
/// absolute address.
///
/// Generated code:
///
///    {expr_from}
///    {expr_to}
///    {expr_step}
///     push    {addr of step_var's value}
///	push    {source}
///	call	rt_forchk
///	cmp	[eax+OFFSET_ID],ERROR_ID
///	je	exit
///
///	push	eax
///	mov	[eax+REPLIMIT],eax
///	call	deuse
///
///	cmp	eax,0
///	je	skip_for	][	; <-branch2
///
///	push	eax
///	push	ebp
///	call    rt_repeat_enter
/// $for:
//===================================================
int asm_for_prologue( context_t* ctx, atom_t step_var, atom_t source, int* branch2 )
{
    
  asm_push_value_addr( ctx, step_var );
  
  INFO( "" );
  INFO( "check validity of for count" );
  PUSH_SOURCE( source ); REM( "source" );
  CALL( rt_forchk, TEXT("rt_forchk") ); REM( "exit if invalid" );
  CMP_ID_ERROR;
  JE_EXIT;

  PUSH_EAX; REM( "free for count expression" );
  MOV_EAX_MEM_EAX( OFFSET_INT ); REM( "get for count" );
  CALL( deuse, TEXT("deuse") );

  CMP_EAX_CONST( 0 ); REM( "exit if it is 0" );
  JE( 0, TEXT("skip_for") );
  *branch2 = LOCAL_IP-4;

  PUSH_EAX; REM( "for count" );
  PUSH_EBP; REM( "static link" );
  CALL( rt_repeat_enter, TEXT("rt_repeat_enter") ); REM( "initialize for" ); // reuse rt_repeat_enter

  asm_label( ctx, TEXT("$for:") );

  return IP;
}

//===================================================
/// \fn 	void asm_for_epilogue( context_t* ctx, int branch, int branch2 )
/// \brief		generates epilogue for FOR command
/// \param	ctx	compilation context
/// \param	branch	branch for for's loop
/// \param	branch2 branch for skipping for
///
/// Generates epilogue code for the \c for command.
///
/// The epilogue of \c for decrements the
/// repeat count and if it is not zero, then
/// loops back to the branch address, which is
/// provided as parameter.
///
/// \c branch2 contains the address where a branch
/// is made over the whole body of the for when 
/// the repetition number is 0.
///
/// Generated code:
/// (note: {body} is not a part of the epilogue)
///
/// $for:
///    {body}
///	mov	eax,[ebp+REPEATCHAIN]
///	mov	eax,[eax+CAR_OFFSET]
///	inc	[eax+REPCOUNT]
///	dec	[eax+REPLIMIT]
///	jnz	$for
///
/// $skip:
///	push	ebp
///	call	rt_repeat_exit  // reuses repeat_exit
//===================================================
//void rt_for_debug ( int replefts, int repcount, atom_t data )
//{
//  printf("atom=");dump_atom(data,1);
//  printf("repcount=%d/%d replefts=%d/%d\n",repcount,REPCOUNT(CAR(data)),replefts,REPLIMIT(CAR(data)));
//}

void asm_for_epilogue( context_t* ctx, int branch, int branch2 )
{

  INFO( "" );
  INFO( "calculate next loop of for" );
  MOV_EAX_MEM_EBP( BASE_OFFSET_REPEATCHAIN ); REM( "get repeat chain" );
  MOV_EAX_MEM_EAX( CAR_OFFSET ); REM( "get current repeat node" );
  INC_EAX_MEM( OFFSET_REPCOUNT ); REM( "increase done repeats" );
  DEC_EAX_MEM( OFFSET_REPLIMIT ); REM( "decrease left repeats" );

  //PUSH_MEM_EBP(BASE_OFFSET_REPEATCHAIN);
  //PUSH_MEM_EAX(OFFSET_REPCOUNT);
  //PUSH_MEM_EAX(OFFSET_REPLIMIT);
  //CALL( rt_repeat_debug, TEXT("") );

  JNZ( branch, TEXT("$for") );

  if( branch2!=-1 )
    {
      INFO( "" );
      asm_label( ctx, TEXT("skip_for:") );
      INFO( "no more loops" );
      asm_fix( ctx, branch2 );
    }

  PUSH_EBP; REM( "static link" );
  CALL( rt_repeat_exit, TEXT("rt_repeat_exit") ); // same as exit from repeat
}



//===================================================
/// \fn		void asm_external_function( context_t* ctx, atom_t func )
/// \brief	generates trampoline to external function
/// \param ctx	compilation context
/// \param func	function which trampoline to generate
///
/// Generates a trampoline to call an external
/// function. The trampoline's code converts Logo
/// parameters into C-type data and prepares the
/// stack to be as expected by a sdtcall function.
///
/// Generated code:
///
/// func:
///   push  ebp
///   mov   ebp,esp
///
///   ; clear error flag
///   mov   [error_flag],0
///
///   ; for each parameter do:
///   push  [ebp+{ofs}]   ; push param
///   call  atom_to_xxx   ; convert to c-type
///   push  edx           ; push c-type hi word (if 64-bit)
///   push  eax           ; push c-type lo word
///   fstp  [esp]         ; store 32- or 64-bit fp number to stack
///
///   call  func
///
///   ; check error flag
///   cmp   [error_flag],0
///   je    ok
///   mov   eax,[error_flag]
///   jmp   exit
///
///   ; push result
/// ok:
///   push  edx           ; push c-type hi word (if 64-bit)
///   push  eax           ; push c-type lo word
///   fstp  [esp]         ; store 32- or 64-bit fp number to stack
///   call  xxx-to_atom   ; convert back to atom
///
/// exit:
///   mov   esp,ebp
///   pop   ebp
///   ret
//===================================================
void asm_external_function( context_t* ctx, atom_t func )
{
  INFO( "" );
  INFO( "=========================" );
  INFO( " EXTERNAL FUNCTION %a", NAME(func) );
  INFO( "=========================" );
  INFO( "" );

  LABEL( "%a:", NAME(func) );
  PUSH_EBP;	REM( "function prologue" );
  MOV_EBP_ESP;

  INFO( "" );
  MOV_MEM_OFS_CONST( &error_flag, 0 );
  REM( "clear error flag" );
  INFO( "" );
  
  INFO( "convert parameters to C-types" );

  atom_t params;
  for( params=LOCALS(func); IS_NOT_EMPTY(params); params=CDR(params) )
  {
    atom_t param = CAR(params);
    if( !IS_EXTERNAL(param) ) continue;

    int c_type = VARCLASS( param );

    // push param's atom and convert it to c-type
    PUSH_MEM_EBP( OFFSET(param) ); REM( "convert %a", NAME(param) );
    CALL( c_types[c_type].atom_to_c, c_types[c_type].atom_to_c_name );

    // push the result to the stack. If the result is 1..32 bits, then
    // it is in EAX. If it is 33..64 bits, then it is stored in EDX:EAX.
    if( c_types[c_type].size>32 ) { PUSH_EDX; }
    PUSH_EAX;
    
    // floating-point numbers are stored in FPU's st0 register.
    // Pop the value from the register and store it in the stack.
    if( c_types[c_type].class==C_TYPE_FLOAT )
      {
	if( c_types[c_type].size>32 )
	  { FSTPD_MEM_ESP; } // store 64-bit float
	else
	  { FSTPF_MEM_ESP; } // store 32-bit float
      }
  } //params=LOCALS(func)
 
  INFO( "" );
  INFO( "call to external function" );
  CALL_ATOM( ADDRESS(func), NAME(func) );

  INFO( "" );
  INFO( "check error flag" );

  CMP_MEM_OFS_CONST_4( &error_flag, 0 );
  JE( 0, TEXT("$ok") ); REM( "continue if no error" );
  int branch=LOCAL_IP-4;
  MOV_EAX_MEM_OFS( &error_flag );
  JMP( 0, TEXT("exit") );
  int branch2=LOCAL_IP-4;

  // function result must be converted back into atom
  INFO( "" );
  INFO( "convert function result" );
LABEL( "$ok:" );
  asm_fix( ctx, branch );
  int c_type = VARCLASS( func );
  if( c_types[c_type].size>32 ) { PUSH_EDX; }
  PUSH_EAX;
  if( c_types[c_type].class==C_TYPE_FLOAT )
    {
      if( c_types[c_type].size>32 )
        { FSTPD_MEM_ESP; } // store 64-bit float
      else
        { FSTPF_MEM_ESP; } // store 32-bit float
    }
  CALL( c_types[c_type].c_to_atom, c_types[c_type].c_to_atom_name );
  INFO( "" );

  // fix the stack by removing all c-type data
  // (actually this is not needed, because the
  // next MOV ESP,EBP command fixes the stack
LABEL( "exit:" );
  asm_fix( ctx, branch2 );
  MOV_ESP_EBP;
  POP_EBP;
  RET;	REM( "end of %a", NAME(func) );
}




//===================================================
/// \fn		void asm_internal_function( context_t* ctx, int static_link, atom_t func )
/// \brief	generates trampoline to external function
/// \param ctx	compilation context
/// \param static_link	static link from the current frame
/// \param func	function which trampoline to generate
///
/// Generates a trampoline to call an external
/// function. The trampoline's code converts Logo
/// parameters into C-type data and prepares the
/// stack to be as expected by a sdtcall function.
///
/// Generated code:
///
/// func:
///
/// exit:
///	ret
//===================================================
void asm_internal_function( context_t* ctx, int static_link, atom_t func )
{
  //INT_3;
  INFO( "" );
  INFO( "=========================" );
  INFO( " INTERNAL FUNCTION %a", NAME(func) );
  INFO( "=========================" );
  INFO( "" );
  asm_prologue( ctx, func, 0 );

  INFO( "convert parameters to atoms" );
  atom_t params;
  int offset = 8; // this is offset from a C calling code
  int count = 0;
  for( params = LOCALS(func); IS_NOT_EMPTY(params); params=CDR(params))
  {
    atom_t param = CAR( params );
    if( !IS_INTERNAL(param) ) continue;
    int c_type = VARCLASS( param );

    PUSH_MEM_EBP( offset );
    CALL( c_types[c_type].c_to_atom, c_types[c_type].c_to_atom_name );
    PUSH_EAX;
    offset += c_types[c_type].size/8;
    count++;
  }

  PUSH_CONST( count );

  //MOV_EBP_CONST( static_link );
  //MOV_ESI_CONST( static_link );
  MOV_MEM_OFS_EBP( &backup_frame );
  MOV_EBP_MEM_OFS( &root_frame );
  MOV_ESI_EBP;
  INFO( "" );
  INFO( "call to internal function" );
  //INT_3;
  CALL_ATOM( ADDRESS(func), NAME(func) );
  MOV_EBP_MEM_OFS( &backup_frame );

  CMP_ID_ERROR;
  JNE( 0, TEXT("$skip") );
  int branch = LOCAL_IP-4;
  MOV_MEM_OFS_EAX( &error_flag ); REM( "store result in error flag");
  asm_fix( ctx, branch );
LABEL( "$skip" );

  INFO( "" );
  INFO( "convert function result" );
  int c_type = VARCLASS( func );
  PUSH_EAX;
  PUSH_EAX;
  CALL( c_types[c_type].atom_to_c, c_types[c_type].atom_to_c_name );
  INFO( "" );
  CALL( deuse, TEXT("deuse result") );

  POP_EDX;
  for( params=LOCALS(func); IS_NOT_EMPTY(params); params=CDR(params) )
    if( IS_INTERNAL(CAR(params)) )
      CALL( deuse, TEXT("deuse") );

  //INT_3;
  asm_epilogue( ctx, func, COMPILE_AS_NON_MACRO );
}





//===================================================
/// \fn		void asm_int_3( context_t* ctx )
/// \brief	generates Interrupt 3
/// \param ctx	compilation context
///
/// Generates instruction for trapping debugger.
///
/// Generated code:
///
///   int   3
//===================================================
void asm_int_3( context_t* ctx )
{
  INT_3;
}



//===================================================
/// \fn		void asm_nop( context_t* ctx )
/// \brief	generates NOP
/// \param ctx	compilation context
///
/// Generates instruction for no operation.
///
/// Generated code:
///
///   nop
//===================================================
void asm_nop( context_t* ctx )
{
  NOP;
}
