//
// Project:	Lhogho
// File:	lists.c
//
// Copyright (C) 2007 P.Boytchev
//
// Revision history:
//   2007-06-09	- file created from atoms.c
//   2007-08-13 - support for PRINTDEPTHLIMIT, PRINTWIDTHLIMIT
//   2007-09-21 - list_copy_n, list_copy_but_last, list_length, get_at_list
//   2008-01-17 - extended atoms
//   2008-01-23 - Fixed bug #1836433 Example with IFTRUE fails
//   2012-09-23 - Added DEBUG_ATOM_LIST macro
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//



#include "globals.h"
#ifdef UNICODE_CHARS
  #include <wchar.h>	// UNICODE support: wcslen()
#endif //UNICODE_CHARS

#include <assert.h>	// assert()
#include <errno.h>	// errno, ENOMEM
#include <stdio.h>	// fopen(), fread(), fclose()
#include <stdlib.h>     // malloc()
#include <string.h>     // memcpy()
#include <locale.h>	// LC_ALL

#include "atoms.h"
#include "words.h"
#include "lists.h"
#include "errors.h"
#include "pools.h"
#include "unicode.h"
#include "translate.h"
#include "vars.h"
#include "parser.h"
#include "options.h"
#include "compiler.h"
#include "numbers.h"

atom_t empty_list;		///< single instance of empty list [ ]
int print_width_limit = -1;	///< integer of PRINTWIDTHLIMIT
int print_depth_limit = -1;	///< integer of PRINTDEPTHLIMIT



//===================================================
/// \fn		void delete_list( atom_t a ) 
/// \brief	deletes list atom
/// \param a	atom to delete
///
/// Deletes the first node of list atom and its data.
/// If needed deleted the next nodes.
//===================================================
void delete_list( atom_t a )
{

//  printf("DELETE_LIST(%x) ->",(int)a); dump_atom(a,1); printf("\n");

  while( 1 ) 
    {
      atom_t next = CDR(a);
      //printf(">>>   DEUSE CAR={"); dump_atom(CAR(a),1); printf("}\n");
      DEUSE( CAR(a) );
      //printf(">>>   DONE DEUSE CAR\n");
      if( IS_EXTENDED(a) )
      {
	//printf(">>>   DEUSE POS={");
	//dump_atom(POS(a),1);
	//printf("}\n");
	DEUSE( POS(a) ); 
	return_to_pool( &data_pool_ex, a ); // extended atom
      }
      else
	{
	  return_to_pool( &data_pool, a );
	}

      a = next;
      if( IS_EMPTY(a) ) return;

      #if defined DEBUG_ATOM || defined DEBUG_ATOM_LIST
        printf("<ATOM>  [%08x] ref-1\n",(int)a);
      #endif //DEBUG_ATOM
      if( --REF(a) ) return;

      #ifdef ADVANCED
	stats[ID(a)].deallocs++;
	stats_free++;
      #endif //ADVANCED
    };
}




//===================================================
/// \fn 	atom_t new_list( atom_t car, atom_t cdr )
/// \brief		creates a new list atom
/// \param	car	value associated with the new list node
/// \param	cdr	next node of the list
/// \return		list atom
///
/// Creates a list atom with reference count 1. The
/// reference counts of \c car and \c cdr are not
/// changed.
//===================================================
atom_t new_list( atom_t car, atom_t cdr )
{
  atom_t a = (atom_t)take_from_pool( &data_pool );

  REF(a)   = 1;
  ID(a)    = LIST_ID;
  CAR(a)   = car;
  CDR(a)   = cdr;
  FLAGS(a) = 0;

  #ifdef SAFEMODE
    assert( car );
    assert( cdr );
  #endif //SAFEMODE

  #if defined DEBUG_ATOM || defined DEBUG_ATOM_LIST
    if( IS_EMPTY(cdr) )
      {
	printf("<ATOM>  [%08x] list=[%08x]",(int)a,(int)car);
	printf(" CAR="); dumpln(car);
      }
    else
      printf("<ATOM>  [%08x] list=[%08x %08x]\n",(int)a,(int)car,(int)cdr);
  #endif //DEBUG_ATOM

  #ifdef ADVANCED
    stats[ID(a)].allocs++;
    if( stats[ID(a)].max<(stats[ID(a)].allocs-stats[ID(a)].deallocs) )
      stats[ID(a)].max=(stats[ID(a)].allocs-stats[ID(a)].deallocs);
    stats_free--;
  #endif //ADVANCED

  #ifdef DEBUG_RUNTIME_ATOMS
  if( running_compiled_code )
    {
      outter( TEXT("<RUNTIME> new  "), -1 );
      dump_atom_address( a );
      dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
  #endif
  #ifdef DEBUG_COMPILETIME_ATOMS
  if( compiling_code )
    {
      outter( TEXT("<COMPILETIME> new  "), -1 );
      dump_atom_address( a );
      dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
  #endif

  return a;
}




//===================================================
/// \fn 	atom_t new_list_ex( atom_t car, atom_t cdr )
/// \brief		creates a new extended list atom
/// \param	car	value associated with the new list node
/// \param	cdr	next node of the list
/// \return		list atom
///
/// Creates an extended list atom with reference count 1.
/// The reference counts of \c car and \c cdr are not
/// changed. The code should be the same as new_list()
/// except that the extended pool is used, and the
/// ::POS field is set to unbound
//===================================================
atom_t new_list_ex( atom_t car, atom_t cdr )
{
  atom_t a = (atom_t)take_from_pool( &data_pool_ex ); // extended atom
  //printf("============>new_list_ex at %x\n",(int)a);

  REF(a)   = 1;
  ID(a)    = LIST_ID;
  CAR(a)   = car;
  CDR(a)   = cdr;
  POS(a)   = unbound;			// extended atom
  FLAGS(a) = FLAG_EXTENDED_NODE;	// extended atom

  #ifdef SAFEMODE
    assert( car );
    assert( cdr );
  #endif //SAFEMODE

  #if defined DEBUG_ATOM || defined DEBUG_ATOM_LIST
    if( IS_EMPTY(cdr) )
      printf("<ATOMX> [%08x] list=[%08x]\n",(int)a,(int)car);
    else
      printf("<ATOMX> [%08x] list=[%08x %08x]\n",(int)a,(int)car,(int)cdr);
  #endif //DEBUG_ATOM

  #ifdef ADVANCED
    stats[ID(a)].allocs++;
    if( stats[ID(a)].max<(stats[ID(a)].allocs-stats[ID(a)].deallocs) )
      stats[ID(a)].max=(stats[ID(a)].allocs-stats[ID(a)].deallocs);
    stats_free--;
  #endif //ADVANCED

  #ifdef DEBUG_RUNTIME_ATOMS
  if( running_compiled_code )
    {
      outter( TEXT("<RUNTIME> newx "), -1 );
      dump_atom_address( a );
      dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
  #endif
  #ifdef DEBUG_COMPILETIME_ATOMS
  if( compiling_code )
    {
      outter( TEXT("<COMPILETIME> newx "), -1 );
      dump_atom_address( a );
      dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
  #endif

  return a;
}




//===================================================
/// \fn		void dump_list( atom_t a, int level ) 
/// \brief	dumps list atom
/// \param a	atom to dump
/// \param level dump level
///
/// Dumps list atom through the current outter function.
//===================================================
void dump_list( atom_t a, int level )
{
  if( print_depth_limit!=-1 && print_depth_limit<=level )
    {
      outter( TEXT("..."), 3 );
      outter_size += 3;
      return;
    }

  #define DUMP_BUF_SIZE 4
  char_t buf[DUMP_BUF_SIZE];
  int expr;

  expr = IS_EXPRESSION(a);

  #ifdef DEBUG_REF_COUNT
    level=1;
  #endif

  if( expr )
    {
      buf[0] = TEXT('(');
      outter( buf, 1 );
      outter_size += 1;
    }
  else
    if( level ) 
      {
	buf[0] = TEXT('[');
	outter( buf, 1 );
	outter_size += 1;
      }
  #ifdef DEBUG_TOKENS
    printf("\n");
  #endif

  int space = 0;
  int count = 0;

  while( IS_NOT_EMPTY(a) )
    {
      if( space )
	{
	  outter( TEXT(" "), 1 );
	  outter_size += 1;
	}
      else { space = 1; }

      if( 0<=print_width_limit && print_width_limit<=count )
	{
	  outter( TEXT("..."), 3 );
	  outter_size += 3;
	  break;
	}


  #ifdef DEBUG_REF_COUNT
    if( IS_EMPTY(a) )
      {
	outter( TEXT("{||}"), 2 );
      }
    else
      {
	int n;
	char_t buf[64];
	n = SPRINTF( buf, DUMP_BUF_SIZE, TEXT("|%d|"), REF(a) );
	outter( buf, n );
      }
  #endif

      dump_atom( CAR(a), level+1 );

/*        if( IS_EXTENDED(a) ) */
/*  	{ */
/*  	  printf("@{"); */
/*  	  dump_atom( POS(a), 1 );  */
/*  	  printf("}");  */
/*  	}  */


      #ifdef DEBUG_LIST_FLAGS
      if( GET_FLAGS(a,FLAG_NEWLINE+FLAG_AFTER_SPACE+FLAG_BEFORE_SPACE) )
	{ 
	  outter(TEXT("{"),1);
	  if( GET_FLAGS(a,FLAG_NEWLINE) ) outter(TEXT("New"),3); 
	  if( GET_FLAGS(a,FLAG_AFTER_SPACE) ) outter(TEXT("Asp"),3); 
	  if( GET_FLAGS(a,FLAG_BEFORE_SPACE) ) outter(TEXT("Bsp"),3); 
	  outter(TEXT("}"),1);
	}
      #endif // DEBUG_LIST_FLAGS
      #ifdef DEBUG_LIST_TOKEN_FLAGS
      if( GET_FLAGS(a,FLAG_TOKENIZED_DATA+FLAG_TOKENIZED_COMMANDS) )
	{ 
	  //outter(TEXT("{"),1);
	  if( GET_FLAGS(a,FLAG_TOKENIZED_DATA) ) outter(TEXT("'"),1); 
	  if( GET_FLAGS(a,FLAG_TOKENIZED_COMMANDS) ) outter(TEXT("!"),1); 
	  //outter(TEXT("}"),1);
	}
      #endif // DEBUG_LIST_FLAGS
      count++;
      a = CDR(a);
    }

  if( expr )
    {
      buf[0] = TEXT(')');
      outter( buf, 1 );
      outter_size += 1;
    }
  else
    if( level ) 
      {
	buf[0] = TEXT(']');
	outter( buf, 1 );
	outter_size += 1;
      }

  #ifdef DEBUG_TOKENS
  printf("\n");
  #endif

#undef DUMP_BUF_SIZE
}



//===========================================================
/// \fn	      void append( atom_t element, atom_t *first, atom_t *last )
/// \brief		appends atom to list
/// \param    element	atom to be appended
/// \param    first	first node of the list
/// \param    last	last node of the list
///
/// Appends atom \c element to the end of the list starting
/// with atom \c first and ending with atom \c last. If \c
/// first and \c last are empty a new list will be created.
///  All reference counts are kept unchanged.
//===========================================================
void append( atom_t element, atom_t *first, atom_t *last )
{
  atom_t new = new_list( element, empty_list );

  if( IS_NOT_EMPTY(*first) )
    CDR(*last) = new;
  else
    *first = new;

  *last = new;
}



//===========================================================
/// \fn	      void append_ex( atom_t element, atom_t *first, atom_t *last )
/// \brief		appends atom to extended node of a list
/// \param    element	atom to be appended
/// \param    first	first node of the list
/// \param    last	last node of the list
///
/// Appends atom \c element to the end of the list starting
/// with atom \c first and ending with atom \c last. If \c
/// first and \c last are empty a new list will be created.
/// All reference counts are kept unchanged. The new node is
/// created as extended atom.
//===========================================================
void append_ex( atom_t element, atom_t *first, atom_t *last )
{
  atom_t new = new_list_ex( element, empty_list );

  if( IS_NOT_EMPTY(*first) )
    CDR(*last) = new;
  else
    *first = new;

  *last = new;
}



//===========================================================
/// \fn	      atom_t copy_append( atom_t list, atom_t element)
/// \brief    creates a copy of list and append element to end
/// \param    element	atom to be appended
/// \param    list	list to be copied
///
/// Creates a copy of \c list and append \c element to it
//===========================================================
atom_t copy_append (atom_t list, atom_t element)
{
    atom_t result, iter;

    if (!IS_LIST(list))
    {
        return new_error(ERROR_NOT_A_LIST, list);
    }
    if (IS_EMPTY(list))
    {
        return new_list(USE(element), empty_list);
    }

    result = new_list(USE(CAR(list)), empty_list);
    iter = result;
    list = CDR(list);
    while (IS_NOT_EMPTY(list))
    {
        CDR(iter) = new_list(USE(CAR(list)), empty_list);
        iter = CDR(iter);
        list = CDR(list);
    }
    CDR(iter) = new_list(USE(element), empty_list);

    return result;
}


//===========================================================
/// \fn		atom_t behead( atom_t a )
/// \brief		removes the head of a list
/// \param   a	list to be beheaded
/// \return		the list without its head
///
/// Removes and deuses the head of a list and returns list's
/// buthead. This functionality is useful for descructive
/// iteration over a list.
//===========================================================
atom_t behead( atom_t a )
{
  atom_t buthead = CDR(a);
  CDR(a) = empty_list;
  DEUSE(a);
  return buthead;
}


//===========================================================
/// \fn		int list_length (atom_t a)
/// \brief		returns count of list elements
/// \param   a	list
/// \return		the number of elements of list
///
/// Returns number of elements in the list.
//===========================================================
int list_length (atom_t a)
{
    int res = 0;
    if (!IS_LIST(a))
    {
        return -1;
    }
    while (IS_NOT_EMPTY(a))
    {
        ++res;
        a = CDR(a);
    }
    return res;
}


//===========================================================
/// \fn		    atom_t list_copy_n (atom_t a, unsigned int n)
/// \brief		copies a list, but no more than n elents
/// \param   a	list
/// \param   n  max number of elements to be copied
/// \return		the new copy of the list
///
/// Copies list, but no more than \c n elements.
/// New list contains same data - only CDR-s are copied.
//===========================================================
atom_t list_copy_n (atom_t a, unsigned int n)
{
    atom_t result, iter;

    if (!IS_LIST(a))
    {
        return new_error(ERROR_NOT_A_LIST, a);
    }
    if (IS_EMPTY(a) || n == 0)
    {
        return empty_list;
    }

    result = new_list(USE(CAR(a)), empty_list);
    iter = result;
    a = CDR(a);
    while (IS_NOT_EMPTY(a) && --n)
    {
        CDR(iter) = new_list(USE(CAR(a)), empty_list);
        iter = CDR(iter);
        a = CDR(a);
    }

    return result;
}


//===========================================================
/// \fn		    atom_t list_copy_but_last (atom_t a)
/// \brief		copies a list, but without last element
/// \param   a	list
/// \return		the new copied list
///
/// Copies list, but without its last element. Linearly
//===========================================================
atom_t list_copy_but_last (atom_t a)
{
    atom_t result, iter;

    if (!IS_LIST(a))
    {
        return new_error(ERROR_NOT_A_LIST, a);
    }
    if (IS_EMPTY(a))
    {
        return new_error(ERROR_MISSING_VALUE, a);
    }
    if (IS_EMPTY(CDR(a)))
    {
        return empty_list;
    }

    result = new_list(USE(CAR(a)), empty_list);
    iter = result;
    a = CDR(a);
    while (IS_NOT_EMPTY(CDR(a)))
    {
        CDR(iter) = new_list(USE(CAR(a)), empty_list);
        iter = CDR(iter);
        a = CDR(a);
    }

    return result;
}


//===========================================================
/// \fn         atom_t get_at_list (atom_t a, int n)
/// \brief      returns element at specified position in list
/// \param  a   list
/// \param  n   position index
/// \return		element at position or last if position is -1
///
/// Returns element at specified position in the list.
/// If position is negative returns last element in the list
//===========================================================
atom_t get_at_list (atom_t a, int n)
{
    if (!IS_LIST(a))
    {
        return new_error(ERROR_NOT_A_LIST, a);
    }
    if (IS_EMPTY(a))
    {
        return new_error(ERROR_MISSING_VALUE, a);
    }
    if (n < 0)
    {
        while (IS_NOT_EMPTY(CDR(a)))
            a = CDR(a);
    }
    else
    {
        while (IS_NOT_EMPTY(a) && n--)
            a = CDR(a);
        if (IS_EMPTY(a))
            return new_error(ERROR_MISSING_VALUE, a);
    }
    return CAR(a);
}

