//
// Project:	Lhogho
// File:	mems.c
//
// Copyright (C) 2007 P.Boytchev
//
// Revision history:
//   2007-06-19	- file created from atoms.c
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//





#include <errno.h>	// errno, ENOMEM
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>     // malloc()

#include "atoms.h"
#include "pools.h"
#include "options.h"
#include "mems.h"
#include "unicode.h"
#include "lists.h"
#include "atoms.h"
#include "vars.h"
#include "compiler.h"
#include "numbers.h"



//===================================================
/// \fn		void delete_mem( atom_t a ) 
/// \brief	deletes mem atom
/// \param a	atom to delete
///
/// Deletes the mem atom and deallocates memory
/// used by it.
//===================================================
void delete_mem( atom_t a )
{
  DEUSE( ATOMS(a) );
  DEALLOC( MEMORY(a) ); 
  return_to_pool( &data_pool, a );
}




//===================================================
/// \fn 	atom_t new_mem( uint_t size)
/// \brief		creates a new mem atom
/// \param	size	size of memory to allocate
/// \return		mem atom
///
/// Creates a memory atom with reference count 1.
//===================================================
atom_t new_mem( uint_t size)
{
  atom_t a = take_from_pool( &data_pool );

  REF(a)      = 1;
  ID(a)       = MEM_ID;
  MEMORY(a)   = XALLOC( (size_t)size );
  ATOMS(a)    = empty_list;

  #ifdef SAFEMODE
    assert( MEMORY(a) );
  #endif //SAFEMODE

  #ifdef DEBUG_ATOM
    printf("<ATOM>  [%08x] mem={%d bytes}\n",(int)a,size);
  #endif //DEBUG_ATOM

  #ifdef ADVANCED
    stats[ID(a)].allocs++;
    if( stats[ID(a)].max<(stats[ID(a)].allocs-stats[ID(a)].deallocs) )
      stats[ID(a)].max=(stats[ID(a)].allocs-stats[ID(a)].deallocs);
    stats_free--;
  #endif //ADVANCED

  #ifdef DEBUG_RUNTIME_ATOMS
  if( running_compiled_code )
    {
      outter( TEXT("<RUNTIME> new  "), -1 );
      dump_atom_address( a );
      dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
  #endif
  #ifdef DEBUG_COMPILETIME_ATOMS
  if( compiling_code )
    {
      outter( TEXT("<COMPILETIME> new  "), -1 );
      dump_atom_address( a );
      dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
  #endif

  return a;
}




//===================================================
/// \fn		void dump_mem( atom_t a, int level ) 
/// \brief	dumps mem atom
/// \param a	atom to dump
/// \param level dump level
///
/// Dumps mem through the current outter function.
//===================================================
void dump_mem( atom_t a, int level )
{
  #define DUMP_BUF_SIZE 128
  char_t buf[DUMP_BUF_SIZE];
  int n;

  n = SPRINTF( buf, DUMP_BUF_SIZE, FORMAT_MEM, (int)MEMORY(a) );

  outter( buf, n );

  #undef DUMP_BUF_SIZE
}




