//
// Project:	Lhogho
// File:	numbers.c
//
// Copyright (C) 2007 P.Boytchev
//
// Revision history:
//   2007-06-09	- file created from atoms.c
//   2007-06-25 - atom_to_float()
//   2007-07-08 - fixed bug #1749858 INF and NaN in Windows
//   2007-08-13 - atom_to_int()
//   2007-11-05 - atom_to_integer()
//   2007-11-26 - fixed bug #1838504 User commands without inputs are miscompiled
//   2007-12-21 - fixed bug #1851865 Locale information not considered
//   2009-07-15 - fixed request #2804558 Correct display of integers>999999
//   2012-02-06 - Fixed bug #3484953 Printing power 10 200 crashes
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#undef __STRICT_ANSI__


#include <assert.h>	// assert()
#include <errno.h>	// errno, ENOMEM
#include <stdio.h>	// fopen(), fread(), fclose()
#include <stdlib.h>     // malloc()
#include <string.h>     // memcpy()
#include <locale.h>	// LC_ALL
#include <math.h>

#include "globals.h"
#include "compiler.h"

#ifdef UNICODE_CHARS
  #include <wchar.h>
#endif

#include "atoms.h"
#include "numbers.h"
#include "words.h"
#include "lists.h"
#include "errors.h"
#include "pools.h"
#include "unicode.h"
#include "translate.h"
#include "vars.h"
#include "parser.h"
#include "options.h"


#ifdef LINUX
#ifndef signbit
extern int signbit( float64_t x );
#endif
#endif


#define EPSILON 1E-64 ///< used for numerical comparison


//===================================================
/// \fn		void delete_numeric( atom_t a ) 
/// \brief	deletes numeric atom
/// \param a	atom to delete
///
/// Deletes integer or float atom by returning it
/// back to the data pool.
//===================================================
void delete_numeric( atom_t a )
{
  return_to_pool( &data_pool, a );
}




//===================================================
/// \fn 	atom_t new_integer( int64_t data ) 
/// \brief		creates a new integer atom
/// \param	data	64-bit integer
/// \return		integer atom
///
/// Creates an integer atom with reference count 1.
//===================================================
atom_t new_integer( int64_t data )
{
  atom_t a = (atom_t)take_from_pool( &data_pool );

  REF(a)     = 1;
  ID(a)	     = INTEGER_ID;
  INTEGER(a) = data;

  #ifdef DEBUG_ATOM
    printf("<ATOM>  [%08x] integer=%I64d\n",(int)a,data);
  #endif //DEBUG_ATOM

  #ifdef ADVANCED
    stats[ID(a)].allocs++;
    if( stats[ID(a)].max<(stats[ID(a)].allocs-stats[ID(a)].deallocs) )
      stats[ID(a)].max=(stats[ID(a)].allocs-stats[ID(a)].deallocs);
    stats_free--;
  #endif //ADVANCED

  #ifdef DEBUG_RUNTIME_ATOMS
  if( running_compiled_code )
    {
      outter( TEXT("<RUNTIME> new  "), -1 );
      dump_atom_address( a );
      dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
  #endif
  #ifdef DEBUG_COMPILETIME_ATOMS
  if( compiling_code )
    {
      outter( TEXT("<COMPILETIME> new  "), -1 );
      dump_atom_address( a );
      dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
  #endif

  return a;
}




//===================================================
/// \fn 	atom_t new_float( float64_t data ) 
/// \brief		creates a new floating-point atom
/// \param	data	64-bit floating-point
/// \return		float atom
///
/// Creates an floating-point atom with reference
/// count 1.
//===================================================
atom_t new_float( float64_t data )
{
  atom_t a = (atom_t)take_from_pool( &data_pool );

  REF(a)   = 1;
  ID(a)    = FLOAT_ID;
  FLOAT(a) = data;

//if( IS_FLOAT(a) )
//{
//printf("SET.FLT[%08x] %0.14f ref %d\n", (int)a, FLOAT(a), REF(a) );
//}

  #ifdef DEBUG_ATOM
    printf("<ATOM>  [%08x] float=%Lf\n",(int)a,data);
  #endif //DEBUG_ATOM

  #ifdef ADVANCED
    stats[ID(a)].allocs++;
    if( stats[ID(a)].max<(stats[ID(a)].allocs-stats[ID(a)].deallocs) )
      stats[ID(a)].max=(stats[ID(a)].allocs-stats[ID(a)].deallocs);
    stats_free--;
  #endif //ADVANCED

  #ifdef DEBUG_RUNTIME_ATOMS
  if( running_compiled_code )
    {
      outter( TEXT("<RUNTIME> new  "), -1 );
      dump_atom_address( a );
      dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
  #endif
  #ifdef DEBUG_COMPILETIME_ATOMS
  if( compiling_code )
    {
      outter( TEXT("<COMPILETIME> new  "), -1 );
      dump_atom_address( a );
      dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
  #endif

  return a;
}




//===================================================
/// \fn		void dump_integer( atom_t a, int level ) 
/// \brief	dumps integer atom
/// \param a	atom to dump
/// \param level dump level
///
/// Dumps integer atom through the current outter function.
//===================================================
void dump_integer( atom_t a, int level )
{
  if( IS_UNBOUND(a) )
    {
      outter( TEXT("(unbound)"), 9 );
      outter_size += 9;
    }
  else
  if( IS_STOPPED(a) )
    {
      outter( TEXT("(stopped)"), 9 );
      outter_size += 9;
    }
  else
    {
      #define DUMP_BUF_SIZE 128
      char_t buf[DUMP_BUF_SIZE];
      int n;

      n = SPRINTF( buf, DUMP_BUF_SIZE, FORMAT_INT, INTEGER(a) );

      int limit = print_width_limit;

      if( 0<=limit && limit<10 ) limit=10;

      if( 0<=limit && limit<n )
	{
	  outter( buf, limit );
	  outter( TEXT("..."), 3 );
	  outter_size += limit+3;
	}
      else
	{
	  outter( buf, n );
	  outter_size += n;
	}

      #undef DUMP_BUF_SIZE
    }
}




//===================================================
/// \fn		void dump_integer_const( int a ) 
/// \brief	dumps integer constant
/// \param a	value to dump
///
/// Dumps integer constant through the current outter function.
//===================================================
void dump_integer_const( int a )
{
  #define DUMP_BUF_SIZE 128
  char_t buf[DUMP_BUF_SIZE];
  int n;

  n = SPRINTF( buf, DUMP_BUF_SIZE, TEXT("%d"), a );
  outter( buf, n );

  #undef DUMP_BUF_SIZE
}


#if defined(DEBUG_RUNTIME_ATOMS) || defined(DEBUG_COMPILETIME_ATOMS)
//===================================================
/// \fn		void dump_atom_address( atom_t a ) 
/// \brief	dumps atom address in hex
/// \param a	value to dump
///
/// Dumps atom address in hex through the current outter function.
//===================================================
void dump_atom_address( atom_t a )
{
  #define DUMP_BUF_SIZE 128
  char_t buf[DUMP_BUF_SIZE];
  int n;

  n = SPRINTF( buf, DUMP_BUF_SIZE, TEXT(" [%x]"), (int)a );
  outter( buf, n );
  if( REF(a)>300 )
    outter( TEXT(":?"), 2 );
  else
    {
      n = SPRINTF( buf, DUMP_BUF_SIZE, TEXT(":%d = "), REF(a) );
      outter( buf, n );
    }

  #undef DUMP_BUF_SIZE
}
#endif //DEBUG_RUNTIME_ATOMS || DEBUG_COMPILETIME_ATOMS



//===================================================
/// \fn		void dump_float( atom_t a, int level ) 
/// \brief	dumps float atom
/// \param a	atom to dump
/// \param level dump level
///
/// Dumps float atom through the current outter function.
//===================================================
void dump_float( atom_t a, int level )
{
  // In Window infinities are printed as "1.#INF" instead
  // of just "inf", and "-1.#IND" instead of "nan". So let's fix this.
  // In some Linuxes sqrt(-1) is "nan" in others it is "-nan".
  // Let's make "-nan" be "nan"
  #ifdef WINDOWS
  if( isinf(FLOAT(a))!=0 && FLOAT(a)>0 )
    {
      outter( TEXT("inf"), 3 );
      outter_size += 3;
      return;
    }

  if( isinf(FLOAT(a))!=0 && FLOAT(a)<0 )
    {
      outter( TEXT("-inf"), 4 );
      outter_size += 4;
      return;
    }
  #endif //WINDOWS

  // Convert "-nan" to "nan"
  if( isnan(FLOAT(a))!=0 )
   {
      outter( TEXT("nan"), 3 );
      outter_size += 3;
      return;
    }

  #define DUMP_BUF_SIZE 128
  char_t buf[DUMP_BUF_SIZE];
  int n;
  float64_t x = FLOAT(a);
  if( x==0 && signbit(x) ) x = 0;

  // remove trailing zeros and decimal point
  if( (x>=1e23) || (x<=-1e23) || ((x>=-1e-23) && (x<=1e-23)) )
    n = SPRINTF( buf, DUMP_BUF_SIZE, FORMAT_EXTRA_FP, x );
  else
    {
      n = SPRINTF( buf, DUMP_BUF_SIZE, FORMAT_FP, x );
	  
      //if( !buf[n-1] ) n--;	// remove trailing null character (should not appear, but it does!)
      while( buf[n-1]==L'0' ) n--;	// remove trailing zero
      if( buf[n-1]==L'.' ) n--;	// remove default decimal point
      if( (unsigned char)buf[n-1]<128
          && buf[n-1]==*(locale_info->decimal_point) ) n--;
    }


  int limit = print_width_limit;

  if( 0<=limit && limit<10 ) limit=10;

  if( 0<=limit && limit<n )
    {
      outter( buf, limit );
      outter( TEXT("..."), 3 );
      outter_size += limit+3;
    }
  else
    {
      outter( buf, n );
      outter_size += n;
    }


  #undef DUMP_BUF_SIZE
}




//===================================================
/// \fn		int atom_to_float( atom_t a, float64_t* np )
/// \brief	convert atom to a float
/// \param a	atom to convert
/// \param np	pointer where to put the number
/// \return	returns 1 if conversion was successful.
///
/// Converts an atom to a floating point number. If
/// conversion is possible, then *np will contain the
/// number and the return value will be 1. Otherwise
/// *np will be uninitialized and the return value
/// will be 0.
//===================================================
int atom_to_float( atom_t a, float64_t* np )
{
  // test for unbound
  if( IS_UNBOUND(a) )
    {
      return 0;
    }

  // convert integer->float
  if( IS_INTEGER(a) ) 
    {
      *np = INTEGER(a);
      return 1;
    }

  // convert float->float
  if( IS_FLOAT(a) ) 
    {
      *np = FLOAT(a);
      return 1;
    }

  // fail if not a word
  if( !IS_ANY_WORD(a) )
    {
      return 0;
    }

  // convert word->float
  int     cnt = LENGTH(a);
  chars_t ptr = STRING(a);
  char_t  chr = *(ptr+cnt);
  char_t  *end;
  *(ptr+cnt) = TEXT('\0');	/* set artificial end of string */

  if( (*ptr==TEXT('0')) && ((*(ptr+1)==TEXT('x')) || (*(ptr+1)==TEXT('X'))))
    { // it is a hex number
      *np = STRTOL(ptr,&end);
    }
  else
    { // it is a dec number
      *np = STRTOD(ptr,&end);
    }
  *(ptr+cnt) = chr;		/* recover the original string  */

  // test for successful conversion
  if( end-ptr==cnt ) return 1;


  // test for infinities/nonanumbers in Windows
#ifdef WINDOWS
  if( LENGTH(a)==3 &&
      same_strings(1,STRING(a),TEXT("inf"),3) )
    {
      *np = INFINITY;
      return 1;
    }
  if( LENGTH(a)==4 &&
      same_strings(1,STRING(a),TEXT("-inf"),4) )
    {
      *np = -INFINITY;
      return 1;
    }
  if( LENGTH(a)==3 &&
      same_strings(1,STRING(a),TEXT("nan"),3) )
    {
      *np = NAN;
      return 1;
    }
#endif //WINDOWS

  return 0;
}




//===================================================
/// \fn		int atom_to_int( atom_t a, int64_t* np )
/// \brief	convert atom to an integer
/// \param a	atom to convert
/// \param np	pointer where to put the number
/// \return	returns 1 if conversion was successful.
///
/// Converts an atom to a 64-bit integer number. If
/// conversion is possible, then *np will contain the
/// number and the return value will be 1. Otherwise
/// *np is uninitialized and the return value will be
/// set to 0.
//===================================================
int atom_to_int( atom_t a, int64_t* np )
{
  // test for unbound
  if( IS_UNBOUND(a) )
    {
      return 0;
    }

  // convert integer->float
  if( IS_INTEGER(a) ) 
    {
      *np = INTEGER(a);
      return 1;
    }

  // convert float->float
  if( IS_FLOAT(a) ) 
    {
      *np = FLOAT(a);
      return abs(*np-FLOAT(a))<EPSILON;
    }

  // fail if not a word
  if( !IS_ANY_WORD(a) )
    {
      return 0;
    }

  // convert word->float
  int     cnt = LENGTH(a);
  chars_t ptr = STRING(a);
  char_t  chr = *(ptr+cnt);
  char_t  *end;
  *(ptr+cnt) = TEXT('\0');	/* set artificial end of string */
  *np = STRTOL(ptr,&end);
  *(ptr+cnt) = chr;		/* recover the original string  */

  // test for successful conversion
  if( end-ptr==cnt ) return 1;

  return 0;
}



//===================================================
/// \fn		int atom_to_integer( atom_t a, int* np )
/// \brief	convert atom to an 32-bit integer
/// \param a	atom to convert
/// \param np	pointer where to put the number
/// \return	returns 1 if conversion was successful.
///
/// Converts an atom to a 32-bit integer number. If
/// conversion is possible, then *np will contain the
/// number and the return value will be 1. Otherwise
/// *np will be set to -1 and the return value will
/// be set to 0.
//===================================================
int atom_to_integer( atom_t a, int* np )
{
  *np = -1;

  // test for unbound
  if( IS_UNBOUND(a) )
    {
      return 0;
    }

  // convert integer->float
  if( IS_INTEGER(a) ) 
    {
      *np = INTEGER(a);
      return 1;
    }

  // convert float->float
  if( IS_FLOAT(a) ) 
    {
      *np = FLOAT(a);
      return abs(*np-FLOAT(a))<EPSILON;
    }

  // fail if not a word
  if( !IS_ANY_WORD(a) )
    {
      return 0;
    }

  // convert word->float
  int     cnt = LENGTH(a);
  chars_t ptr = STRING(a);
  char_t  chr = *(ptr+cnt);
  char_t  *end;
  *(ptr+cnt) = TEXT('\0');	/* set artificial end of string */
  *np = STRTOL(ptr,&end);
  *(ptr+cnt) = chr;		/* recover the original string  */

  // test for successful conversion
  if( end-ptr==cnt ) return 1;

  return 0;
}
