//
// Project:	Lhogho
// File:	options.c
//
// Copyright (C) 2007 P.Boytchev
//
// Revision history:
//   2007-05-31	- file created
//   2007-06-01 - case insensitive option
//   2007-06-02 - option_source_filename
//              - option_memory_statistics
//              - use ADVANCED define to filter options
//   2007-06-03 - errors renamed
//              - ERROR_CROWDED_SOURCES
//              - option_variables
//   2007-06-07 - definitions spread to where they belong
//   2007-06-09 - fixed bug #1734091 "Bad compiler option is misinterpreted"
//   2007-06-11 - added option -Zuv (-Zuser-variables)
//          		- added option -x (--executable)
//   2007-06-13	- fixed bug #1736021 "Alloc/dealloc statistics"
//   2007-06-19 - added mem atoms
//          		- added XALLOC() and xalloc()
//   2007-06-21 - OPTION_ASM, OPTION_ASM_LONG
//   2007-07-09 - OPTION_RUNTIME, OPTION_RUNTIME_LONG
//   2007-07-11 - OPTION_TRADITIONAL, OPTION_TRADITIONAL_LONG
//   2007-07-29 - set_option()
//   2007-09-23 - OPTION_MEMORY_ALL, OPTION_MEMORY_ALL_LONG,
//		            options -Zmad, --Zmemory-all-details
//   2007-10-06 - added REALLOC() and realloc()
//   2009-05-18 - OPTION_EXECCOMP, OPTION_EXECCOMP_LONG
//   2010-02-07 - Fixed bug #3174945 OS alloc balance is not 0
//          		- Fixed bug #3174840 Invalid options cause a crash
//   2011-12-02 - Added notification of out-of-memory situations
//   2011-12-26 - Error messages taken from string, stored to array of atoms
//   2012-01-02 - Added output_compiler_name()
//                Command-line options are translatable
//   2012-10-14 - Fixed bug #314 -x compile on Windows needs lhogho.exe next to .lgo source file
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#undef __STRICT_ANSI__


#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <sys/stat.h> // stat, S_ISDIR, ...
#include "globals.h"
#include "atoms.h"
#include "words.h"
#include "lists.h"
#include "options.h"
#include "parser.h"
#include "translate.h"
#include "unicode.h"
#include "errors.h"
#include "vars.h"
#include "compiler.h"

#ifdef LINUX
#include <sys/mman.h> // for mprotect()
#endif

#define CHECK(PARAM1,PARAM2,BODY)	 \
  if( strcmp(*argv,PARAM1)==0 || strcmp(*argv,PARAM2)==0 ) \
  BODY; ///< check parameter and execute code if it matches

int    option_turned_on[OPTIONS_COUNT]; ///< Values for options
atom_t option_texts[4][OPTIONS_COUNT];  ///< Texts for short options [0]-short [1]-long [2]-alternative [3]-description

char*	option_source_filename;		        ///< ASCII name of current source file
chars_t option_source_filename_chars;	  ///< name of current source file
char*	option_compiler_filename;	        ///< ASCII name of current compiler
chars_t	option_compiler_filename_chars;	///< name of current compiler

int standalone;		                      ///< indicate using of standalone compiler (i.e. -x option)
atom_t standalone_arguments;            ///< list of command-line arguments

#ifdef ADVANCED
char* id_names[MAX_ID] = { 
  "Integers",
  "Floats",
  "Lists",
  "Words",
  "Subwords",
  "Errors",
  "Var",
  "Memory" }; ///< array of names for each atom type

int	  stats_free;	///< statistics for free atoms (of any type)
int     stats_allocs; ///< statistics for usage of ALLOC and DEALLOC
stats_t stats[MAX_ID];///< statistics for each type of atoms
#endif //ADVANCED




//===================================================
/// \fn		void init_options( )
/// \brief		initializes the Options module
///
/// Initalizes options with their default values.
//===================================================
void init_options( int argc, char*argv[] )
{
  // clear string options
  option_source_filename = NULL;
  option_source_filename_chars = NULL;

  option_compiler_filename = NULL;
  option_compiler_filename_chars = NULL;

  standalone = 0;
  standalone_arguments = empty_list;
  
  // load options texts
  atom_t names = new_word( TR_OPTIONS, UNKNOWN );
  atom_t tokens = tokenize( names, TOKENIZE_DATA );
  atom_t t = tokens; 

  // create options texts
  int i;
  for( i = 0; i<OPTIONS_COUNT; i++ )
  {
    option_turned_on  [i] = 0;
    int j;
    for( j=0; j<4; j++ )
    {
      #ifdef SAFEMODE
      assert( IS_NOT_EMPTY(t) );            // too few words in TR_OPTIONS
      #endif
      option_texts[j][i] = USE(CAR(t));
      t=CDR(t);
    }
  }

  #ifdef SAFEMODE
  assert( IS_EMPTY(t) ); // too many words in TR_OPTIONS
  #endif

  DEUSE( names );
  DEUSE( tokens );
}




//===================================================
/// \fn		void dump_statistics( )
/// \brief	dump statistics
///
/// Dumps statistics of memory usage
//===================================================
#ifdef ADVANCED
void dump_statistics( )
{
  int i;
  int stats_used=0;
  PRINT("%s\n","");
  PRINT("   %10s %8s %8s %8s %8s\n","Atom type","Max","Allocs","Deallocs","Balance");
  PRINT("   %10s %8s %8s %8s %8s\n","----------","--------","--------","--------","--------");
  for( i=MIN_ID; i<MAX_ID; i++ )
  {
    stats_used += stats[i].allocs-stats[i].deallocs;
    if( stats[i].allocs )
    {
      PRINT("   %10s %8d %8d %8d %8d\n",
      id_names[i],
      stats[i].max,
      stats[i].allocs,
      stats[i].deallocs,
      stats[i].allocs-stats[i].deallocs );
    }
  }
  PRINT(" %39s\n","----------------");
  PRINT(" %30s %8d\n","Used",stats_used);
  PRINT(" %30s %8d\n","Free",stats_free);
  PRINT(" %39s\n","----------------");
  PRINT(" %30s %8d\n","OS alloc balance",stats_allocs);
}
#endif

//===================================================
/// \fn		void finit_options( )
/// \brief	finalizes the Options module
///
/// Dumps statistics if -Zm option was switched on
//===================================================
void finit_options( )
{
  int i;
  for( i = 0; i<OPTIONS_COUNT; i++ )
  {
    int j;
    for( j = 0; j<4; j++ )
    {
      DEUSE(option_texts[j][i]);
    }
  }
  DEUSE( standalone_arguments );

  DEALLOC( option_compiler_filename );
  DEALLOC( option_compiler_filename_chars );

  if( option_source_filename_chars )
  {
    DEALLOC( option_source_filename_chars );
  }

#ifdef ADVANCED
  int stats_used=0;
  if( OPTION_DETAILED_MEMORY_STATISTICS )
  {
    dump_statistics();
  }

  if( OPTION_MEMORY_STATISTICS )
  {
    for( i=MIN_ID; i<MAX_ID; i++ )
    stats_used += stats[i].allocs-stats[i].deallocs;
    PRINT("{MEM#%d:%d}\n",stats_used,stats_allocs);
  }
#endif //ADVANCED
}



//===================================================
/// \fn		void output_compiler_name( int show_help )
/// \brief		outputs the name of the compier
/// \param show_help if set also output commandline options
///
/// Outputs the name of the compiler, its language,
/// its platform and build date. If \c show_help is
/// not zero, output a list of command-line options
//===================================================
void output_compiler_name( int show_help )
{

  atom_t texts = new_word( TR_HELP_TEXTS, UNKNOWN );
  atom_t help_texts = tokenize( texts, TOKENIZE_DATA );
  atom_t t = help_texts;
  
  dump( CAR(t) ); t = CDR(t);       // LHOGHO - The LOGO Compiler
  outter( TEXT(" [" VERSION_STR ", " OS_NAME "-" PROCESSOR_NAME "(" LANG_STR "), " __DATE__ "]\n\n"), -1 );

  if( show_help )
  {
    dump( CAR(t) ); t = CDR(t);       // Usage
    
    int i;
    for( i=0; i<OPTIONS_COUNT; i++ )
    {
      int len;

      if( i==FIRST_ADVANCED_OPTION )
      {
        #ifndef ADVANCED
          t = CDR(t);
        #endif
        dumpln( CAR(t) ); t = CDR(t);       // options
      }
      
      // show short version
      len = LENGTH(option_texts[0][i]);
      outter( TEXT("  "), -1 );
      dump( option_texts[0][i] );
      for( ; len<6; len++ ) outter( TEXT(" "), -1 );

      // show long version
      len = LENGTH(option_texts[2][i]);
      if( len )
      {
        dump( option_texts[2][i] );
      }
      else
      {
        len = LENGTH(option_texts[1][i]);
        dump( option_texts[1][i] );
      }
      for( ; len<30; len++ ) outter( TEXT(" "), -1 );

      // show description      
      dumpln( option_texts[3][i] );
    }
  }
  
  DEUSE( texts );
  DEUSE( help_texts );
}



//===================================================
/// \fn		void set_compiler_filename( int argc, char*argv[] )
/// \brief		sets the full name of the compiler
/// \param argc	\c argc parameter from the \c main function
/// \param argv	\c argv parameter from the \c main function
///
/// If the compiler is executed through a search
/// in the PATH, then try to find the full name
//===================================================
void set_compiler_filename( int argc, char*argv[] )
{
  // get the name of the compiler (append extension if needed)
  option_compiler_filename = (char*)ALLOC( strlen(*argv)+strlen(EXE_EXT) );
  strcpy( option_compiler_filename, *argv );

  char* extension = option_compiler_filename+strlen(*argv);
  if( strcasecmp(extension-strlen(EXE_EXT),EXE_EXT)==0 )
	strcpy( extension-strlen(EXE_EXT), EXE_EXT );
  else
	strcpy( extension, EXE_EXT );

  //printf("compiler filename='%s'\n",option_compiler_filename);
  //printf("search path='%s'\n",getenv("PATH"));
  
  struct stat buffer;
  int err;

#ifdef WINDOWS
	// In WINDOWS the current folder is always checked first
	// So, before checking PATH, check the current folder
	err = stat( option_compiler_filename, &buffer );
	//printf( "full name(err=%d)='%s'\n", err, option_compiler_filename );
	if( !err ) goto ready;
#endif

  char* envpath;
  envpath = strdup( getenv("PATH") );
  char* path = strtok(envpath,PATH_DELIM);
  
  while( path )
  {
	//printf( "dir='%s'\n", path );
	char filename[strlen(path)+1+strlen(option_compiler_filename)+10];
	strcpy( filename, path );
	strcpy( filename+strlen(path), PATH_SLASH );
	strcpy( filename+strlen(path)+1, option_compiler_filename );
	strcpy( filename+strlen(path)+1+strlen(option_compiler_filename), "\0" );

	err = stat( filename, &buffer );
	//printf( "full name(err=%d)='%s'\n", err, filename );
	if( !err )
	{
		// We found where the file is loaded from, remember it
		DEALLOC( option_compiler_filename );
		option_compiler_filename = (char*)ALLOC( strlen(filename)+1 );
		strcpy( option_compiler_filename, filename );
		goto free_and_ready;
	}
	
	path = strtok( NULL, PATH_DELIM );
  }
  
free_and_ready:
  free( envpath );

#ifdef WINDOWS
ready:
#endif
  option_compiler_filename_chars = UNFILENAME(option_compiler_filename);
}

//===================================================
/// \fn		void set_options( int argc, char*argv[] )
/// \brief		sets options from command-line parameters
/// \param argc	\c argc parameter from the \c main function
/// \param argv	\c argv parameter from the \c main function
///
/// Reads options from the \c argv array and sets
/// appropriate parameters.
//===================================================
void set_options( int argc, char*argv[] )
{
  atom_t last = empty_list;

  set_compiler_filename( argc, argv );
  
  // get standalone mode
  int   ptr;
  unsigned char* code = load_file( option_compiler_filename_chars, &ptr );
  if( !code )
  {
    //atom_t error = new_os_error( option_compiler_filename_chars );
	return; 
  }

  standalone = (*(int*)(code+ptr-4)==MAGIC_NUMBER);
  DEALLOC( code );

  // get options only if not in standalone mode,
  // otherwise collect them in a list
  if( standalone )
  { // standalone mode -- collect arguments in a list
    for( argc--,argv++; argc--; argv++ )
    {
      atom_t word = decode_word( (unsigned char*)*argv, strlen(*argv), 0 );
      append( word, &standalone_arguments, &last );
    }
  }
  else
  { // compiler mode -- process options till file name and then collect arguments
    for( argc--,argv++; argc--; argv++ )
    {
      if( option_source_filename_chars )
      {
        atom_t word = decode_word( (unsigned char*)*argv, strlen(*argv), 0 );
        append ( word, &standalone_arguments, &last );
      }
      else
      {
        int found = 0;
        chars_t option = UNFILENAME(*argv);
        int len = STRLEN(option);
        int i;
        for( i=0; i<OPTIONS_SUPPORTED_COUNT; i++ )
        {
          int j;
          for( j=0; j<4; j++ )
          {
            if( len==LENGTH(option_texts[j][i]) )
            {
              if( same_strings(1,option,STRING(option_texts[j][i]),len) )
              {
                found = 1;
                option_turned_on[i] = 1;
                break;
              }
            }
          }
        }
        DEALLOC( option );

        if( OPTION_HELP )
        {
          output_compiler_name( 1 );
          exit(0);
        }

        if( found ) continue;
        
        // check for invalid option or second file name
        if( **argv=='-'  )
        {
          chars_t name = UNFILENAME(*argv); // has ALLOC inside
          atom_t word = new_word( name, UNKNOWN );
          atom_t err = new_error( ERROR_UNKNOWN_OPTION, word );
          dumpln( err );
          DEUSE( word );
          DEALLOC( name );
          continue;
        }

        // set file name
        option_source_filename = *argv;
        option_source_filename_chars = UNFILENAME(*argv);
      }
    } // for
  }
}




//===================================================
/// \fn		void* alloc( int size )
/// \brief		allocates dynamic memory
/// \param	size	size of memory (in bytes) to allocate
/// \return		pointer to allocated memory
///
/// This function is used to allocate memory.
/// Depending on the existence of DEBUG_HEAP symbol
/// it either just allocates memory or allocates and
/// dump on the screen allocation data.
///
///
/// \fn		void dealloc( void* ptr )
/// \brief		deallocates dynamic memory
/// \param	ptr	pointer to the  memory to deallocate
///
/// This function is used to deallocate memory.
/// Depending on the existence of DEBUG_HEAP symbol
/// it either just deallocates memory or deallocates
/// and dumps on the screen deallocation data.
//===================================================
#ifdef DEBUG_HEAP
void* alloc( int size, const char* func, const char* file, const int line )
{
  void* buffer = malloc( size );
  printf( "<HEAP> +[%08x] %d bytes %s(%s:%d)\n",(int)buffer,size,func,file,line );
  if( !buffer ) { dumpln(error_texts[ERROR_OUT_OF_MEM]); exit(1); }
  #ifdef ADVANCED
  stats_allocs++;
  #endif
  return buffer;
}

void* reallocate( void * mem, int size, const char* func, const char* file, const int line )
{
  #ifdef ADVANCED
  if( !mem ) stats_allocs++;
  #endif
  void* buffer = realloc (mem, size);

  printf( "<HEAP> +[%08x] extended with %d bytes %s(%s:%d)\n", (int)buffer, size, func, file, line);
  if( !buffer ) { dumpln(error_texts[ERROR_OUT_OF_MEM]); exit(1); }

  return buffer;
}

void* xalloc( int size, const char* func, const char* file, const int line )
{
  #if defined(WINDOWS) || defined(APPLE)
  void* buffer = malloc( size );
  #else
  void* buffer = valloc( size );
  if( buffer ) {mprotect( buffer, size, PROT_READ | PROT_WRITE | PROT_EXEC );}
  #endif

  printf( "<HEAP> +[%08x] %d bytes %s(%s:%d)\n",(int)buffer,size,func,file,line );
  if( !buffer ) { dumpln(error_texts[ERROR_OUT_OF_MEM]); exit(1); }
  
  #ifdef ADVANCED
  stats_allocs++;
  #endif
  return buffer;
}
void dealloc( void* ptr, const char* func, const char* file, const int line )
{
  printf( "<HEAP> -[%08x] %s(%s:%d)\n",(int)ptr,func,file,line );
  #ifdef ADVANCED
  stats_allocs--;
  #endif
  free( ptr );
}
#else //DEBUG_HEAP
void* alloc( int size )
{
  void* buffer = malloc( size );
  if( !buffer ) { dumpln(error_texts[ERROR_OUT_OF_MEM]); exit(1); }
  #ifdef ADVANCED
  stats_allocs++;
  #endif
  return buffer;
}

void* reallocate(void * mem, int size)
{
  #ifdef ADVANCED
  if( !mem ) stats_allocs++;
  #endif
  void* buffer = realloc(mem, size);
  if( !buffer ) { dumpln(error_texts[ERROR_OUT_OF_MEM]); exit(1); }
  return buffer;
}

void* xalloc( int size )
{
  #if defined(WINDOWS) || defined(APPLE)
  void* buffer = malloc( size );
  #else
  void* buffer = valloc( size );
  mprotect( buffer, size, PROT_READ | PROT_WRITE | PROT_EXEC );
  #endif

  if( !buffer ) { dumpln(error_texts[ERROR_OUT_OF_MEM]); exit(1); }

  #ifdef ADVANCED
  stats_allocs++;
  #endif
  return buffer;
}
void dealloc( void* ptr )
{
  #ifdef ADVANCED
  stats_allocs--;
  #endif
  free( ptr );
}
#endif //DEBUG_HEAP

