//
// Project:	Lhogho
// File:	options.h
//
// Copyright (C) 2007 P.Boytchev
//
// Revision history:
//   2007-05-31	- file created
//   2007-06-01 - case insensitive option
//   2007-06-02 - option_source_filename
//		- OPTION_MEMORY, OPTION_MEMORY_LONG
//		- option_variables
//   2007-06-07 - definitions spread to where they belong
//   2007-06-11 - added option -Zuv (--Zuser-variables)
//		- added option -x (--executable)
//   2007-06-19 - added XALLOC() and xalloc()
//   2007-06-21 - OPTION_ASM, OPTION_ASM_LONG
//   2007-07-09 - OPTION_RUNTIME, OPTION_RUNTIME_LONG
//   2007-07-11 - OPTION_TRADITIONAL, OPTION_TRADITIONAL_LONG
//   2007-07-29 - set_options()
//   2007-09-23 - OPTION_MEMORY_STAT, OPTION_MEMORY_STAT_LONG
//   2007-10-06 - added REALLOC() and realloc()
//   2009-05-18 - OPTION_EXECCOMP, OPTION_EXECCOMP_LONG
//   2012-01-02 - Command-line options are translatable
//	 2012-10-14 - Adding PATH_SLASH, PATH_DELIM
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef LHOGHO_OPTIONS_H
#define LHOGHO_OPTIONS_H

#include "atoms.h"




// ===========================================================
/// \name Names of compiler options
// ===========================================================
// @{
#define OPTION_HELP                       option_turned_on[0]  ///< option "-h"  and "--help"                print this help message
#define OPTION_MAKE_EXECUTABLE            option_turned_on[1]  ///< option "-x"  and "--executable"          create executable file
#define OPTION_MAKE_EXECUTABLE_COMPILER   option_turned_on[2]  ///< option	"-xc" and "--executable-compiler" generate executable compiler
#define OPTION_CASE_INSENSITIVE           option_turned_on[3]  ///< option "-ci" and "--case-insensitive"    switch to case-insensitive mode
#define OPTION_TRADITIONAL                option_turned_on[4]  ///< option "-t"  and "--traditional"         switch to traditional mode

#define OPTION_ASSEMBLER                  option_turned_on[5]  ///< option "-Za"   and "--Zassembler"          dump assembler code
#define OPTION_MEMORY_STATISTICS          option_turned_on[6]  ///< option "-Zm"   and "--Zmemory"             dump short memory statistics
#define OPTION_DETAILED_MEMORY_STATISTICS option_turned_on[7]  ///< option "-Zmad" and "--Zmemory-all-details" dump detailed memory statistics
#define OPTION_RUNTIME                    option_turned_on[8]  ///< option "-Zrt"  and "--Zrun-time"	          dump evaluations at runtime
#define OPTION_DUMP_AST                   option_turned_on[9]  ///< option "-Zt"   and "--Ztree"	              dump abstract syntax trees
#define OPTION_VARIABLES                  option_turned_on[10] ///< option "-Zv"   and "--Zvariables"          dump created variables
#define OPTION_USER_VARIABLES             option_turned_on[11] ///< option "-Zuv"  and "--Zuser-variables"     dump user variables

#define FIRST_BASIC_OPTION          0
#define LAST_BASIC_OPTION           4

#define FIRST_ADVANCED_OPTION       5
#define LAST_ADVANCED_OPTION        11

#define OPTIONS_COUNT (LAST_ADVANCED_OPTION-FIRST_BASIC_OPTION+1)

#ifdef ADVANCED
  #define OPTIONS_SUPPORTED_COUNT OPTIONS_COUNT
#else
  #define OPTIONS_SUPPORTED_COUNT (LAST_BASIC_OPTION-FIRST_BASIC_OPTION+1)
#endif

// @}


extern int    option_turned_on  [OPTIONS_COUNT];  ///< Values for options


#ifdef WINDOWS
  #define EXE_EXT		".exe"	///< default extension for executable files
  #define PATH_SLASH	"\\"	///< default slash in paths
  #define PATH_DELIM	";"		///< default delimiter in paths in environment variables (like PATH)
#else
  #define EXE_EXT		""		///< default extension for executable files
  #define PATH_SLASH	"/"		///< default slash in paths
  #define PATH_DELIM	":"		///< default delimiter in paths in environment variables (like PATH)
#endif
#define EXE_EXT2 ".run"		///< alternative extension for executable files used to avoid overwriting the source file




extern int option_dump_ast;
extern int option_make_executable;
extern int option_make_executable_compiler;
extern int option_case_insensitive;
extern int option_traditional;
extern int option_memory_statistics;
extern int option_detailed_memory_statistics;
extern int option_variables;
extern int option_user_variables;
extern int option_assembler;
extern int option_runtime;

extern char*	option_compiler_filename;
extern chars_t	option_compiler_filename_chars;
extern char*	option_source_filename;
extern chars_t	option_source_filename_chars;

extern int standalone;
extern atom_t standalone_arguments;

extern void init_options( );
extern void finit_options( );
extern void set_options( int argc, char*argv[] );
extern void output_compiler_name( int show_help );

#ifdef ADVANCED
extern void dump_statistics( );
#endif


// ===========================================================
/// \struct		stats_t atoms.c
/// \brief		statistics structure
///
/// This structure contains statistics for allocation and
/// deallocation of a given atom type. It is used as an element
/// of the ::stats[] array.
// ===========================================================
#ifdef ADVANCED
typedef struct {
  int allocs;	///< total number of allocations
  int deallocs;	///< total number of deallocations
  int max;	///< MAX(allocs-deallocs)
} stats_t;

extern int     stats_allocs;
extern stats_t stats[MAX_ID];///< statistics for each type of atoms
extern void dump_statistics( );
#endif




// ===========================================================
/// \name Memory tracing
// ===========================================================
#ifdef DEBUG_HEAP
  #define ALLOC(size)	alloc(   size,		\
			         __FUNCTION__,	\
			         __FILE__,	\
			         __LINE__) ///< allocate memory

  #define REALLOC(ptr, size) reallocate(ptr, \
                   size,                     \
                   __FUNCTION__,             \
                   __FILE__,                 \
                   __LINE__) ///< reallocate memory

  #define XALLOC(size)	xalloc(   size,		\
			         __FUNCTION__,	\
			         __FILE__,	\
			         __LINE__) ///< allocate executable memory

  #define DEALLOC(ptr)	dealloc( ptr,		\
				 __FUNCTION__,	\
				 __FILE__,	\
				 __LINE__) ///< deallocate memory


  extern void* alloc(   int  size, const char* func, const char* file, const int line );
  extern void* reallocate( void* ptr, int size, const char* func, const char* file, const int line );
  extern void* xalloc(   int  size, const char* func, const char* file, const int line );
  extern void  dealloc( void* ptr, const char* func, const char* file, const int line );
#else
  #define ALLOC(size)	alloc(size)             ///< allocate memory
  #define REALLOC(ptr, size) reallocate(ptr, \
                                        size)   ///< reallocate memory
  #define XALLOC(size)	xalloc(size)            ///< allocate executable memory
  #define DEALLOC(ptr)	dealloc(ptr)            ///< deallocate memory

  extern void* alloc( int size );
  extern void* reallocate( void* ptr, int size);
  extern void* xalloc( int size );
  extern void  dealloc( void* ptr );
#endif //DEBUG_HEAP
#endif //LHOGHO_OPTIONS_H



