//
// Project:	Lhogho
// File:	pools.c
//
// Copyright (C) 2007 P.Boytchev
//
// Revision history:
//   2006-09-29	- Pools module created
//   2006-09-30	- Separated free atoms from blocks
//   2006-10-10	- More detailed statistics
//   2007-02-27	- Module renamed from \b Pool to \b Pools
//   2007-05-17	- Added license info
//   2007-05-21	- Added doxygen-friendly documentation
//   2007-06-02 - STATISTICS merged into ADVANCED
//   2007-06-05 - Simplified interface
//   2007-06-05 - definitions spread to where they belong
//   2007-06-13	- fixed bug #1736021 "Alloc/dealloc statistics"
//   2007-10-09 - fixed bug #1806186 Atoms no returned to the pool
//   2007-12-12 - dump_pool()
//   2008-01-18 - data_pool_ex
//   2008-01-23 - Fixed bug #1836433 Example with IFTRUE fails
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//




#include <stdlib.h>
#include <stdio.h>
#include <assert.h>	// assert()
#include "globals.h"
#include "pools.h"
#include "atoms.h"
#include "lists.h"
#include "words.h"
#include "options.h"
#include "compiler.h"
#include "unicode.h"
#include "atoms.h"
#include "numbers.h"

#define DEFAULT_BLOCK_SIZE	8192

pool_t data_pool;	///< pool for data atoms
pool_t data_pool_ex;	///< pool for extended data atoms


typedef struct {
  void* next; ///< next free atom in the chain
} free_atom_t;



//===================================================
/// \fn void init_pool( pool_t* pool, int atom_size ) 
/// \brief		initializes a pool
/// \param pool		pool to initialize
/// \param atom_size	size of atom in bytes
///
/// Initializes a pool by setting size of atoms and
/// maximal size of the pool. Initially the pool has
/// no any allocated atoms. Aom will be	allocated
/// during the first request to take an atom.
//===================================================
void init_pool( pool_t* pool, int atom_size )
{
  static int id = 0;
  pool->id		 = ++id;
  pool->atom_size	 = atom_size;
  pool->block_size	 = DEFAULT_BLOCK_SIZE;
  pool->free_block_atoms = 0;
  pool->free_atom	 = NULL;
  pool->free_block	 = NULL;
  pool->first_atom	 = NULL;

  #ifdef DEBUG_POOL
    printf( "<POOL> Create pool %d (size=%d)\n",
	    pool->id, atom_size );
  #endif //DEBUG_POOL
}




//===================================================
/// \fn	void expand_pool( pool_t* pool ) 
/// \brief		expands a pool when requested
/// \param pool		pool to be expanded
///
/// Assumes that both the list and the block of free
/// free atoms are empty and allocates a new block.
//===================================================
void expand_pool( pool_t* pool )
{
  pool->first_atom = ALLOC( pool->atom_size * pool->block_size );
  pool->free_block = pool->first_atom;

#ifdef DEBUG_CLEAR_FREED_MEM
  memset( pool->first_atom, 0xFF, pool->atom_size * pool->block_size );
#endif //DEBUG_CLEAR_FREE_MEM


  #ifdef SAFEMODE
    assert( pool->free_block );
  #endif //SAFEMODE

  pool->free_block_atoms = pool->block_size;

  #ifdef DEBUG_POOL
    printf( "<POOL> Expand pool %d by %d bytes (%d atoms)\n",
	    pool->id, pool->atom_size*pool->block_size,pool->block_size );
  #endif //DEBUG_POOL

  #ifdef ADVANCED
    stats_free += pool->block_size;
    stats_allocs--; // do not count pool allocs
  #endif //ADVANCED
}




//===================================================
/// \fn void return_to_pool( pool_t* pool, void* a ) 
/// \brief		returns atom back to its pool
/// \param pool		pool that will accept the atom
/// \param a		atom to be returned
///
/// Returns atom to the pool. The function assumes
/// that the atom size is the same as the pool's atom
/// size. Returned atoms are attached to the list and
/// never to the block.
//===================================================
void return_to_pool( pool_t* pool, void* a )
{
  //printf("!");
  #ifdef DEBUG_ATOM
  printf("<ATOM> -[%08x]\n",(int)a);
  #endif //DEBUG_ATOM

  #ifdef DEBUG_RUNTIME_ATOMS
  if( running_compiled_code )
    {
      outter( TEXT("<RUNTIME> free "), -1 );
      dump_atom_address( a );
      //dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
  #endif
  #ifdef DEBUG_COMPILETIME_ATOMS
  if( compiling_code )
    {
      outter( TEXT("<COMPILETIME> free "), -1 );
      dump_atom_address( a );
      //dump_atom( a, 1 );
      outter( TEXT("\n"), -1 );
    }
  #endif

#ifdef DEBUG_CLEAR_FREED_MEM
  memset( a, 0xFF, pool->atom_size );
#endif //DEBUG_CLEAR_FREE_MEM

  free_atom_t* b = a;
  b->next = pool->free_atom;
  pool->free_atom = b;

  //printf("delete %x\n",(int)b);

  //*((void**)a) = pool->free_atom;
  //pool->free_atom = a;

}




//===================================================
/// \fn void* take_from_pool( pool_t* pool )
/// \brief		takes new atom from a pool
/// \param pool		pool from which to allocate atom
/// \return		new atom from the pool
///
/// Takes atom from the pool. First attempts to take
/// it from the list. If empty it takes from the block.
/// If it is empty too then expands the	pool with a new
/// fresh block and takes atom from it.
//===================================================
void* take_from_pool( pool_t* pool )
{
  // take atom from the list
  free_atom_t* new = pool->free_atom;
  //printf(",");
  if( new )
    { // there is free atom in the list
      //printf("alloc  %x (next free is %x)\n",(int)new,(int)new->next);
      pool->free_atom = new->next;
    }
  else
    { // take atom from the block
      if( !pool->free_block ) expand_pool( pool );

      new = (free_atom_t*)((char*)pool->free_block + (-- pool->free_block_atoms)*pool->atom_size);
      if( pool->free_block_atoms == 0 )
	pool->free_block = NULL;
    }

  #ifdef DEBUG_ATOM
  printf( "<ATOM> +[%08x]\n",(int)new );
  #endif //ATOM

  #ifdef DEBUG_RUNTIME_ATOMS
  if( running_compiled_code )
    {
      outter( TEXT("<RUNTIME> alloc"), -1 );
      dump_atom_address( (atom_t)new );
      outter( TEXT("\n"), -1 );
    }
  #endif
  #ifdef DEBUG_COMPILETIME_ATOMS
  if( compiling_code )
    {
      outter( TEXT("<COMPILETIME> alloc"), -1 );
      dump_atom_address( (atom_t)new );
      outter( TEXT("\n"), -1 );
    }
  #endif

  return new;
}



#ifdef DEBUG_MEMORY_LEAKS
//===================================================
/// \fn void dump_pool( ) 
/// \brief		dumps pool's contents
///
/// Scans the pool and dumps all atoms which are
/// free. Assumes that the pool contains atom_t
/// elements.
//===================================================
void dump_pool( )
{
  void dump_list( atom_t a )
  {
    if( ID(a)==0xFF )
      printf("???");
    else
    if( IS_LIST(a) )
      {
	printf("[ ");
	while( ID(a)!=0xFF )
	  {
	    if( ID(CAR(a))!=0xFF )
	      {
		printf("|%d|{%x=",REF(a),(int)CAR(a));
		dump_list(CAR(a));
		printf("} ");
	      }
	    else
	      printf("??? ");
	    a = CDR(a);
	  }
	printf("]");
      }
    else
      dump(a);
  }

  void _dump_pool( pool_t p )
  {
    int i;
    atom_t a = p.first_atom;
    if( !a ) return;
    for( i=p.block_size; i; i--,a++ )
      {
	if( ID(a)!=0xFF )
	  {
	    switch( ID(a) )
	      {
	      case INTEGER_ID: printf("INT "); break;
	      case FLOAT_ID: printf("FLP "); break;
	      case LIST_ID: printf("LST "); break;
	      case WORD_ID: printf("WRD "); break;
	      case SUBWORD_ID: printf("SUB "); break;
	      case ERROR_ID: printf("ERR "); break;
	      case VAR_ID: printf("VAR "); break;
	      case MEM_ID: printf("MEM "); break;
	      default : printf("??? "); break;
	      }
	    printf("#%d[%x]=",REF(a),(int)a);
	    dump_list(a);
	    printf("\n");
	  }
      }
  }

  printf("\nUNUSED ATOMS:\n");
  _dump_pool( data_pool );
  printf("\nUNUSED EXTENDED ATOMS:\n");
  _dump_pool( data_pool_ex );
}
#endif //DEBUG_MEMORY_LEAKS
