
//
// Project:	Lhogho
// File:	unicode.c
//
// Copyright (C) 2007 P.Boytchev
//
// Revision history:
//   2006-10-09	- file created
//   2006-10-10 - fixed memory leaks
//   2006-10-11 - added UNICODE macro
//		  added ASCII_to_ASCII()
//   2007-02-27 - module renamed to UNICODE
//   2007-05-17 - added license info
//   2007-05-22	- added doxygen-friendly documentation
//   2007-06-07 - definitions spread to where they belong
//   2007-06-13	- fixed bug #1736021 "Alloc/dealloc statistics"
//   2010-02-07 - Fixed bug #3174945 OS alloc balance is not 0
//		          - Fixed bug #3174840 Invalid options cause a crash
//   2011-03-02 - Fixed bug #3188211 LOAD crashes if file size is >1MB
//   2012-10-14 - Fixed bug #314 -x compile on Windows needs lhogho.exe next to .lgo source file
//   2013-07-04 - If rt_libload() fails, it tries in lib subfolder of the compiler 
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include "globals.h"
#ifdef UNICODE_CHARS
  #include <wchar.h>	// UNICODE support: wcslen()
#endif //UNICODE_CHARS

#include <stdio.h>	// fopen(), fread(), fclose()
#include <stdlib.h>     // malloc()
#include <string.h>     // memcpy
#include <errno.h>	// errno
#include <assert.h>	// assert()
#include <sys/stat.h>	// stat()
#include <limits.h>	// PATH_MAX
#include <libgen.h>	// dirname()
 #if defined(WINDOWS) || defined(LINUX)
  #include <malloc.h>	// alloca()
#endif
#include "atoms.h"
#include "unicode.h"
#include "options.h"




//===================================================
/// \fn		char* UTF16_to_ASCII (chars_t ws)
/// \brief		converts UTF-16 to ASCII
/// \param	ws	characters to convert
/// \return		converted string
/// \note		function defined only if
//			UNICODE_CHARS symbol is defined
///
/// Converts string of widechar UTF-16LE encoding to
/// ASCII encoding. The input string is not freed.
//===================================================
#ifdef UNICODE_CHARS
char* UTF16_to_ASCII (chars_t ws)
{
  size_t len = wcslen( ws );
  char* buffer = alloca( 4*(len+1) ); // assume one utf16 can expand to 4 bytes max
  char* buf = buffer;
  mbstate_t state;
  size_t nbytes;

  memset (&state, '\0', sizeof (state));
  while (len>0)
    {
      nbytes = wcrtomb (buf, *ws, &state);
      buf += nbytes;
      len -= 1;
      ws += 1;
    }
  *buf = '\0';

  len = buf-buffer+1;
  buf = ALLOC( len );
  memcpy( buf, buffer, len );
  return buf;
}
#endif //UNICODE_CHARS




//===================================================
/// \fn		chars_t ASCII_to_UTF16 (const char *s)
/// \brief		converts ASCII to UTF-16
/// \param	s	characters to convert
/// \return		converted string
/// \note		defined only if UNICODE_CHARS
//			symbol is defined
///
/// Converts string of ASCII encoding to widechar
/// UTF-16LE encoding.
//===================================================
#ifdef UNICODE_CHARS
chars_t ASCII_to_UTF16 (const char *s)
{
  //printf("===%s===\n",s);

  size_t len = strlen(s);
  wchar_t *buffer = ALLOC( CHAR_SIZE*(len+1) );
  wchar_t *buf = buffer;
  mbstate_t state;
  size_t nbytes;
  memset (&state, '\0', sizeof (state));
  while (len>0)
    {
      nbytes = mbrtowc (buf, s, len, &state);
      buf++;
      len -= nbytes;
      s += nbytes;
    }
  *buf = L'\0';

  len = CHAR_SIZE*(buf-buffer+1);
  buf = ALLOC( len );
  memcpy( buf, buffer, len );
  DEALLOC( buffer );
  return buf;
}
#endif //UNICODE_CHARS




//===================================================
/// \fn		chars_t ASCII_to_ASCII (const char *s)
/// \brief		converts ASCII to ASCII
/// \param	s	characters to convert
/// \return		converted string
/// \note		defined only if UNICODE_CHARS
///			symbol is \b not defined
///
/// Converts string of ASCII encoding to ASCII. Actually
/// does not covert anything. This function is used
/// because it uses the ::ALLOC() macro which helps
/// tracing memory allocation.
//===================================================
chars_t ASCII_to_ASCII (const char *s)
{
#ifndef UNICODE_CHARS
  size_t len = STRLEN(s)+1;
  char* buf = ALLOC( len );
  memcpy( buf, s, len );
  return buf;
#else
  return (chars_t)0;
#endif //UNICODE_CHARS
}




//===================================================
/// \fn		chars_t UTF8_to_UTF16 ( byte_t* source )
/// \brief		converts UTF-8 to UTF-16
/// \param	source	characters to convert
/// \return		converted string
/// \note		defined only if UNICODE_CHARS
///			symbol is defined
///
/// Converts string of multibyte UTF-8 encoding to
/// widechar UTF-16LE encoding.
//===================================================
#ifdef UNICODE_CHARS
byte_t* UTF16_to_UTF8 (chars_t source )
{
  int len = STRLEN( source );
  byte_t *buffer = alloca( len+1 );
  byte_t *buf = buffer;
  
  while( len>0 )
    {
      char_t wc = *source;
      //printf("\nCODE=%4x|",wc);
      if( wc < 0x0080 )
	{ // 0000-007F
	  // from: [0xxxxxxx]
	  //   to: [0xxxxxxx]
	  *buf++ = (byte_t)wc; 
	}
      else if( wc < 0x0800 )
	{ // 0080-07FF
	  // from: [00000yyy yyxxxxxx]
	  //   to: [110yyyyy] [10xxxxxx]
	  *buf++ = 0xC0 | (byte_t)(wc >> 6);
	  *buf++ = 0x80 | (byte_t)(wc & 0x3F); 
	}
      else
	{ // 0800-FFFF
	  // from: [zzzzyyyy yyxxxxxx]
	  //   to: [1110zzzz] [10yyyyyy] [10xxxxxx]
	  *buf++ = 0xE0 | (byte_t)(wc >> 12);
	  *buf++ = 0x80 | (byte_t)((wc >> 6) & 0x3F);
	  *buf++ = 0x80 | (byte_t)(wc & 0x3F);
	}
      len--;
      source++;
    }

  *buf = '\0';

  len = buf-buffer+1;
  buf = ALLOC( len );
  memcpy( buf, buffer, len );

  return buf;
}

chars_t UTF8_to_UTF16 ( byte_t* source )
{
  int len = strlen ((char*)source);
  wchar_t *buffer = alloca( CHAR_SIZE*(len+1) );
  wchar_t *buf = buffer;
  unsigned long wc;

  while( len>0 )
    {
      if( (*source & 0x80)==0x00 )
	{ // 00-7F		[0zzz-zzzz]
	  wc = (byte_t)*source++;
	  len -= 1;
	}
      else if( (*source & 0xE0)==0xC0 )
	{ // 080-7FF		[110y-yyyy] [10zz-zzzz]
	  wc = *source & 0x1F;
	  source++;
	  wc = (wc<<6) + (*source & 0x3F);
	  source++;
	  len -= 2;
	}
      else if( ((byte_t)*source & 0xF0)==0xE0 )
	{ // 0800-FFFF		[1110-xxxx] [10yy-yyyy] [10zz-zzzz]
	  wc = *source & 0x1F;
	  source++;
	  wc = (wc<<6) + (*source & 0x3F);
	  source++;
	  wc = (wc<<6) + (*source & 0x3F);
	  source++;
	  len -= 3;
	}
      else
	{ // 01000-10FFFF	[1111-wwww] [10xx-xxxx] [10yy-yyyy] [10zz-zzzz]
	  wc = *source & 0x1F;
	  source++;
	  wc = (wc<<6) + (*source & 0x3F);
	  source++;
	  wc = (wc<<6) + (*source & 0x3F);
	  source++;
	  wc = (wc<<6) + (*source & 0x3F);
	  source++;
	  len -= 4;
	}
      *buf = (unsigned short)wc;
      buf++;
    }

  *buf = L'\0';

  len = CHAR_SIZE*(buf-buffer+1);
  buf = ALLOC( len );
  memcpy( buf, buffer, len );

  return buf;
}
#endif //UNICODE_CHARS




//===================================================
/// \fn		void* load_file( chars_t wfilename, int* filesize )
/// \brief		load text file into a word atom
/// \param	wfilename file name
/// \param	filesize  file size
/// \return		word atom
///
/// Loads a text file which can be ASCII, multibyte
/// UTF-8 or widechar UTF-16LE encoding. The size of
/// the file is returned in \c filesize so that the
/// caller can append null character if needed.
//===================================================
void* load_file( chars_t wfilename, int* filesize )
{
  FILE* file;		// file stream
  void* buffer;		// file buffer
  struct stat st_info;	// file attributes

  // convert filename to multibyte
  char* filename = FILENAME(wfilename);

  errno = 0;
  
  // open file
  file = fopen( filename, "rb" );
  if( errno )
	{
		// searching failed, try again looking in subfolder
		// lib of the folder where the compiler is
		char buf[PATH_MAX+1];
		strncpy(buf,option_compiler_filename,strlen(option_compiler_filename));
		char* path = dirname(buf);
		int pathlen = strlen(path);
		strncpy(buf,path,pathlen);
		
		int filelen = strlen(filename);
		if( pathlen+filelen+6 > PATH_MAX ) filelen=0;
#ifdef WINDOWS
		strncpy(buf+pathlen,"\\lib\\",5);
#else
		strncpy(buf+pathlen,"/lib/",5);
#endif
		strncpy(buf+5+pathlen,filename,filelen);
		buf[pathlen+5+filelen] = '\0';
		
		//printf("failed %s, try '%s' while compiler is %s\n",filename,buf,option_compiler_filename);
		errno = 0;
		file = fopen( buf, "rb" );
		if( errno )
		{
			//printf("failed again\n");
			return NULL;
		}
	}

  // get file size
  fstat( fileno(file), &st_info );
  *filesize = st_info.st_size;

  // file name is not needed any more
  #ifdef UNICODE_CHARS
    DEALLOC( filename );
  #endif

  // allocate buffer
  buffer = ALLOC(*filesize+1);
  #ifdef SAFEMODE
  if( !buffer )
    {
      fclose( file );
      errno = ENOMEM;
      return NULL;
    }
  #endif //SAFEMODE

  // read file into the buffer
  if( *filesize && !fread(buffer,1,*filesize,file ) )
    {
      #ifdef SAFEMODE
        DEALLOC( buffer );
        return NULL;
      #endif //SAFEMODE
    }

  fclose( file );
  #ifdef SAFEMODE
  if( errno )
    {
      free( buffer );
      return NULL;
    }
  #endif //SAFEMODE

  return buffer;
}






