//
// Project:	Lhogho
// File:	unicode.h
//
// Copyright (C) 2007 P.Boytchev
//
// Revision history:
//   2006-10-08	- file created
//   2006-10-11 - UNICODE macro
//		  STRLEN,STRNCPY,SPRINTF,PUTCHAR,FILENAME macros
//		  STR,NULL_CHAR macros
//		  extern to ASCII_to_ASCII()
//		  SPACE, TAB macros
//   2006-10-17 - CHR macro
//   2006-10-26 - FORMAT_ERR
//   2007-02-25 - split FORMAT_ERR to FORMAT_ERR_CODE
//		  and FORMAT_ERR_POS
//   2007-02-27 - module renamed to UNICODE
//		- TEXT()
//   2007-05-17 - license info
//   2007-05-22	- doxygen-friendly documentation
//   2007-05-26 - FORMAT_ERR_TEXT macro
//		- CLONE macro
//   2007-06-02 - UNFILENAME macro
//   2007-06-06 - definitions spread to where they belong
//   2007-06-25 - STRTOD macro
//   2007-06-30 - fixed bug #1745627 Option -Zm does not show memory (all Linuxes)
//		- fixed bug #1745786 -Zm option does not work in Cygwin
//   2007-09-14 - added TOUPPER and TOLOWER
//   2007-10-28 - added ISDIGIT and STRCHR
//   2009-06-02 - added GETCHAR
//   2012-01-23 - added STRFTIME
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#ifndef LHOGHO_UNICODE_H
#define LHOGHO_UNICODE_H


#include "globals.h"
#include "atoms.h"
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#ifdef UNICODE_CHARS
  #include <wchar.h>
#endif




//===================================================
/// \name Converters
/// Used to convert one string to another. In ASCII
/// mode all conversions are like identities - i.e.
/// no conversion is actually done.
//===================================================
#ifdef UNICODE_CHARS
  #define TEXT(a)	L##a		  ///< fix string constants
  #define FILENAME(x)	UTF16_to_ASCII(x) ///< fix file names
  #define UNFILENAME(x)	ASCII_to_UTF16(x) ///< unfix file names
#else
  #define TEXT(a)	a		  ///< fix string constants
  #define FILENAME(x)	(x)		  ///< fix file names
  #define UNFILENAME(x)	ASCII_to_ASCII(x) ///< unfix file names
#endif




//===================================================
/// \name Format strings
//===================================================
#define NULL_CHAR	TEXT('\0')
#ifdef UNICODE_CHARS
  #define STR	"%S"
  #define CHR	"%C"
#else
  #define STR	"%s"
  #define CHR	"%c"
#endif //UNICODE_CHARS




//===================================================
/// \name String functions
///
/// They are used to maintain dual UNICODE/ASCII
/// processing. This is needed because there are
/// different names for mutibyte and widechar functions.
//===================================================
#ifdef UNICODE_CHARS
  #define NO_MORE		WEOF
  #define PUTCHAR(x,y)		  putwc(x,y)
  #define GETCHAR(x)		    getwc(x)
  #define STRLEN(x)		      wcslen(x)
  #define STRNCPY(x,y,z)	  wcsncpy(x,y,z)
//#define STRNCMP(x,y,z)	  wcsncmp(x,y,z)
  #define STRCMP(x,y)	      wcscmp(x,y)
  #define STRTOD(x,y)		    wcstod(x,y)
  #define STRTOL(x,y)		    wcstoll(x,y,0)
  #define TOUPPER(x)        towupper(x)
  #define TOLOWER(x)        towlower(x)
  #define ISDIGIT(x)        iswdigit(x)
  #define STRCHR(x,y)       wcschr(x,y)
  #define PRINT(x,...)	    printf(x,__VA_ARGS__)
  #define STRFTIME(x,y,z,t) wcsftime(x,y,z,t)
  #ifdef WINDOWS
    #define SPRINTF(x,n,y,z)	swprintf(x,y,z)
    #define SPRINT(x,n,y,...)	swprintf(x,y,__VA_ARGS__)
  #else
    #define SPRINTF(x,n,y,z)	swprintf(x,n,y,z)
    #define SPRINT(x,n,y,...)	swprintf(x,n,y,__VA_ARGS__)
  #endif
#else
  #define NO_MORE		EOF
  #define STRFTIME(x,y,z,t) strftime(x,y,z,t)
  #define PRINT(x,...)	  	printf(x,__VA_ARGS__)
  #define PUTCHAR(x,y)	  	putc(x,y)
  #define GETCHAR(x)    		getc(x)
  #define STRLEN(x)	  	    strlen(x)
  #define TOUPPER(x)	    	toupper(x)
  #define TOLOWER(x)  		  tolower(x)
  #define STRCHR(x,y)	    	strchr(x,y)
  #define ISDIGIT(x)  	  	isdigit(x)
  #define STRNCPY(x,y,z)  	strncpy(x,y,z)
//  #define STRNCMP(x,y,z)	strncmp(x,y,z)
  #define STRCMP(x,y)	      strcmp(x,y)
  #define STRTOD(x,y)   		strtod(x,y)
  #define STRTOL(x,y)		    strtoll(x,y,0)
//  #define USTRNCMP(x,y,z)	strncasecmp(x,y,z)
#ifdef WINDOWS
  #define SPRINTF(x,n,y,z)	_snprintf(x,n,y,z)
#else
  #define SPRINTF(x,n,y,z)	snprintf(x,n,y,z)
#endif
  #define SPRINT(x,n,y,...)	snprintf(x,n,y,__VA_ARGS__)
#endif //UNICODE_CHARS


//===================================================
/// \name Functions
//===================================================
#ifdef UNICODE_CHARS
extern chars_t UTF8_to_UTF16  ( byte_t* source );
extern byte_t* UTF16_to_UTF8 (chars_t source );
extern char*   UTF16_to_ASCII ( chars_t ws );
extern chars_t ASCII_to_UTF16 (const char* s);
#ifdef LINUX
extern long long int wcstoll (__const wchar_t *__restrict __nptr,
			      wchar_t **__restrict __endptr, int __base);
extern int swprintf (wchar_t *__restrict __s, size_t __n,
		     __const wchar_t *__restrict __format, ...);
extern size_t wcslen (__const wchar_t *__s) __THROW __attribute_pure__;
extern int fwide (__FILE *__fp, int __mode) __THROW;
extern int wprintf (__const wchar_t *__restrict __format, ...);
extern int wcsncmp (__const wchar_t *__s1, __const wchar_t *__s2, size_t __n)
     __THROW __attribute_pure__;
extern int wcsncasecmp (__const wchar_t *__s1, __const wchar_t *__s2,
			size_t __n) __THROW;
extern wchar_t *wcsncpy (wchar_t *__restrict __dest,
			 __const wchar_t *__restrict __src, size_t __n)
     __THROW;
#endif //LINUX
#endif //UNICODE_CHARS

extern chars_t ASCII_to_ASCII ( const char* s );
extern void*   load_file( chars_t wfilename, int* filesize );


#endif //LHOGHO_UNICODE_H
