//===================================================
// Project:TestSystem   Author: Peter Armianov
//===================================================
//
// PARAM_PARSERS.C
//
//  This file contains definitions of funcions which
// parse one configuration parameter line
//
// Revision history:
//   2007-05-21	- file created
//   2007-07-12 - additional comments in Doxygen style
//===================================================
#include <stdio.h>
#include <stdlib.h>

#include "error.h"
#include "TestMaker.h"
#include "internal.h"
#include "param_parsers.h"

//---------------------------------------------------
//  Parser functions implementation
// For each new parameter here must be placed  
// function that parse the line.
// Function name must be such like
// <param>_parser
// where <param> is the name used in PARSERS_LIST macro
//---------------------------------------------------

//===================================================
/// \brief	Parse name pameter
/// \param  parameter_name  name of the parameter to parse
/// \param  parameter_optipons  the text passed for parsing
/// \param  test_case_params    palce where extracted name will be stored
/// \return RESULT code (see error.h)
///
/// Exract the name of the test case. 
//===================================================
RESULT name_parser(const TCHAR * parameter_name,
                   const TCHAR * parameter_optipons,
                   test_case_info * test_case_params)
{
    /* if this is not parameter that can be handled here */
    if (m_strcmp(parameter_name, TEXT("name")))
    {
        return ERR_INVALID_ARG;
    }

    return extract_string(parameter_optipons, &test_case_params->test_name);
}

//===================================================
/// \brief	Parse commands for compiler
/// \param  parameter_name  name of the parameter to parse
/// \param  parameter_optipons  the text passed for parsing
/// \param  test_case_params    palce where extracted comand options will be stored
/// \return RESULT code (see error.h)
///
/// Exract the name of the test case. 
//===================================================
RESULT cmd_parser(const TCHAR * parameter_name,
                  const TCHAR * parameter_optipons,
                  test_case_info * test_case_params)
{
    /* if this is not parameter that can be handled here */
    if (m_strcmp(parameter_name, TEXT("cmd")))
    {
        return ERR_INVALID_ARG;
    }

    return extract_string(parameter_optipons, &test_case_params->command_line_param);
}


//----------------------------------------------------
// Basic functions implementation
//----------------------------------------------------

//===================================================
/// \fn RESULT extract_string(const TCHAR * parameter_optipons, TCHAR ** extracted_value)
/// \brief	Extracts a string to space form parameter buffer
/// \param  parameter_optipons  input string 
/// \param  extracted_value  result value. Memory is allocated for it
/// \return RESULT code (see error.h)
///
/// Exract the the data to the first space. 
/// If space is between " " it is ecraned.
/// Allocate memory for result string
//===================================================
RESULT extract_string(const TCHAR * parameter_optipons, TCHAR ** extracted_value)
{
    UINT32      start = 0;
    UINT32      end = 0;
    BOOL        escape_spaces = FALSE;

    while (parameter_optipons[start] && m_isspace(parameter_optipons[start]))
        ++start;
    
    if (parameter_optipons[start] == SPACE_ESC_SYMBOL)
    {
        ++start;
        escape_spaces = TRUE;
    }
    end = start;

    while (parameter_optipons[end] && 
           !((escape_spaces && parameter_optipons[end] == SPACE_ESC_SYMBOL) ||
             (!escape_spaces && m_isspace(parameter_optipons[end]))))
        ++end;

    return m_strndup(extracted_value, parameter_optipons + start, end - start);
}

