//===================================================
// Project:TestSystem   Author: Peter Armianov
//===================================================
//
// STRING_FUNCTIONS.C
//
//  This file contains definitions of helper funcions 
// working with TCHAR type
//
// Revision history:
//   2007-05-21	- file created
//   2007-06-25 - several bug fixes
//   2007-06-30 - several bug fixes
//   2007-07-12 - additional comments in Doxygen style
//===================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>

#include "error.h"
#include "TestMaker.h"


//===========================================================
/// \fn		size_t m_strlen(const TCHAR * str)
/// \brief		calculates string length
/// \param   str	the string
/// \return	length of the string in symbols
///
/// Calculates string length in symbols. Works with char and w_char
//===========================================================
size_t m_strlen(const TCHAR * str)
{
    const TCHAR * start = str;
    while (*str) ++str;
    return str - start;
}


//===========================================================
/// \fn		int m_strcmp(const TCHAR * str1, const TCHAR * str2)
/// \brief		compares two strings
/// \param   str1	first string
/// \param   str2	second string
/// \return comparison result similar to standart strcmp result
///
/// Comare two strings. 
/// Return negative value if first is less than second
/// positive value if first is great than second
/// and zero if strings are equal
//===========================================================
int m_strcmp(const TCHAR * str1, const TCHAR * str2)
{
    while (*str1 && *str1 == *str2)
    {
        ++str1;
        ++str2;
    }
    return *str1 - *str2;
}


//===========================================================
/// \fn     TCHAR * m_strcpy(TCHAR * dest, const TCHAR * src)
/// \brief  copy string
/// \param  dest	destination string
/// \param  src     source string
/// \return pointer to destination
///
/// Copy string value to given pointer.
/// User must supply enought memory
//===========================================================
TCHAR * m_strcpy(TCHAR * dest, const TCHAR * src)
{
    TCHAR * mem = dest;
    while (*src)
    {
        *dest++ = *src++;
    }
    *dest = 0;
    return mem;
}


//===========================================================
/// \fn     BOOL m_isspace(TCHAR chr)
/// \brief  test if char is space
/// \param  chr char to test
/// \return result of the test
///
/// Test if symbol is space symbol (Space, new line or tab)
//===========================================================
BOOL m_isspace(TCHAR chr)
{
    return (chr == TEXT(' ')  || chr == TEXT('\t') || 
        chr == TEXT('\r') || chr == TEXT('\n'));
}


//===========================================================
/// \fn     RESULT m_strndup(TCHAR ** dest_ptr, const TCHAR * src, size_t num_chars)
/// \brief  duplicate string, but no more than \c num_chars symbols
/// \param  dest_ptr pointer to destination string
/// \param  src     source string
/// \param  num_chars number of characters to copy.
/// \return RESULT value. Error on memory problems
///
/// Copy string value to given pointer, byt no more than \c num_chars symbols.
/// Allocates memory for destination. 
/// Not copy the 0 byte if source is longer than wanted length
//===========================================================
RESULT m_strndup(TCHAR ** dest_ptr, const TCHAR * src, size_t num_chars)
{
    *dest_ptr = (TCHAR *)malloc((num_chars + 1) * sizeof(TCHAR));
    if (!*dest_ptr)
    {
        return ERR_MEMORY;
    }
    memcpy(*dest_ptr, src, num_chars * sizeof(TCHAR));
    (*dest_ptr)[num_chars] = 0;
    return SUCCESS_FULL;
}


//===========================================================
/// \fn     RESULT m_strdup(TCHAR ** dest_ptr, const TCHAR * src)
/// \brief  duplicate string. Allocate memory for destination
/// \param  dest_ptr pointer to destination string
/// \param  src     source string
/// \return RESULT value. Error on memory problems
///
/// Duplicate string. Allocate memory for destination. 
/// If memory problems occured return error.
//===========================================================
RESULT m_strdup(TCHAR ** dest_ptr, const TCHAR * src)
{
    size_t len = m_strlen(src);
    return m_strndup(dest_ptr, src, len);
}


//===========================================================
/// \fn     TCHAR * m_fgets(FILE* file, TCHAR * buffer, size_t buffer_size, size_t * out_size_ptr)
/// \brief  Read a line from file
/// \param  file file to read from
/// \param  buffer buffer to read in
/// \param  buffer_size size of space in the buffer
/// \param  out_size_ptr size of line readed
/// \return pointer to readed string or NULL if error
///
/// Read characters from given open file until new line symbol is reached
/// or input buffer is full. 
/// Returns the count of really readed symbols in parameter and pointer to the buffer.
/// If error occured during read process returns NULL
//===========================================================
TCHAR * 
m_fgets(FILE* file, TCHAR * buffer, size_t buffer_size, size_t * out_size_ptr)
{
    TCHAR c = 0;
    size_t pos = 0;
    if (!file || !buffer || !buffer_size || !out_size_ptr)
    {
        LOG_ERROR("Invalid args supplied to m_fgets");
        if (out_size_ptr)
        {
            *out_size_ptr = 0;
        }
        return NULL;
    }

    while (pos < buffer_size - 1)
    {
        if (!fread(&c, sizeof(TCHAR), 1, file))   
        {
            break;
        }
        buffer[pos++] = c;

        if (c == TEXT('\r') || c == TEXT('\n'))
        {
            TCHAR c1;
            size_t readed;
            if ((readed = fread(&c1, sizeof(TCHAR), 1, file)) &&
                (c1 == TEXT('\r') || c1 == TEXT('\n')) && c1 != c)
            {
                buffer[pos++] = c1;
            }
            else
            {
                if (readed)
                {
                    fseek(file, -1l, SEEK_CUR);
                }
            }
            break;
        }
    }

    buffer[pos] = 0;
    *out_size_ptr = pos;
    return pos ? buffer : NULL;
}


//===========================================================
/// \fn     void m_fputs(FILE* file, const TCHAR * buffer)
/// \brief  prints a line to file
/// \param  file file to write in
/// \param  buffer string to write
///
/// Writes a line to the file. Buffer must be a valid nul-terminated string
//===========================================================
void
m_fputs(FILE* file, const TCHAR * buffer)
{
    size_t len = m_strlen(buffer);
    fwrite(buffer, sizeof(TCHAR), len, file);
}

//===========================================================
/// \fn     void m_fputs_ascii(FILE* file, const char * buffer)
/// \brief  prints a line to file, convert it to UNICODE if needed
/// \param  file file to write in
/// \param  buffer string to write
///
/// Writes a line to the file. If program works in unicode mode 
/// the string is converted from ASCII to unicode.
//===========================================================
void
m_fputs_ascii(FILE* file, const char * buffer)
{
    char * buf = (char *)buffer;
    size_t len = strlen(buffer) * sizeof(TCHAR);
    
#if defined(UNICODE)
     buf = (char*)malloc(len);
     while (*buffer) *buf++ = *buffer++, *buf++ = 0;
#endif
    fwrite(buf, sizeof(TCHAR), len, file);

#if defined(UNICODE)
    free(buf);
#endif
}


//===========================================================
/// \fn     void m_fputc(FILE* file, TCHAR tchar)
/// \brief  prints a symbol to file
/// \param  file file to write in
/// \param  tchar symbol to write
///
/// Writes a symbol to file.
//===========================================================
void
m_fputc(FILE* file, TCHAR tchar)
{
#if defined (UNICODE)
    putwc(tchar, file);
#else
    putc(tchar, file);
#endif
}
